unit s_changemodel;

{$if defined(lazarus)}
{$mode objfpc}{$H+}
{$ifend}

interface

uses
  Classes, SysUtils,
  Slib, s_base, s_dat, digraph, s_constants;


type

 ChangeModelSpecification = object(modelSpecification)
   procedure FileChange(var outf,lf: TFileStream; st3: shortstring; var OK: boolean);
   procedure LineChange(var outf,lf: TFileStream; st2,st3,st4,st5: shortstring; var OK: boolean);
 end; (* ChangeModelSpecification *)

implementation

procedure ChangeModelSpecification.LineChange(var outf,lf: TFileStream; st2,st3,st4,st5: shortstring; var OK: boolean);
var pst2, pst3, pst4, pst5, code, action: integer;
    rst3, rst5: real;
    x: string;
begin
//  writeln(lf,'LineChange ',pname,' ',st2,' ',st3,' ',st4,' ',st5);
  pst2 := 0;
  OK := true;
  if (st2 = 'p') then
    begin
    pst2 := 1;
    pst3 := -1;
    if (st3 = 'r') then pst3 := 1;
    if (st3 = 'v') then pst3 := 2;
    if (st3 = 'e') then pst3 := 3;
    if (st3 = 'p') then pst3 := 4;
    end;
  if (st2 = 'st') then
    begin
    pst2 := 2;
    val(st3,pst3,code);
    if (code <> 0) then pst3 := -1;
    end;
  if (st2 = 'mc') then
    begin
    pst2 := 3;
    val(st3,pst3,code);
    if (code <> 0) then pst3 := 999;
    end;
  if (st2 = 'est') then
    begin
    pst2 := 4;
    val(st3,pst3,code);
    if (code <> 0) then pst3 := 999;
    end;
  if (st2 = 'muf') then
    begin
    pst2 := 5;
    val(st3,rst3,code);
    if (code <> 0) then rst3 := 999.0;
    end;
  if (st2 = 'nsub2') then
    begin
    pst2 := 6;
    val(st3,rst3,code);
    if (code <> 0) then rst3 := 999.0;
    end;
  if (st2 = 'nrun3') then
    begin
    pst2 := 7;
    val(st3,rst3,code);
    if (code <> 0) then rst3 := 999.0;
    end;
  if (st2 = 'seed') then
    begin
    pst2 := 8;
    val(st3,rst3,code);
    if (code <> 0) then rst3 := 999.0;
    end;
  if (st2 = 'ig') then
    begin
    pst2 := 9;
    val(st3,rst3,code);
    if (code <> 0) then rst3 := 999.0;
    end;

  case pst2 of
    1: begin    (* parameter value *)
       if (st5 >= '') then
         begin
         val(st4,pst4,code);
         if (code <> 0) then
           begin
           x := 'Error: the fourth parameter for option <p> must be an integer number.'+eol;
           lf.Write(pchar(x)^,length(x));
           OK := false;
           end;
         action := 0;
         code := 0;
         if (st5 = '-') then action := 1;
         if (st5 = '+') then action := 2;
         if (st5 = 't') then action := 3;
         if (st5 = 'e') then action := 4;
         if (action = 0) then
           begin
           val(st5,rst5,code);
           if (code <> 0) then
             begin
             x := 'Error: the fifth parameter for option <p> must be -, +, or a real number.'+eol;
             lf.Write(pchar(x)^,length(x));
             wwriteln(lf,'parameters '+st2+', '+st3+', '+st4+', '+st5+'; code = '+inttostr(code));
             wwriteln(lf,'ord(-) = '+inttostr(ord('-')));
             OK := false;
             exit;
             end;
           pst5 := round(rst5);
           end; (* if *)

         case pst3 of
           1: begin (* rate function *)
              if ((pst4 >= 1) and (pst4 <= nxa*NetworkEffects_l + nza*ActionEffects_l)) then
                case action of
                    0: alpa_l[pst4] := rst5;
                    1: effects_l[pst4] := false;
                    2: effects_l[pst4] := true;
                    3: begin
                       fix_l[pst4] := true;
                       test_l[pst4] := true;
                       end;
                    4: begin
                       fix_l[pst4] := false;
                       test_l[pst4] := false;
                       end;
                    end (* case *)
                 else
                   begin
                   x := format('%s%d%s',['Parameter number ',pst4,' does not exist for the rate function.'])+eol;
                   lf.Write(pchar(x)^,length(x));
                   OK := false;
                   end;
                end;
           2: begin (* evaluation function *)
              if ((pst4 >= 1) and (pst4 <= nxa*NetworkEffects_f + nza*ActionEffects_f)) then
                case action of
                  0: alpa_f[pst4] := rst5;
                  1: effects_f[pst4] := false;
                  2: effects_f[pst4] := true;
                  3: begin
                     fix_f[pst4] := true;
                     test_f[pst4] := true;
                     end;
                  4: begin
                     fix_f[pst4] := false;
                     test_f[pst4] := false;
                     end;
                  end (* case *)
                 else
                   begin
                   x := format('%s%d%s',['Parameter number ',pst4,' does not exist for the evaluation function.'])+eol;
                   lf.Write(pchar(x)^,length(x));
                   OK := false;
                   end;
              end;
           3: begin (* endowment function *)
              if ((pst4 >= 1) and (pst4 <= nxa*NetworkEffects_g + nza*ActionEffects_g)) then
                case action of
                  0: alpa_g[pst4] := rst5;
                  1: effects_g[pst4] := false;
                  2: effects_g[pst4] := true;
                  3: begin
                     fix_g[pst4] := true;
                     test_g[pst4] := true;
                     end;
                  4: begin
                     fix_g[pst4] := false;
                     test_g[pst4] := false;
                     end;
                  end (* case *)
                 else
                   begin
                   x := format('%s%d%s',['Parameter number ',pst4,' does not exist for the endowment function.'])+eol;
                   lf.Write(pchar(x)^,length(x));
                   OK := false;
                   end;
              end;
           4: begin (* parameter parameter *)
              if ((pst4 >= 1) and (pst4 <= nxa*NetworkEffects_f + nza*ActionEffects_f)) then
                begin
                with NewNetworkEffects do
                  begin
                  SetParam(pst4,pst5);
                  param_effect[pst4] := pst5;
{            else
             begin
             writeln(lf,'Parameter value ',pst5,' is not permitted for effect number ',pst4,'.');
             OK := false;
 }                end;
                end
               else
                begin
                x := format('%s%d%s',['Parameter number ',pst3,' does not exist for the evaluation function.'])+eol;
                lf.Write(pchar(x)^,length(x));
                OK := false;
                end;
              end;
           end; (* case pst3 *)
         end
        else
         begin
         OK := false;
         end;
       end; (* case pst2 = 1 *)
    2: begin   (* starting value *)
       case pst3 of
         0: Initialstandard := false;
         1: Initialstandard := true;
        else
         begin
         x := 'For code <st>, which means change of initial value option,'+eol;
         lf.Write(pchar(x)^,length(x));
         if (st3 = '') then  begin
           x := 'a third code string is required after <st>, but none was found.'+eol;
           lf.Write(pchar(x)^,length(x));
         end
       else begin
          x := 'the next code was read as <'+st3+'>, but only values 0 and 1 are allowed.'+eol;
          lf.Write(pchar(x)^,length(x));
       end;
         OK := false;
         end;
        end; (* case pst3 *)
       end; (* case pst2 = 2 *)
    3: begin    (* model code *)
       if (pst3 = 999) then
         begin
         OK := false;
         end
        else
         begin
         Gibbscode := pst3;
         if signed then Modeltype := -pst3 else Modeltype := pst3;
         (* The following options are so defined that Modeltype
            will not be changed during WriteModelOptions *)
         OutputType := 0;
         ReciprocityModel := false;
         StaySameType := 0;
         end;
         end; (* case pst2 = 3 *)
    4: begin   (* estimation method *)
       if (pst3 = 999) then
         begin
         OK := false;
         end
        else
         begin
         maxdegree := pst3 div 1000;
         MakeStep3 := (maxdegree >= 1);
         if MakeStep3 then pst3 := pst3 - 1000;
         maxdegree := pst3 div 100;
         if (maxdegree >= 1) then pst3 := pst3 - 100*maxdegree;
         if (maxdegree <= 1) then maxdegree := high(integer);
         CConditional := ((pst3=1) or (pst3>=21));
         if (exo and Cconditional) then  CConditional := false;
         if Cconditional
         then begin
           if pst3=1 then pst3:=21;
           if ((pst3 - 20) > (nxa+nza))
           then begin
             ConditioningNetworkVariable := 1;
             ConditioningActionVariable := 0;
           end (* if *)
           else begin
             ConditioningNetworkVariable := 0;
             ConditioningActionVariable := 0;
             if ((pst3 - 20) <= nxa)
             then ConditioningNetworkVariable := pst3 -20
             else ConditioningActionVariable := pst3 - nxa - 20;
           end (* else *)
         end;

         MaxLike := ((pst3 = 2) and (observations >= 2));
         Bayesian := ((pst3 = 3) and (observations >= 2));
         if (MaxLike or Bayesian) then
           begin
           CConditional := false;
           continuous := true;
           if exo then
             begin
             MaxLike := false;
             Bayesian := false;
             end;
           end;

         if (not cconditional) then
           begin
           ConditioningNetworkVariable := 0;
           ConditioningActionVariable := 0;
           end;

         VertexPar := ((pst3 in [10,11]) and (observations = 1));
         end;
         end; (* case pst2 = 4 *)
    5: begin   (* multiplication factor *)
       if (rst3 <> 999.0) then rstep0 := rst3;
       end; (* case pst2 = 5 *)
    6: begin   (* number of subphases in phase 2 *)
       if (rst3 <> 999.0) then nsubphase := trunc(rst3);
       end; (* case pst2 = 6 *)
    7: begin   (* number of runs in phase 3 *)
       if (rst3 <> 999.0) then nrun3 := trunc(rst3);
       end; (* case pst2 = 7 *)
    8: begin   (* random number seed *)
       if (rst3 <> 999.0) then modelseed := trunc(rst3);
       end; (* case pst2 = 8 *)
    9: begin   (* initial gain factor *)
       if (rst3 <> 999.0) then firstgain := rst3;
       end; (* case pst2 = 9 *)
      else
       begin
       OK := false;
       end;
     end; (* case pst2 *)

    if not OK then
      begin
      x := 'No valid interpretation of codes '+pname+' '+st2+' '+st3+' '+st4+' '+st5+eol;
      lf.Write(pchar(x)^,length(x));
      x := 'Siena03 is called, with the following codes, which could not be interpreted:'+eol;
      outf.Write(pchar(x)^,length(x));
      x := pname+' '+st2+' '+st3+' '+st4+' '+st5+eol;
      outf.Write(pchar(x)^,length(x));
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      end
     else   begin
        x := 'Siena03 executed with codes '+pname+' '+st2+' '+st3+' '+st4+' '+st5+eol;
       lf.Write(pchar(x)^,length(x));
     end;
end; (*  ChangeModelSpecification.LineChange *)

procedure ChangeModelSpecification.FileChange(var outf,lf: TFileStream; st3: shortstring;
                        var OK: boolean);
var  s2, s3, s4, s5: shortstring;
     f: textfile;
     a: char;
     stn: integer;
     OKf, prevbl: boolean; (* prevbl = previous blank, to allow separation by multiplpe blanks *)
     x: string;
begin
  OK := false;
  if not FileExists(st3) then
    begin
    x := 'No valid interpretation of codes '+pname+' '+'f'+' '+st3+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'Siena03 is called, with the following codes, which could not be interpreted:'+eol;
    lf.Write(pchar(x)^,length(x));
    x := pname+' '+'f'+' '+st3+eol;
    lf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    exit;
    end;

  assignfile(f,st3);
  reset(f);
  while not(eof(f)) do
    begin
    s2 := '';
    s3 := '';
    s4 := '';
    s5 := '';
    stn := 2;
    prevbl := false;
    while not eoln(f) do
      begin
      read(f,a);
      if (a = ' ') then
        begin
        if (not prevbl) then inc(stn);
        prevbl := true;
        end
       else
        begin
        prevbl := false;
        if (ord(a) >= 33) then
          case stn of
          2: s2 := s2+a;
          3: s3 := s3+a;
          4: s4 := s4+a;
          5: s5 := s5+a;
          end; (* case *)
        end;
      end; (* while *)
    LineChange(outf,lf,s2,s3,s4,s5,OKf);
    if Okf then OK := true;
    readln(f);
    end; (* while *)
  closefile(f);
end; (*  ChangeModelSpecification.FileChange *)



end.

