                        (* S_desc.pas *)

unit s_desc;

(* Some decriptive statistics for SIENA *)

interface
uses Forms, Classes, SysUtils, eight, slib, s_dat, s_base, digraph;

type intmat = array of array of integer;

type
 Three = 0..3;
 TriadVector = array[1..16] of Real;
 TriadType = array[1..3] of Three;
   (* Triads indicated in the Holland-Leinhardt notation by
      MAN or MANX, where M,A,N are numbers 0,1,2,3 and X is C,T,U, or D,
      are coded here by M,A,X since 2*M+A=3.
      Codes for X are: nothing = 0, C = 0, T = 1, U = 1, D = 2 *)
 TriadCountType = array[1..16] of Longint;
 TriadArrayType = array[1..16] of TriadType;

 DataDescription = object(ModelSpecification)
  missing: boolean; (* indicates that any are missing *)

  procedure Describe(var outf,lf: textfile);
(* Procedure Describe is called in Siena_2. *)
  procedure MatrixSum(var matsum: Intmat);
  (* sum of digraphs over the observations *)
  function HTrans(v,s: real): real;
(* Transforms degree variance and degree sum s to standardized H *)
  function rowvar(k: integer): real;
(* row degree variance *)
  function colvar(k: integer): real;
(* column degree variance *)
  function Isolates(k: integer): integer;
(* number of isolates *)
  function Components(k: integer): integer;
(* number of components consisting of >= 2 vertices *)
  procedure DetermineTriad(var i,j,k: integer; m: integer; var t: TriadType);
  procedure TriadCount(m: integer; var c: triadcounttype;
                        var p: TriadVector; var tottriad: cardinal);
  procedure Spatials(var outf:textfile; sumf: TFileStream);
(* Spatial autocorrelation statistics *)
  procedure DescribeStatistics(var outf,lf: textfile;sumf:TFileStream);
  procedure DescribeWithCovariates(var outf: textfile);

end; (* DataDescription *)

 function TriadName(var tr: TriadType): string;
 function TriadNumber(var tr: TriadType): Byte;
 function TriadCode(i: byte): string;


implementation

var  TriadKind: TriadArrayType;


procedure InitTriad(var Triad: TriadArrayType);
var t: TriadType;
begin
  t[1] := 0; t[2] := 0; t[3] := 0; Triad[1] := t;
  t[1] := 0; t[2] := 1; t[3] := 0; Triad[2] := t;
  t[1] := 1; t[2] := 0; t[3] := 0; Triad[3] := t;
  t[1] := 0; t[2] := 2; t[3] := 2; Triad[4] := t;
  t[1] := 0; t[2] := 2; t[3] := 1; Triad[5] := t;
  t[1] := 0; t[2] := 2; t[3] := 0; Triad[6] := t;
  t[1] := 1; t[2] := 1; t[3] := 2; Triad[7] := t;
  t[1] := 1; t[2] := 1; t[3] := 1; Triad[8] := t;
  t[1] := 0; t[2] := 3; t[3] := 1; Triad[9] := t;
  t[1] := 0; t[2] := 3; t[3] := 0; Triad[10] := t;
  t[1] := 2; t[2] := 0; t[3] := 0; Triad[11] := t;
  t[1] := 1; t[2] := 2; t[3] := 2; Triad[12] := t;
  t[1] := 1; t[2] := 2; t[3] := 1; Triad[13] := t;
  t[1] := 1; t[2] := 2; t[3] := 0; Triad[14] := t;
  t[1] := 2; t[2] := 1; t[3] := 0; Triad[15] := t;
  t[1] := 3; t[2] := 0; t[3] := 0; Triad[16] := t;
end;

function TriadName(var tr: TriadType): string;
begin
  case tr[1] of
    0:  case tr[2] of 0: TriadName := '003 ';
                      1: TriadName := '012 ';
                      2: case tr[3] of 0: TriadName := '021C';
                                       1: TriadName := '021U';
                                       2: TriadName := '021D';
                                       else TriadName := 'XXXX';
                         end;
                      3: case tr[3] of 0: TriadName := '030C';
                                       1: TriadName := '030T';
                                       else TriadName := 'XXXX';
                         end;
                      else TriadName := 'XXXX';
        end;
    1:  case tr[2] of 0: TriadName := '102 ';
                      1: case tr[3] of 1: TriadName := '111U';
                                       2: TriadName := '111D';
                                       else TriadName := 'XXXX';
                         end;
                      2: case tr[3] of 0: TriadName := '120C';
                                       1: TriadName := '120U';
                                       2: TriadName := '120D';
                                       else TriadName := 'XXXX';
                         end;
                      else TriadName := 'XXXX';
        end;
    2:  case tr[2] of 0: TriadName := '201 ';
                      1: TriadName := '210 ';
                      else TriadName := 'XXXX';
        end;
    3:  if ((tr[2] = 0) and (tr[3] = 0)) then TriadName := '300 '
                                         else TriadName := 'XXXX';
   end; (* case tr[1] *)
end; (* TriadName *)

function TriadNumber(var tr: TriadType): Byte;
begin
  TriadNumber := 0;
  case tr[1] of
    0:  case tr[2] of 0: TriadNumber := 1;
                      1: TriadNumber := 2;
                      2: case tr[3] of 0: TriadNumber := 6;
                                       1: TriadNumber := 5;
                                       2: TriadNumber := 4;
                                       else TriadNumber := 0;
                         end;
                      3: case tr[3] of 0: TriadNumber := 10;
                                       1: TriadNumber := 9;
                                       else TriadNumber := 0;
                         end;
                      else TriadNumber := 0;
        end;
    1:  case tr[2] of 0: TriadNumber := 3;
                      1: case tr[3] of 1: TriadNumber := 8;
                                       2: TriadNumber := 7;
                                       else TriadNumber := 0;
                         end;
                      2: case tr[3] of 0: TriadNumber := 14;
                                       1: TriadNumber := 13;
                                       2: TriadNumber := 12;
                                       else TriadNumber := 0;
                         end;
                      else TriadNumber := 0;
        end;
    2:  case tr[2] of 0: TriadNumber := 11;
                      1: TriadNumber := 15;
                      else TriadNumber := 0;
        end;
    3:  if ((tr[2] = 0) and (tr[3] = 0)) then TriadNumber := 16 else TriadNumber := 0;
   end; (* case tr[1] *)
end; (* TriadNumber *)

function TriadCode(i: byte): string;
begin
  if ((i <= 0) or (i>= 17)) then TriadCode := '----'
   else TriadCode := TriadName(TriadKind[i]);
end; (* TriadCode *)

procedure DataDescription.Describe(var outf,lf: textfile);
var ms: intmat;
//  sumfile: textfile;
sumfile: TFileStream;
begin
if not FileExists('Siena_desc.out')  then
begin
sumfile:= TFileStream.Create('Siena_desc.out',fmCreate);
sumfile.free;
sumfile:= TFileStream.Create('Siena_desc.out',fmOpenWrite or fmShareDenyWrite);
end
else
begin
  sumfile:= TFileStream.Create('Siena_desc.out',fmOpenReadWrite or fmShareDenyWrite);
  sumfile.Seek(0,soFromEnd);
end;
  //assignfile(sumfile,'Siena_desc.out');
  //if fileexists('Siena_desc.out') then append(sumfile) else rewrite(sumfile);
  DescribeStatistics(outf,lf,sumfile);
//  DescribeWithCovariates(outf); gives potentially a lot of output
  if (nza >=1) then Spatials(outf,sumfile);
  SienaToPajek(lf);
 // writeln(sumfile);
 sumfile.write(pchar(string(eol))^,2);
 // closefile(sumfile);
 sumfile.free;
end;

procedure DataDescription.MatrixSum(var matsum: Intmat);
var i,j,h,s: integer;
  datf: text;
  datfn: shortstring;
begin
  SetLength(matsum,n+1,n+1);
  for i:=1 to n do for j:=1 to n do begin
    if (i = j) then matsum[i,i] := 0 else begin
      s := 0;
      for h:= 1 to observations do if ynm(i,j,h) then inc(s);
      matsum[i,j] := s;
     end;
   end;
  datfn :=  pname+'.sum';
  assign(datf,datfn);
  rewrite(datf);
  for i:=1 to n do begin
    for j:=1 to n do
      if (observations <= 9) then write(datf,' ',matsum[i,j]:1)
       else  write(datf,' ',matsum[i,j]:2);
    writeln(datf);
   end;
  closefile(datf);
end; (* MatrixSum *)

function DataDescription.rowvar(k: integer): real;
(* row degree variance *)
var s1,s2: real;
 d,i,j: integer;
begin
  s1 := 0; s2 := 0;
  for i:=n downto 1 do begin
    d := 0;
    for j:=n downto 1 do if ynm(i,j,k) then inc(d);
    s1 := s1 + d;
    s2 := s2+(d*d);
   end; (* i *)
  rowvar := (s2/n) - sqr(s1/n);
end;

function DataDescription.colvar(k: integer): real;
(* column degree variance *)
var s1,s2: real;
 d,i,j: integer;
begin
  s1 := 0; s2 := 0;
  for i:=n downto 1 do begin
    d := 0;
    for j:=n downto 1 do if ynm(j,i,k) then inc(d);
    s1 := s1 + d;
    s2 := s2+(d*d);
   end; (* i *)
  colvar := (s2/n) - sqr(s1/n);
end;

function DataDescription.HTrans(v,s: real): real;
(* Transforms degree variance and degree sum s to standardized H *)
var h,d,vmax,ev,r1: real;
begin
  h := n-1.0;
  d := s/(n*h);
  r1 := Int(n*d);
  vmax := sqr(h)*(r1 + sqr(n*d-r1) - n*sqr(d))/n;
  ev := (s*(n*h-s)*h)/(sqr(n+0.0)*(n*h-1.0));
  HTrans := divi(Root(v) - Root(ev),Root(vmax) - Root(ev));
end;

function DataDescription.Isolates(k: integer): integer;
(* number of isolates *)
var i,j,s: integer;
  isola: boolean;
begin
  s := 0;
  for i:=n downto 1 do begin
    isola := true;
    j := n;
    repeat
     if (ynm(i,j,k) or ynm(j,i,k)) then isola := false;
     dec(j);
    until ((not isola) or (j <= 0));
    if isola then inc(s);
   end;
  Isolates := s;
end;

function DataDescription.Components(k: integer): integer;
(* number of components consisting of >= 2 vertices *)
var i,j,h,reached,pow,comps: integer;
 m, m2: array of array of Boolean;
 reach: array of Boolean;
 mm: Boolean;
begin
  SetLength(m,n+1,n+1);
  SetLength(m2,n+1,n+1);
  SetLength(reach,n+1);
  for i:=n downto 1 do for j:=n downto 1 do
(* symmetrize & "true" diagonal *)
    m2[i,j] := (ynm(i,j,k) or ynm(j,i,k) or (i=j));
  pow := 1;
  repeat
    for i:=n downto 1 do for j:=n downto 1 do m[i,j] := m2[i,j];
    for i:=n downto 1 do for j:=n downto 1 do begin
      mm := m[i,j];
      if (not mm) then begin
        h := n;
        repeat
         if (m[i,h] and m[j,h]) then mm := true;
         dec(h);
        until (mm or (h <= 0));
       end; (* if not mm *)
      m2[i,j] := mm;
     end; (* i, j *)
    pow := pow*2;
(* Now m2 is the Boolean power ynm^pow *)
   until (pow >= n);
(* Now m2[i,j] indicates whether there is a path between i and j *)
 for i:=1 to n do reach[i] := false;
 i := 1;
 reached := 0;
 comps := 0;
 repeat
   inc(comps);
   for j:=1 to n do if (m2[i,j] and (not reach[j])) then begin
     reach[j] := true;
     inc(reached);
    end;
   if (reached < n) then
    repeat inc(i) until ((not reach[i]) or (i > n));
  until ((reached >= n) or (i > n));
  if (i <= n) then Components := comps - isolates(k) else Components := 9999;
(* There is an error in the last case. It should be impossible.  *)
  finalize(m);
  finalize(m2);
end;


procedure DataDescription.DetermineTriad(var i,j,k: integer; m: integer; var t: TriadType);
var mut,sum: cardinal;
function arc(i,j: integer): boolean;
begin
 arc := ynm(i,j,m);
end;
function Cycle: boolean;
begin
  Cycle := (((Arc(i,j)) and (Arc(j,k)) and (Arc(k,i)))
            or
            ((Arc(j,i)) and (Arc(i,k)) and (Arc(k,j))));
  end; (* Cycle *)
function TwoIn: Boolean;
(* At least one in-degree 2. *)
  begin
  TwoIn := (((Arc(i,j)) and (Arc(k,j))) or
            ((Arc(i,k)) and (Arc(j,k))) or
            ((Arc(j,i)) and (Arc(k,i))));
  end; (* TwoIn *)
function TwoOut: Boolean;
(* At least one out-degree 2. *)
  begin
  TwoOut := (((Arc(j,i)) and (Arc(j,k))) or
             ((Arc(k,i)) and (Arc(k,j))) or
             ((Arc(i,j)) and (Arc(i,k))));
  end; (* TwoOut *)
function ZeroIn: Boolean;
(* At least one in-degree 0. *)
  begin
  ZeroIn :=(((not Arc(i,j)) and (not Arc(k,j))) or
            ((not Arc(i,k)) and (not Arc(j,k))) or
            ((not Arc(j,i)) and (not Arc(k,i))));
  end; (* ZeroIn *)

begin (* DetermineTriad *)
mut := 0; sum := 0;
  if (Arc(i,j)) then begin
   Inc(sum);
   if (Arc(j,i)) then
     begin Inc(mut); Inc(sum); end;
  end else if (Arc(j,i)) then Inc(sum);
 if (Arc(i,k)) then begin
   Inc(sum);
   if (Arc(k,i)) then
     begin Inc(mut); Inc(sum); end;
  end else if (Arc(k,i)) then Inc(sum);
 if (Arc(j,k)) then begin
   Inc(sum);
   if (Arc(k,j)) then
     begin Inc(mut); Inc(sum); end;
  end else if (Arc(k,j)) then Inc(sum);
t[1] := mut;
t[2] := sum - 2*mut;
t[3] := 0;
(* Because of the last assignment, only the triads MANX with X = T, U, or D
   need to have a further t[3] assignment. *)
 if (mut <= 0) then begin (*A1*)
  if (sum >= 3) then
    (* triadtype MAN = 030 *)
     begin if not Cycle then t[3] := 1; end
    else begin (*B*)
     if (sum >= 2) then begin (*C*)
       (* triadtype MAN = 021 *)
        if TwoIn then t[3] := 1 else
          begin if TwoOut then t[3] := 2; end;
        end (*C*) ; end (*B*) ;
    end (*A1*) else begin (*A2*)
      if (mut <= 1) then begin (*B*)
        if (sum >= 4) then begin (*C1*)
          (* triadtype MAN = 120 *)
         if not Cycle then
          begin if ZeroIn then t[3] := 2 else t[3] := 1; end;
         end (*C1*) else begin (*C2*)
           if (sum >= 3 ) then
            (* triadtype MAN = 111 *)
            begin if TwoOut then t[3] := 1 else t[3] := 2; end;
           end (*C2*); end (*B*); end (*A2*);
end; (* DetermineTriad *)

procedure DataDescription.TriadCount(m: integer; var c: triadcounttype;
                        var p: TriadVector; var tottriad: cardinal);
var i,j,k: integer;
    t: TriadType;
    mut,sum,tot: cardinal;
    cc: array[0..3,0..3,0..3] of Integer;
function arc(i,j: integer): boolean;
begin
 arc := ynm(i,j,m);
end;
function Cycle: boolean;
  begin
  Cycle := (((Arc(i,j)) and (Arc(j,k)) and (Arc(k,i)))
            or
            ((Arc(j,i)) and (Arc(i,k)) and (Arc(k,j))));
  end; (* Cycle *)
function TwoIn: Boolean;
(* At least one in-degree 2. *)
  begin
  TwoIn := (((Arc(i,j)) and (Arc(k,j))) or
            ((Arc(i,k)) and (Arc(j,k))) or
            ((Arc(j,i)) and (Arc(k,i))));
  end; (* TwoIn *)
function TwoOut: Boolean;
(* At least one out-degree 2. *)
  begin
  TwoOut := (((Arc(j,i)) and (Arc(j,k))) or
             ((Arc(k,i)) and (Arc(k,j))) or
             ((Arc(i,j)) and (Arc(i,k))));
  end; (* TwoOut *)
function ZeroIn: Boolean;
(* At least one in-degree 0. *)
  begin
  ZeroIn :=(((not Arc(i,j)) and (not Arc(k,j))) or
            ((not Arc(i,k)) and (not Arc(j,k))) or
            ((not Arc(j,i)) and (not Arc(k,i))));
  end; (* ZeroIn *)

begin (* TriadCount *)
(* cc is a version of c with indices that make it more easily accessible *)
 for i:= 0 to 3 do for j:= 0 to 3 do for k:= 0 to 3 do cc[i,j,k] := 0;
 tot := 0;
 for i:= 1 to n do for j:= i+1 to n do for k:= j+1 to n do
   if (not (mis(i,j,m) or mis(j,i,m) or mis(i,k,m) or mis(k,i,m)
            or mis(j,k,m) or mis(k,j,m))) then begin
 (* determine mut = number of mutuals in this triad;
              sum = number of arcs in this triad *)
   inc(tot);
   mut := 0; sum := 0;
   if (Arc(i,j)) then begin
    Inc(sum);
    if (Arc(j,i)) then
      begin Inc(mut); Inc(sum); end;
   end else if (Arc(j,i)) then Inc(sum);
  if (Arc(i,k)) then begin
    Inc(sum);
    if (Arc(k,i)) then
      begin Inc(mut); Inc(sum); end;
   end else if (Arc(k,i)) then Inc(sum);
  if (Arc(j,k)) then begin
    Inc(sum);
    if (Arc(k,j)) then
      begin Inc(mut); Inc(sum); end;
   end else if (Arc(k,j)) then Inc(sum);
  t[1] := mut;
  t[2] := sum - 2*mut  ;
  t[3] := 0;
  (* Because of the last assignment, only the triads MANX with X = T, U, or D
     need to have a further t[3] assignment. *)
   if (mut <= 0) then begin (*A1*)
    if (sum >= 3) then
      (* triadtype MAN = 030 *)
       begin if not Cycle then t[3] := 1; end
      else begin (*B*)
       if (sum >= 2) then begin (*C*)
         (* triadtype MAN = 021 *)
          if TwoIn then t[3] := 1 else
            begin if TwoOut then t[3] := 2; end;
          end (*C*) ; end (*B*) ;
      end (*A1*) else begin (*A2*)
        if (mut <= 1) then begin (*B*)
          if (sum >= 4) then begin (*C1*)
            (* triadtype MAN = 120 *)
           if not Cycle then
            begin if  ZeroIn then t[3] := 2 else t[3] := 1; end;
           end (*C1*) else begin (*C2*)
             if (sum >= 3 ) then
              (* triadtype MAN = 111 *)
              begin if TwoOut then t[3] := 1 else t[3] := 2; end;
             end (*C2*); end (*B*); end (*A2*);
        Inc(cc[t[1],t[2],t[3]]);
   end; (* of i,j,k loops; *)
  tottriad := tot;
  for i:=1 to 16 do begin
    c[i] := cc[TriadKind[i][1],TriadKind[i][2],TriadKind[i][3]];
    p[i] := c[i]/tot;
   end;
end; (* TriadCount *)

procedure DataDescription.Spatials(var outf: textfile;sumf: tFileStream);
(* Spatial autocorrelation statistics *)
var i,j,h,k,nn,zcrange,zz,nmis: integer;
  s1,s2,s3,s4,s5,s6,s5i,s6ij,ym,Gc,Mi,MiMean,MiVar,GcVar: real;
  yav: array of real;
  ynn: array of boolean;
  zn, iso: array of integer;
  z1: array of real;

begin
  SetLength(ynn,n+1);
  SetLength(yav,n+1);
  writeln(outf);
  Heading(2,outf,'Spatial autocorrelation statistics');
  for h:=1 to nza do
    begin
    Heading(3,outf,'Dependent variable '+ZNames[nz+h]);
  sumf.write(pchar(eol)^,length(eol));
    for k:=1 to observations do
      begin
    //  writeln(outf,'Observation moment ',k);
(* Computation of Moran's I and Geary's c *)
(* Actors i with missing zc[i,h,k] are disregarded;
   missings tie variables are set to 0;
   tie variables are weights. These are not necessarily symmetric.
   For the formulae, see e.g.
   B. Ripley, "Spatial Statistics", Wiley 1981, p. 98-99. *)
      s1 := 0.0; s2 := 0.0; ym := 0.0;
      s3 := 0.0; s4:= 0.0; nn := 0;
      s5 := 0.0; s6 := 0.0;
      for i:=1 to n do
        if (not miszc(i,h,k)) then begin
          ym := ym + zc[i,h,k];
          nn := nn+1;
         end;
      if (nn >= 1) then ym := ym/nn
       else ym := 0.0; (* mean zc *)
(* Correspondence to Ripley's symbols:
   s2 = Omega, s5 = S2, s6 = S1.*)

      for i:=1 to n do begin
        s5i := 0.0;
        for j:=1 to n do begin
          s6ij := 0;
          if (ynm(i,j,k) and (not (mis(i,j,k) or miszc(i,h,k) or miszc(j,h,k)))) then begin
            s1 := s1 + sqr(zc[i,h,k]-zc[j,h,k]);
            s2 := s2 + 1;
            s3 := s3 + (zc[i,h,k] - ym)*(zc[j,h,k] - ym);
            s5i := s5i + 1;
            s6ij := s6ij + 1;
           end; (* if  *)
          if (ynm(j,i,k) and (not (mis(j,i,k) or miszc(i,h,k) or miszc(j,h,k)))) then begin
            s5i := s5i + 1;
            s6ij := s6ij + 1;
           end;
          s6 := s6 + sqr(s6ij);
         end; (* for j *)
        s5 := s5 + sqr(s5i);
        if (not miszc(i,h,k))
          then s4 := s4 + sqr(zc[i,h,k] - ym);
       end; (* for i *)
      s6 := 0.5*s6;
      if (nn >= 2) and (s2*s4 > 1e-50) then begin
        Gc := (nn-1)*s1/(2*s2*s4);
        Mi := nn*s3/(s2*s4);
        MiMean := -1/(nn-1);
        MiVar :=
          (sqr(nn)*s6 - nn*s5 + 3*sqr(s2))/((sqr(nn)-1)*sqr(s2)) - sqr(MiMean);
        GcVar := ((2*s6 + s5)*(nn-1) - 4*sqr(s2))/(2*(nn+1)*sqr(s2));
       end else begin
         Gc := 1.0;
         Mi := 0.0;
         MiMean := 0.0;
         MiVar := 0.0;
         GcVar := 0.0;
        end;
      writeln(outf,'Moran',chr(39),'s I = ',Mi:7:4,'; null mean ',MiMean:7:4,
                   ', null S.D. ',root(MiVar):7:4,
                   ', standardized ',divi(Mi-MiMean,root(MiVar)):7:4);
      writeln(outf,'Geary',chr(39),'s c = ',Gc:7:4,'; null mean ',1.0:7:4,
                   ', null S.D. ',root(GcVar):7:4,
                   ', standardized ',-divi(Gc-1.0,root(GcVar)):7:4);
      writeln(outf);
      sumf.write(pchar(format('%7.4f %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f ',
      [Mi,MiMean,root(MiVar),divi(Mi-MiMean,root(MiVar)),Gc,root(GcVar),
                   -divi(Gc-1.0,root(GcVar))]))^,56);
     // write(sumf,Mi:7:4,' ',MiMean:7:4,' ',root(miVar):7:4,' ',
      //           divi(Mi-MiMean,root(MiVar)):7:4,' ', Gc:7:4,' ',
      //           root(GcVar):7:4,' ',-divi(Gc-1.0,root(GcVar)):7:4,' ');

(* Computation of average value yav of neighbors of each actor;
   both i=>j and j=>i make j a neighbor of i *)
      for i:= 1 to n do begin
        nn := 0; s1 := 0.0;
        for j:= 1 to n do begin
         if ((not miszc(j,h,k))
             and ((ynm(i,j,k) and (not mis(i,j,k)))
                  or (ynm(j,i,k) and (not mis(j,i,k))))) then begin
           Inc(nn);
           s1 := s1 + zc[j,h,k];
          end; end; (* for j if *)
        if (nn >= 1) then begin
          ynn[i] := true;
          yav[i] := s1/nn;
         end else begin
          ynn[i] := false;
          yav[i] := 0.0;
         end;
       end; (* for i *)
(* average yav[i] for all i with the same value zc[i,h,k] =  zz *)
      zcrange := round(zcmax[h] - zcmin[h]) + 1;
      Setlength(z1,zcrange+1);
      Setlength(zn,zcrange+1);
      Setlength(iso,zcrange+1);
      for i:=0 to zcrange do begin
        z1[i] := 0.0;
        zn[i] := 0;
        iso[i] := 0;
       end;
      nmis := 0;
      for i:=1 to n do
        begin
        zz := round(zc[i,h,k]);
(* first rescale to values 1..zcrange *)
        if ((zz >= zcmin[h]) and (zz <= zcmax[h]) and (not miszc(i,h,k))) then
          begin
          zz := round(zz-zcmin[h])+1;
          if ynn[i] then
            begin
            zn[zz] := zn[zz] + 1;
            z1[zz] := z1[zz] + yav[i];
            end (* if ynn[i] *)
           else iso[zz] := iso[zz] + 1;
          end
         else inc(nmis);
        end; (* for i *)
      for i:=1 to zcrange do if (zn[i] >= 1) then z1[i] := z1[i]/zn[i];
      writeln(outf,'Average values of neighbors of nodes with value z :');
      write(outf,'z       ');
      for i:=1 to zcrange do write(outf, round(zcmin[h]+i-1):5,' ');
      writeln(outf);
      write(outf,'average ');
      for i:=1 to zcrange do
        if (zn[i] >= 1) then write(outf, z1[i]:5:1,' ')
                        else write(outf,'  --- ');
      writeln(outf);
      write(outf,'number  ');
      for i:=1 to zcrange do write(outf, zn[i]:5,' ');
      writeln(outf);
      iso[0] := 0;
      for i:=1 to zcrange do inc(iso[0],iso[i]);
      if (iso[0] >= 1) then
        begin
        write(outf,'isolates');
        for i:=1 to zcrange do write(outf, iso[i]:5,' ');
        end;
      writeln(outf);
      if (nmis >= 1) then writeln(outf,'Missings ',nmis:4);
      writeln(outf);
      end; (* for k *)
      writeln(outf);
    end; (* for h *)
end; (* Spatials *)


procedure DataDescription.DescribeStatistics(var outf,lf: textfile;sumf: tFileStream);
(* This procedure gives some basic data description in the output file.
   Initially it was copied from InitialStatistics in S_Start,
   deleting all parts that have to do with parameter estimation.
   But a lot was added.
   The triad count was taken from GRSTAT in the ZO collection.
   *)
var i,j,k,h: integer;
  ntot,ndtot,n1,nd1,ne1,n11,null,asym,mut,ndy,comp: array of cardinal;
  nmisd,nlu,ttot: cardinal;
  n01,n10,n00: array of integer;
  xk: array of real;
  tc: TriadCountType;
  tv: TriadVector;
  mystring:string;
begin (* DescribeStatistics *)
 Heading(1,outf,'Data description.');
 setlength(n1,observations+1);
 setlength(n11,observations+1);
 setlength(n01,observations+1);
 setlength(n10,observations+1);
 setlength(n00,observations+1);
 setlength(ntot,observations+1);
 setlength(ndtot,observations+1);
 setlength(nd1,observations+1);
 setlength(ne1,observations+1);
 setlength(null,observations+1);
 setlength(asym,observations+1);
 setlength(mut,observations+1);
 setlength(ndy,observations+1);
 setlength(xk,observations+1);
 setlength(comp,observations+1);
 missing := false;
 if length(pname) <= 7 then
//  write(sumf,pname:7,' ',n:7,' ',nza:7,' ',observations:7,' ')
begin
mystring := format('%7s %7d %7d %7d ',[pname,n,nza,observations]);
sumf.write(pchar(mystring)^,length(mystring));
end
  else
  // write(sumf,pname,' ',n:7,' ',nza:7,' ',observations:7,' ');
 sumf.write( pchar(format('%s %7:d %7:d %7:d ',[pname,n,nza,observations]))^,25+length(pname));
(*  note that missing cells have the value 0=false *)
   for k := 1 to observations do begin
     n1[k] := 0;
     n11[k] := 0;
     nd1[k] := 0;
     ne1[k] := 0;
     ntot[k] := 0;
     ndtot[k] := 0;
     null[k] := 0;
     asym[k] := 0;
     mut[k] := 0;
     ndy[k] := 0;
(* Statistics for periods separately : *)
    for i:=1 to n do for j:=1 to n do begin
       if ynm(i,j,k) then inc(n1[k]);
       if not(mis(i,j,k) or (i=j)) then inc(ntot[k]);
       if ((i < j) and (not mis(i,j,k)) and (not mis(j,i,k))) then begin
         h := 0;
         if ynm(i,j,k) then inc(h);
         if ynm(j,i,k) then inc(h);
         case h of
          0: inc(null[k]);
          1: inc(asym[k]);
          2: inc(mut[k]);
          end; (* case *)
          inc(ndy[k]);
        end; (* if *)
      end; (* for i,j *)
     if (ntot[k] < n*(n-1)) then missing := true;
(* Change statistics : *)
     for i:=1 to n do for j:=1 to n do begin
       if ((k <observations) and not(mis(i,j,k) or mis(i,j,k+1) or (i=j))) then
           begin
           inc(ndtot[k]);
           if ynm(i,j,k+1) then inc(nd1[k]);
           if (ynm(i,j,k) and ynm(i,j,k+1)) then inc(n11[k]);
           if ynm(i,j,k) then inc(ne1[k]);
           end;
      end; (* for i,j *)
    end;

  if not missing then writeln(outf,'No missing data in this data set.');

(* first come the descriptive statistics for each observation separately *)
  if (observations = 1) then begin
    if (ntot[1] > 0) then begin
      if missing then begin
        writeln(outf,
          'For the following statistics, missing values in the digraphs are ');
        writeln(outf,'considered as absent arcs.');
      end;
     Heading(2,outf,'Components.');
     writeln(outf,'Number of isolated vertices: ',isolates(1):3,' ');
     comp[1] := Components(1);
     writeln(outf,'Number of weak components of 2 or more vertices :',
                  comp[1]:3,' ');
    // writeln(sumf);
    sumf.write(pchar(eol)^,length(eol));
   //  write(sumf,n1[1]:7,' ',(1.0 - ntot[1]/(n*(n-1))):7:3,' ');
   sumf.write(pchar(format('%7d %7.3f ',[n1[1],(1.0 - ntot[1]/(n*(n-1)))]))^,16);
     xk[1] := rowvar(1);
     Heading(2,outf,'Degree variances / standard deviations.');
     writeln(outf,'Row Degree S.D.  :        ',root(xk[1]):6:1);
     writeln(outf,'Row normalized variance (H):',HTrans(xk[1],n1[1]):6:3);
     xk[1] := colvar(1);
    writeln(outf,'Row Degree S.D.  :        ',root(xk[1]):6:1);
     writeln(outf,'Row normalized variance (H):',HTrans(xk[1],n1[1]):6:3);
     writeln(outf);
    end; (* if *)

  if missing then
    writeln(outf,
 'For the following statistics, missing values in the digraphs are not counted.');
   writeln(outf,'Network density  = ',n1[1],'/',ntot[1],' = ',
                   d(n1[1],ntot[1]):8:4,' .');
   writeln(outf,'Average degree   = ', d((n-1)*n1[1],ntot[1]):8:4,' .');
  // writeln(sumf, d((n-1)*n1[1],ntot[1]):7:3,' .');
  sumf.write(pchar(format('%8.4f .',[d((n-1)*n1[1],ntot[1])])+eol)^,12);
  // writeln(sumf,comp[1]:7);
  sumf.write(pchar(format('%7d',[comp[1]])+eol)^,9);
   Heading(2,outf,'Dyad count.');
   if missing then writeln(outf,ndy[1],' complete dyads:')
     else writeln(outf,ndy[1],' dyads:');
   writeln(outf,null[1],' null, ',asym[1],' asymmetric, ',mut[1],' mutuals.');
   Heading(2,outf,'Triplets and triads.');
   if (ntot[1] > 0) then begin
     write(outf,'Triplets with (i -> j), (j -> h)');
     if missing then writeln(outf,' and (i,h) non-missing:')
       else writeln(outf,':');
     ndy[1] := 0; (* will be total number of triplets  *)
     mut[1] := 0; (* will be number of transitive triplets  *)
     for i:=n downto 1 do for j:=n downto 1 do for h:=n downto 1 do begin
       if (ynm(i,j,1) and ynm(j,h,1) and (not mis(i,h,1))) then begin
         inc(ndy[1]);
         if ynm(i,h,1) then inc(mut[1]);
        end;
      end; (* for i, j, h *)
     xk[1] := divi(mut[1],ndy[1]);
     writeln(outf,'Total           ',ndy[1]:6,' ');
     writeln(outf,'Transitive      ',mut[1]:6,' ');
     if (ndy[1] > 0) then writeln(outf,'Trans. index    ',xk[1]:6:3,' ');
  //   write(sumf,xk[1]:7:3,' ');
     sumf.write(pchar(format('%7.3f ',[xk[1]]))^,8);
     writeln(outf);
     writeln(outf);
    end; (* if ntot[1] > 0 *)
 end else (* observations >= 2 *) begin
  if missing then begin
    writeln(outf,'For the numbers of isolates and of components,');
    writeln(outf,'and the degree s.d.s and variances, ');
    writeln(outf,'missing values in the digraphs are considered as absent arcs.');
    writeln(outf,'For the counts of arcs, dyads, triplets, and triads,');
    writeln(outf,'the missings are not counted.');
    end;
   Heading(2,outf,'Descriptives per observed network.');
   write(outf,'Observation time ');
   for k:=1 to observations do write(outf,k:4,'   ');
   writeln(outf); writeln(outf);
   write(outf,'Density         ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'  --   ')
       else write(outf,n1[k]/ntot[k]:6:3,' ');
   writeln(outf);
   write(outf,'Average degree  ');
   for k:=1 to observations do
     if (ntot[k] = 0) then begin
       write(outf,'  --   ');
       end
      else begin
       write(outf,(n-1)*n1[k]/ntot[k]:6:3,' ');
     end;
   writeln(outf);
   if missing then begin
     write(outf,'Missing fraction');
     for k:=1 to observations do
       if (ntot[k] = 0) then write(outf,' 1.0  ')
         else write(outf,(1.0 - ntot[k]/(n*(n-1))):6:3,' ');
     writeln(outf);
    end; (* if missing *)
   Heading(3,outf,'Dyad count');
   if missing then write(outf,'Complete dyads  ')
     else write(outf,'Dyads           ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'   --  ') else write(outf,ndy[k]:6,' ');
   writeln(outf);
   write(outf,'Mutual          ');
   for k:=1 to observations do
     if (ndy[k] = 0) then write(outf,'   --  ') else write(outf,mut[k]:6,' ');
   writeln(outf);
   write(outf,'Asymmetric      ');
   for k:=1 to observations do
     if (ndy[k] = 0) then write(outf,'   --  ') else write(outf,asym[k]:6,' ');
   writeln(outf);
   write(outf,'Null            ');
   for k:=1 to observations do
     if (ndy[k] = 0) then write(outf,'   --  ') else write(outf,null[k]:6,' ');
   writeln(outf);
   writeln(outf,'Fraction reciprocated [i.e. 2M/(2M+A) ]            ');
   write(outf,'                ');
   for k:=1 to observations do
     if ((mut[k] + asym[k]) = 0) then write(outf,'   --  ')
               else write(outf,2*mut[k]/(2*mut[k]+asym[k]):6:3,' ');
   writeln(outf);
   Heading(3,outf,'Components');
   writeln(outf,'Numbers of isolated vertices, and of weak components of 2 or more vertices :');
   write(outf,'Number isolates ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'   --  ')
       else write(outf,isolates(k):6,' ');
   writeln(outf);
   write(outf,'Number componts.');
   for k:=1 to observations do begin
     if (ntot[k] = 0) then begin
       write(outf,'   --  ');
       comp[k] := 0;
      end else begin
       comp[k] := Components(k);
       write(outf,comp[k]:6,' ');
       end;
      end;
   writeln(outf);
   Heading(3,outf,'Degree variances / standard deviations');
   writeln(outf,'Degree standard deviations and normalized variances (H):');
   write(outf,'Row D.S.D.    ');
   for k:=1 to observations do begin
     xk[k] := rowvar(k);
     if (ntot[k] = 0) then write(outf,'   --  ') else write(outf,root(xk[k]):6:1,' ');
    end;
   writeln(outf);
   write(outf,'Row H           ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'   --  ')
       else write(outf,HTrans(xk[k],n1[k]):6:3,' ');
   writeln(outf);
   write(outf,'Column D.S.D. ');
   for k:=1 to observations do begin
     xk[k] := colvar(k);
     if (ntot[k] = 0) then write(outf,'   --  ') else write(outf,root(xk[k]):6:1,' ');
    end;
   writeln(outf);
   write(outf,'Column H        ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'   --  ')
       else write(outf,HTrans(xk[k],n1[k]):6:3,' ');
   writeln(outf);
   Heading(3,outf,'Triplets and triads');
   write(outf,'Triplets with (i -> j), (j -> h)');
   if missing then writeln(outf,' and (i,h) non-missing:')
     else writeln(outf,':');
   for k:=1 to observations do begin
     ndy[k] := 0; (* will be total number of triplets  *)
     mut[k] := 0; (* will be number of transitive triplets  *)
     for i:=n downto 1 do for j:=n downto 1 do for h:=n downto 1 do begin
       if (ynm(i,j,k) and ynm(j,h,k) and (not mis(i,h,k))) then begin
         inc(ndy[k]);
         if ynm(i,h,k) then inc(mut[k]);
        end;
      end; (* for i, j, h *)
     xk[k] := divi(mut[k],ndy[k]);
    end; (* for k *)
   write(outf,'Total           ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'   --  ') else write(outf,ndy[k]:6,' ');
   writeln(outf);
   write(outf,'Transitive      ');
   for k:=1 to observations do
     if (ntot[k] = 0) then write(outf,'   --  ') else write(outf,mut[k]:6,' ');
   writeln(outf);
   write(outf,'Trans. index    ');
   for k:=1 to observations do begin
     if (ndy[k] = 0) then write(outf,'   --  ') else write(outf,xk[k]:6:3,' ');
   //  writeln(sumf);
   sumf.write(pchar(eol)^,length(eol));
   //  write(sumf,n1[k]:7,' ',(1.0 - ntot[k]/(n*(n-1))):7:3,' ');
   sumf.write(pchar(format('%7d %7.3f ',[n1[k],(1.0 - ntot[k]/(n*(n-1)))]))^,16);
   if (ntot[k] <= 0) then
   //write(sumf,0.0:7:3,' ')
   sumf.write(pchar(format('%7.3f ',[0.0]))^,8)
   else
   //    write(sumf,(n-1)*n1[k]/ntot[k]:7:3,' ');
   sumf.write(pchar(format('%7.3f ',[(n-1)*n1[k]/ntot[k]]))^,8);
      //  write(sumf,comp[k]:7,' ');
   sumf.write(pchar(format('%7d ',[comp[k]]))^,8);
   //  write(sumf,xk[k]:7:3,' ');
   sumf.write(pchar(format('%7.3f ',[xk[k]]))^,8);
    end;
   for k:=1 to observations-1 do begin
 //    writeln(sumf);
       sumf.write(pchar(eol)^,length(eol));
  //   write(sumf, ObservedNetworkDistance[1,k]:7,' ');
       sumf.write(pchar(format('%7d ',[ObservedNetworkDistance[1,k]]))^,8);
    for h:=1 to nza do
    //write(sumf, ObservedActionDistance[h,k]:7,' ');
    sumf.write(pchar(format('%7d ',[ObservedActionDistance[h,k]]))^,8);
    end;
   writeln(outf);
   writeln(outf);
  end; (* if observations >= 2 *)


  writeln(outf,'Triad census');
  writeln(outf);
  for k:=1 to observations do begin
    if (observations>= 2) then writeln(outf,'Observation ', k);
    TriadCount(k,tc,tv,ttot);
    nlu := 3*trunc(ln(n))+1;
    write(outf,'Triad');
    for i:=1 to nlu do write(outf,' ');
    writeln(outf,'Number      Proportion ');
    for i:=1 to 16 do begin
      writeln(outf,TriadCode(i),'      ',tc[i]:nlu,'        ',tv[i]:8:6);
     end;
    if missing then write(outf,'Total non-missing  ')
      else write(outf,'Total     ');
    writeln(outf,ttot:nlu);
    writeln(outf);
   end; (* for k *)

(* Now come the change statistics *)
  if (observations >= 2) then begin
   Heading(2,outf,'Descriptives for changes between networks.');
   if missing then begin
     writeln(outf,
     'For the following statistics, missing values in the digraphs');
     writeln(outf,
     ' for the observations at the start or end of each period are not counted.');
    end ;
   Heading(3,outf,'Arcs');
   writeln(outf,'Numbers of changes between subsequent observations');
   write(outf,'obs. times   0 -> 0    0 -> 1    1 -> 0    1 -> 1   Distance');
   if missing then writeln(outf,'     Missing') else writeln(outf);
   for k:=1 to observations-1 do begin
     n01[k] := nd1[k]-n11[k];
     n10[k] := ne1[k]-n11[k];
     n00[k] := ndtot[k] - n10[k] - n01[k] - n11[k];
     if missing then begin
       nmisd := n*(n-1)-ndtot[k];
       writeln(outf,k:2,' to ',(k+1):2,n00[k]:10,n01[k]:10,n10[k]:10,n11[k]:10,
            ObservedNetworkDistance[1,k]:10, nmisd:10,
            ' (',(100.0*nmisd/(n*(n-1))):2:0,'%)');
      end (* if missing *) else
       writeln(outf,k:2,' to ',(k+1):2,n00[k]:10,n01[k]:10,n10[k]:10,n11[k]:10,
            ObservedNetworkDistance[1,k]:10);
    end; (* for k *)
   Heading(3,outf,'Dyads');
   writeln(outf,'Number of dyad changes between subsequent observations');
   writeln(outf,
'obs. times   M -> M    M -> A    M -> N      M -> any');
   for k:=1 to observations -1 do begin
     null[k] := 0;
     asym[k] := 0;
     mut[k] := 0;
     ndy[k] := 0;
     for i:=1 to n-1 do for j:=i+1 to n do
       if (not (mis(i,j,k) or mis(i,j,k+1) or mis(j,i,k) or mis(j,i,k+1))) then
         if (ynm(i,j,k) and ynm(j,i,k)) then begin
           inc(ndy[k]);
           h := 0;
           if ynm(i,j,k+1) then inc(h);
           if ynm(j,i,k+1) then inc(h);
           case h of
             0: inc(null[k]);
             1: inc(asym[k]);
             2: inc(mut[k]);
            end; (* case *)
      end; (* for i, j if not mis *)
     writeln(outf,
      k:2,' to ',(k+1):2,mut[k]:10,asym[k]:10,null[k]:10,'  ',ndy[k]:10);
    end; (* for k *)
   writeln(outf);
   writeln(outf,
'obs. times   A -> M    A -> A (=)   A -> A (/=)   A -> N      A -> any');
   for k:=1 to observations -1 do begin
     null[k] := 0;
     asym[k] := 0;
     ne1[k] := 0;
     mut[k] := 0;
     ndy[k] := 0;
     for i:=1 to n-1 do for j:=i+1 to n do
       if (not (mis(i,j,k) or mis(i,j,k+1) or mis(j,i,k) or mis(j,i,k+1))) then
         if ((not ynm(i,j,k)) and ynm(j,i,k)) or (ynm(i,j,k) and (not ynm(j,i,k))) then begin
           inc(ndy[k]);
           h := 0;
           if ynm(i,j,k+1) then inc(h);
           if ynm(j,i,k+1) then inc(h);
           case h of
             0: inc(null[k]);
             1: if (ynm(i,j,k) = ynm(i,j,k+1)) then inc(asym[k])
                 else inc(ne1[k]);
             2: inc(mut[k]);
            end; (* case *)
      end; (* for i, j if not mis *)
     writeln(outf,
      k:2,' to ',(k+1):2,mut[k]:10,asym[k]:10,ne1[k]:13,null[k]:14,'  ',ndy[k]:10);
    end; (* for k *)
   writeln(outf);
   writeln(outf,
'obs. times   N -> M    N -> A    N -> N      N -> any');
   for k:=1 to observations -1 do begin
     null[k] := 0;
     asym[k] := 0;
     mut[k] := 0;
     ndy[k] := 0;
     for i:=1 to n-1 do for j:=i+1 to n do
       if (not (mis(i,j,k) or mis(i,j,k+1) or mis(j,i,k) or mis(j,i,k+1))) then
         if (not (ynm(i,j,k) or ynm(j,i,k))) then begin
           inc(ndy[k]);
           h := 0;
           if ynm(i,j,k+1) then inc(h);
           if ynm(j,i,k+1) then inc(h);
           case h of
             0: inc(null[k]);
             1: inc(asym[k]);
             2: inc(mut[k]);
            end; (* case *)
      end; (* for i, j if not mis *)
     writeln(outf,
      k:2,' to ',(k+1):2,mut[k]:10,asym[k]:10,null[k]:10,'  ',ndy[k]:10);
    end; (* for k *)

   Heading(3,outf,'Triplets');
   writeln(outf,'Changes from intransitive triplets:');
   writeln(outf,'Changes for triplets starting with (i -> j), (j -> h), not (i -> h) :');
   writeln(outf,
'obs. times   i -> t    i -> i    i -> v      i -> any');
   for k:=1 to observations-1 do begin
     ndy[k] := 0; (* will be total number of triplets  *)
     mut[k] := 0; (* will be number of transitive triplets  *)
     asym[k] := 0; (* will be number of intransitive triplets *)
     for i:=n downto 1 do for j:=n downto 1 do for h:=n downto 1 do begin
       if (ynm(i,j,k) and ynm(j,h,k) and (not mis(i,h,k)) and (not ynm(i,h,k)))
        and (not (mis(i,j,k+1) or mis(j,h,k+1) or mis(i,h,k+1))) then begin
         inc(ndy[k]);
         if (ynm(i,j,k+1) and ynm(j,h,k+1) and ynm(i,h,k+1)) then inc(mut[k]);
         if (ynm(i,j,k+1) and ynm(j,h,k+1) and (not ynm(i,h,k+1))) then inc(asym[k]);
        end;
      end; (* for i, j, h *)
     writeln(outf,
      k:2,' to ',(k+1):2,mut[k]:10,asym[k]:10,ndy[k]-mut[k]-asym[k]:10,'  ',ndy[k]:10);
    end; (* for k *)
   writeln(outf);
   writeln(outf,'Changes from transitive triplets:');
   writeln(outf,'Changes for triplets starting with (i -> j), (j -> h), (i -> h) :');
   writeln(outf,
'obs. times   t -> t    t -> i    t -> v      t -> any');
   for k:=1 to observations-1 do begin
     ndy[k] := 0; (* will be total number of triplets  *)
     mut[k] := 0; (* will be number of transitive triplets  *)
     asym[k] := 0; (* will be number of intransitive triplets *)
     for i:=n downto 1 do for j:=n downto 1 do for h:=n downto 1 do begin
       if (ynm(i,j,k) and ynm(j,h,k) and ynm(i,h,k))
        and (not (mis(i,j,k+1) or mis(j,h,k+1) or mis(i,h,k+1))) then begin
         inc(ndy[k]);
         if (ynm(i,j,k+1) and ynm(j,h,k+1) and ynm(i,h,k+1)) then inc(mut[k]);
         if (ynm(i,j,k+1) and ynm(j,h,k+1) and (not ynm(i,h,k+1))) then inc(asym[k]);
        end;
      end; (* for i, j, h *)
     writeln(outf,
      k:2,' to ',(k+1):2,mut[k]:10,asym[k]:10,ndy[k]-mut[k]-asym[k]:10,'  ',ndy[k]:10);
    end; (* for k *)
   writeln(outf);
   writeln(outf);

  end; (* if observations >= 2 *)

  writeln(outf);

  if (not anynetworkmissings) then
(* If there are missings: note that the carryforward option imputes missings
    which can lead to inconsistent results with the definition of Dist2. *)  
    begin
    Heading(3,outf,'Indirect connections');
    if symmetric then
      writeln(outf,'Number of pairs having a geodesic distance of 2:')
     else
      writeln(outf,'Number of directed pairs having a geodesic directed distance of 2:');
    if (observations >= 2) then writeln(outf,'obs.    # pairs');
    for k:=1 to observations do
      begin
      if (observations >= 2) then write(outf,k:3,'       ');
      ttot := 0;
      for i:=1 to n do for j:=1 to n do if (i <> j) and Dist2(i,j,k) then inc(ttot);
      if symmetric then writeln(outf,ttot div 2:3) else writeln(outf,ttot:3);;
      end;
    writeln(outf);
    end;

end; (* DescribeStatistics *)

procedure DataDescription.DescribeWithCovariates(var outf: textfile);
(* Gives plots of outdegrees and of indegrees
   to other actors with covariate values below covariate mean, and above mean *)
var h: integer;
begin
  for h:=1 to nz do
    begin
    writeln(outf);
    writeln(outf,'Individual non-changing covariate ',h);
    DescribeNetworksCovariate(n,nz,observations,outf);
    end;
  for h:=1 to nzc do
    begin
    writeln(outf);
    writeln(outf,'Individual changing covariate ',h);
    DescribeNetworksCCovariate(n,nz,nzc,observations,outf);
    end;
end;

begin (* unit *)
  InitTriad(TriadKind);
end.
