(* This is file S_ergm.pas, which is included in S_base.pas *)

(* This file contains some material connected to
   S_base that is used only for the observations = 1 case. *)



procedure ModelSpecification.initialiseERGM(const nn, maxn: integer; const d: real; const ccond: boolean);
var i: integer;
    od: array of integer;
    cst, dinmax, dtot, pinlast: integer;
begin
  setlength(lograt,nn+1);
  lograt[0] := 0.0; (* should not be used *)

  if (gibbs in [4..8]) then
    begin
    if (d < 0.5) then probA := (1.0-2*d)/(2.0-2*d) else probA := 0.0;
    if ccond then
      for i:=1 to nn do lograt[i] := log(1.0 + maxn/i)
     else
      for i:=1 to nn do lograt[i] := log(1.0 + (proba*maxn)/((1.0-proba)*i));
    end;

  if (gibbs in [9..12]) then
    begin
    cst := 3; (* trial value *)
    SetLength(ind,nn+1);
    SetLength(outd,nn+1);
    ind[0] := 0;
    outd[0] := 0;
    dtot := 0;
    dinmax := 0;
    for i := nn downto 1 do
      begin
      ind[i] := Indm(i,1);
      outd[i] := Outdm(i,1);
      inc(dtot,ind[i]);
      if (ind[i] > dinmax) then dinmax := ind[i];
      end;

    if (gibbs in [10,12]) then
      begin
      cst := round(0.1*dinmax);      (* 0.1 is arbitrary *)
      if (cst < 2) then cst := 2;   (* so is 2 *)
      SetLength(pin,nn+1);
      SetLength(pout,nn+1);
      pin[0] := 0;
      pout[0] := 0;
      pinlast := dtot + nn*cst;   (* = pin[nn] defined below *)
      for i:=1 to nn do
        begin
        pin[i] := pin[i-1] + cst + ind[i];
        pout[i] := pout[i-1] + outd[i];
        lograt[i] := log((ind[i]+cst)*(pinlast-ind[i]-cst));
        end;
      if (pinlast <> pin[nn]) then writeln;
      end;
    end; (* gibbs in [9,10,11,12] *)
  ErgmStatistics(CurrentStats,symmetric,od,true,NewNetworkEffects);
end; (* initialiseERGM *)


function ModelSpecification.RandomVertexInComponent
                            (const i: integer; const SameAllowed: boolean): integer;
(* Selects a random vertex in the same component as i;
  if not SameAllowed, it must be but different from i. *)
var j,jj,h,k: integer;
begin
  if (not MoreComponents) then begin
    repeat j := randomd(version,n) + 1
     until (StructActive[j] and (SameAllowed or (i <> j)));
   end else begin
    h := ComponentIndicator[i];
    k := ComponentNumber[h,0];
(* h is the component of i;
   k is number of elements in this component *)
    if (k <= 1) then j := 0
 (* then this is an isolated vertex, should be forbidden *)
     else begin
      repeat
        jj := randomd(version,k) + 1;
        j := ComponentNumber[h,jj];
       until (SameAllowed or (i <> j));
     end; (* else *)
   end; (* MoreComponents *)
  RandomVertexInComponent := j;
end; (* ModelSpecification.RandomVertexInComponent *)



function ModelSpecification.CheckRunningStatistics(var outf,lf: TFileStream): Boolean;
(* In the ERGM case, statistics are calculated by updating,
   based on change statistics.
   The procedure CheckRunningStatistics checks whether these updated values
   are identical to what is produced by ERGMStatistics. *)
var i,cd, sd: integer;
 s: pvec;
 deg: array of integer;
 ok: boolean;
 dif: real;
 x : string;
begin
  ok := true;
  setlength(deg,n+1);
  if cconditional then begin
    sd := 0;
    for i:=n downto 1 do inc(sd,outd_yn(i));
    cd := 0;
    for i:=n downto 1 do inc(cd,outdm(i,1));
    if (cd <> sd) then
      begin
      ok := false;
      x := format('%s%d%s%d%s',['total number of ties: observed ',cd,', simulated ',sd,'.']);
      outf.Write(pchar(x)^,length(x));
      x := ' **** error ****'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    end;

  ERGMStatistics(s,symmetric,deg,true,NewNetworkEffects);
  for i:=1 to ActiveNetworkEffects_f do
    begin
    if (abs(s[i] - currentstats[i]) > 0.1) then  ok := false;
(* The threshold of 0.1 is high, but if it is wrong, then it is far off,
   and small errors should not be counted. *)
   end;
  if (gibbs = 5) then
    begin
    cd := 0;
    sd := 0;
    for i:=1 to n do
        begin
        inc(cd,outd_yn(i));
        inc(sd,abs(cd-cumuldegrees[i]));
        end;
    if (sd > 0) then
      begin
      cd := 0;
      for i:=1 to n do
        begin
        inc(cd,outd_yn(i));
        x := format('%3d%s%12d%s%12d',[i,'. ',cd,' ',CumulDegrees[i]]);
        outf.Write(pchar(x)^,length(x));
        if (abs(cd - cumuldegrees[i] ) > 0)  then
          begin
          ok := false;
          x := '**** error ****'+eol;
          outf.Write(pchar(x)^,length(x));
          end
         else outf.Write(pchar(eol)^,length(eol));
        end;
      end else if ok then begin
        x := 'Check running statistics OK.'+eol;
        lf.Write(pchar(x)^,length(x));
      end;
    end else if ok then begin
      x := 'Check running statistics OK.'+eol;
      lf.Write(pchar(x)^,length(x));
    end;
  if not ok then
    begin
    x := 'There is an error in the calculations of running statistics.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'An incorrect change statistic has been used.'+eol;
    for i:=1 to ActiveNetworkEffects_f do
      begin
      x := format('%3d,%s%12.4f%s%12.4f',[i,'. ', s[i],' ',currentstats[i]]);
      outf.Write(pchar(x)^,length(x));
      if (abs(s[i] - currentstats[i]) > 0.1)
        then begin
        x := '**** error ****'+eol;
        outf.Write(pchar(x)^,length(x));
        end
        else
        outf.Write(pchar(eol)^,length(eol));
       end;
    end
   else
    begin
    dif := 0.0;
    for i:=1 to ActiveNetworkEffects_f do
      begin
      if (abs(s[i] - currentstats[i]) > dif)
        then dif :=  abs(s[i] - currentstats[i]) ;
      currentstats[i] := s[i];  (* against cumulated rounding errors *)
      end;
    x := format('%s%20.12f',['Largest absolute difference is ',dif])+eol;
    lf.Write(pchar(x)^,length(x));
    end;
  CheckRunningStatistics := ok;
(*    if not ok then
      begin
      closefile(outf);
      halt;
      end; *)
end; (* CheckRunningStatistics *)

procedure ModelSpecification.ErgmRunStep(var error: Boolean;
                             var erstr: shortstring; var halterr: Boolean;
                             var outf,lf: TFileStream);
(* one change *)
(* The system random number state must have been set
   before entering this procedure. *)
(* the step consists of a stochastic change of the adjacency matrix;
   the number of changes depends on gibbs and Cconditional. *)
var i,j,h,k,ii,hh,hvar,s,sum,a1,a2,b1,b2,c1,c2: integer;
    tie1, tie2, tie3, cc, change, yih, plus, minpossible, pluspossible: Boolean;
    ynij, ynhk: boolean;
    sp,sp1,sp2,taun,lxi: real;
    hstar: byte;
    stats0, stats1, DeltaStats: pvec;
    eu,eu1,eu2,lnp, lami, lamsum, ran, pro, pro1, pro2,
       p2,p3,p4,p5,p6,p7,p8: real;
    iii : integer;
    ords: intvec;
    tij,thk,tr: shortint;
od: array of integer; // abcd
procedure ChangeStat_VertPars(var ee: real; const tt, ttr: shortint; const i1,j1: integer);
(* Carries out cumulating changes in the change statistics DeltaStats,
   and in the part of the log odds ee depending on vertex parameters,
   corresponding to a toggle of y(i,j).
   tt is the current value of y(i,j), ttr of y(j,i);
   these are included as separate parameters for efficiency purposes. *)
begin
  if VertexPar then
    begin
      if symmetric then
        begin
        if (tt <= 0) then
          ee :=  ee + vertex_a[i1] + vertex_a[j1]
         else  ee :=  ee - vertex_a[i1] - vertex_a[j1];
        end
       else
        begin
        if (tt <= 0) then ee :=  ee + vertex_a[i1] + vertex_b[j1]
         else  ee :=  ee - vertex_a[i1] - vertex_b[j1];
        end;
    end;
  ERGMChangeStatistics(i1,j1,DeltaStats,tt,ttr,NewNetworkEffects);
end; (* ChangeStat_VertPars *)
procedure OtherPair(var ha,ka: integer; const ia,ja: integer; const one: boolean);
(* Returns andom pair (ha,ka) in the same component as (ia,ja),
   but different from (ia,ja);
   if one, this is a random pair for which y(ha,ka) = 1.  *)
begin
  repeat
  if MoreComponents then
    begin
    repeat
      ha := RandomVertexInComponent(ia,true);
      ka := RandomVertexInComponent(ha,false) ;
      until  (not ((ha = ia) and (ka = ja))) and (not(Struct(ha,ka,1)));
    end
   else
    if one then RandomCurrentTie(version,ha,ka) else ha := 0;
    if (ha <= 0) then
      repeat
      ha := randomd(version,n) + 1;
      ka := randomd(version,n) + 1;
      until (ha <> ka);
  until (not ((ha = ia) and (ka = ja)));
end; (* OtherPair *)
procedure OtherPairA(const ia,ja: integer; const t0: shortint;
                     var ha,ka: integer; var ta: shortint);
(* Used if Gibbs = 4 or 6 and Cconditional.
   (ha, ka) = Random pair in the same component as ia;
   t0 = yn(ia,ja); ta = yn(ha,ka).
   if t0 = 0, then the requirement is made that y(ha,ka) = 1.
   selection is not with equal probabilities: version A.
   ha=ka eventually implies exit from ERGMrunstep without doing anything.*)
var kronecker: integer;
begin
  ha := RandomVertexInComponent(ia,true);
  if (t0 >= 1) then
    begin
    if existstruct[1] then
      begin
      if ((ha = ia) or (symmetric and (ha = ja))) then kronecker := 1 else kronecker := 0;
      if (maxnfree*rando(version) > (n - 1 - kronecker - StructRowElements(ha,1)))
        then ka := ha
       else
        repeat
        ka := RandomVertexInComponent(ha,false);
        until (not(Struct(ha,ka,1)) and ((ia <> ha) or (ja <> ka)));
      end
     else
      begin (* not existstruct[1] *)
      if (((ha = ia) or ((symmetric or asymmetric) and (ha = ja)))
          and ((n-1)*rando(version) < 1.0)) then ka := ha
       else
        repeat
         ka := randomd(version,n) + 1;
         until ((ka <> ha) and ((ha <> ia) or (ka <> ja ))
          and ((not (symmetric or asymmetric)) or ((ha <> ja) or (ka <> ia))));
      end;
    if (ha = ka) then exit;
    if yn(ha,ka) then ta := 1 else ta := 0;
    end
   else
    begin (* t0 = 0 *)
    if (outd_yn(ha) >= 1) then
      begin
      ka := RandomCurrentOutTie(version,ha);
      ta := 1;
      end
     else
      ka := ha;
     end;
end; (* OtherPairA *)
procedure UpdateStatistics;    overload;
(* overloaded version 1.
   update of current statistics. *)
var h: integer;
begin
  for h:= ActiveNetworkEffects_f downto 1 do
    CurrentStats[h] := CurrentStats[h] + DeltaStats[h];
  inc(GibbsChange);
end; (* UpdateStatistics 1 *)

procedure UpdateStatistics(const i,j: integer; const d: shortint); overload;
(* overloaded version 2.
    update of current statistics and of cumulative outdegrees;
    the old value of yn(i,j) was d; this value was toggled.
    if symmetric then yn(j,i) was also toggled;
    if asymmetric then yn(j,i) was also toggled.
    This procedure is called very often, therefore
    has been written out for efficiency. *)
var h, ijmin,ijmax: integer;
    dd: shortint;
function dif(const d: shortint): shortint;
(* the difference due to a toggle of the old value d *)
begin
  if (d >= 1) then dif := -1 else dif := 1;
end;
begin
  for h:= ActiveNetworkEffects_f downto 1 do
    CurrentStats[h] := CurrentStats[h] + DeltaStats[h];
  inc(GibbsChange);
  if (symmetric or asymmetric) then
    begin
    if (i < j) then
      begin
      ijmin := i;
      ijmax := pred(j);
      dd := dif(d);
      end
     else
      begin
      ijmin := j;
      ijmax := pred(i);
      if symmetric then dd := d else dd := -dif(d);
      end;
    for h:=ijmin to ijmax do Inc(CumulDegrees[h],dd);
    if symmetric then
      begin
      dd := 2*dd;
      for h:=succ(ijmax) to n do Inc(CumulDegrees[h],dd);
      end;
    end  (* if symmetric or asymmetric *)
   else
    begin
    dd := dif(d);
    for h:=i to n do Inc(CumulDegrees[h],dd);
    end;
end; (* UpdateStatistics 2 *)

procedure UpdateStatistics(const i1,j1,i2,j2: integer;
                                              const d1,d2: shortint); overload;
(* overloaded version 3.
    update of current statistics and of cumulative outdegrees;
    the old value of yn(i1,j1) was d1; this value was toggled.
    similar for i2, j2, d2.
    if (a)symmetric - analogous to version 1.
    This procedure is called very often, therefore
    has been written out for efficiency. *)
var h, hh, ijmin, ijmax: integer;
    dd: shortint;
    ij: array of integer;
    dij: array of shortint;
function dif(const d: shortint): shortint;
begin
  if (d >= 1) then dif := -1 else dif := 1;
end;
begin
  for h:= ActiveNetworkEffects_f downto 1 do
    CurrentStats[h] := CurrentStats[h] + DeltaStats[h];
  inc(GibbsChange);
  if asymmetric then
    begin
    if (i1 < j1) then
      begin
      ijmin := i1;
      ijmax := pred(j1);
      dd := dif(d1);
      end
     else
      begin
      ijmin := j1;
      ijmax := pred(i1);
      dd := -dif(d1);
      end;
    for h:=ijmin to ijmax do Inc(CumulDegrees[h],dd);
    if (i2 < j2) then
      begin
      ijmin := i2;
      ijmax := pred(j2);
      dd := dif(d2);
      end
     else
      begin
      ijmin := j2;
      ijmax := pred(i2);
      dd := -dif(d2);
      end;
    for h:=ijmin to ijmax do Inc(CumulDegrees[h],dd);
    end (* if asymmetric *)
   else
    begin
    if symmetric then
      begin
      setlength(ij,4);
      setlength(dij,4);
      ij[0] := pred(i1); dij[0] := dif(d1);
      ij[1] := pred(j1); dij[1] := dif(d1);
      ij[2] := pred(i2); dij[2] := dif(d2);
      ij[3] := pred(j2); dij[3] := dif(d2);
(* For each hh, for ij[hh] < h <= n,
   cumuldegrees[h] must be increased by dij[hh]. *)
      quicksortr(ij,dij,0,3);
      dd := 0;
      for hh:=1 to 3 do
        begin
        inc(dd,dij[pred(hh)]);
        if (dd <> 0) then
          for h := succ(ij[pred(hh)]) to ij[hh] do  Inc(CumulDegrees[h],dd);
        end;
      inc(dd,dij[3]);
      if (dd <> 0) then for h := succ(ij[3]) to n do Inc(CumulDegrees[h],dd);
      end (* if symmetric *)
     else
      begin (* not symmetric and not asymmetric *)
      if (i1 < i2) then
        begin
        ijmin := i1;
        ijmax := pred(i2);
        dd := dif(d1);
        end
       else
        begin
        ijmin := i2;
        ijmax := pred(i1);
        dd := dif(d2);
        end;
      for h:=ijmin to ijmax do Inc(CumulDegrees[h],dd);
      dd := dif(d1)+dif(d2);
      if (dd <> 0) then for h:=succ(ijmax) to n do Inc(CumulDegrees[h],dd);
      end; (* not symmetric and not asymmetric *)
    end; (* not asymmetric *)
end; (* UpdateStatistics 3 *)

begin (* ERGMRunStep *)
    for h := ActiveNetworkEffects_f downto 1 do DeltaStats[h] := 0.0;
(* select i, j *)
    case Gibbs of
      1..3: (* cycling *)
      if ExistStruct[1] then
        begin
        repeat
        inc(jstar);
        if (jstar = istar) then inc(jstar);
        if (jstar > n) then
           begin
           repeat
           inc(istar);
           if (istar > n) then
             begin
             istar := 1; jstar := 2;
             end
             else jstar := 1;
           until StructActive[istar];
           end;
         until (not struct(istar,jstar,1));
        i := istar; j := jstar;
        ynij := yn(i,j);
        end (* if ExistStruct[1] *)
       else
        begin
        inc(jstar);
        if (jstar = istar) then inc(jstar);
        if (jstar > n) then
          begin
          inc(istar);
          if (istar > n) then
            begin
            istar := 1; jstar := 2;
            end
            else jstar := 1;
          end;
        i := istar; j := jstar;
        ynij := yn(i,j);
        end; (* if not ExistStruct[1]; end case 1..3 *)
      4..8:
        begin
(* First select i *)
        if MoreComponents then
          repeat
            repeat
            i := randomd(version,n) + 1;
            until StructActive[i];
            h := ComponentIndicator[i];
           until (((not symmetric) and (ComponentNumber[h,0] >= 2))
                   or (symmetric and (ComponentNumber[h,0] >= 3)))
          else  (* not MoreComponents *)
           repeat
             i := randomd(version,n) + 1;
           until StructActive[i];
  (* Now select j; Note: Gibbs in [4,6,7] = version  A *)
        if ((Gibbs = 4) and Cconditional and ExistStruct[1]) then
          begin
          if (maxnfree*rando(version) > (n - 1 - StructRowElements(i,1))) then exit;
          end;
        if Cconditional or (not (Gibbs in [4,6,7])) or (rando(version) > probA) then
          begin
          if ((Gibbs in [4,6,7]) and (maxnfree*rando(version) > (n - 1 - StructRowElements(i,1)))) then exit;
          repeat
           j := RandomVertexInComponent(i,false)
          until (not Struct(i,j,1));
        (* latter condition is logically redundant because it is checked below,
               but included anyway for efficiency *)
          ynij := yn(i,j);
          end
         else (* i.e. (not cconditional) and (gibbs in [4,6,7]) and (rando < proba ) *)
          begin
          j := RandomCurrentOutTie(version,i);
          ynij := true;
          end;
        if (j = i) then exit;
        end; (* case 4..8 *)
      9:
        begin
(* First select i *)
        if MoreComponents then
          repeat
            repeat i := randomd(version,n) + 1  until (outd[i] >= 1);
            h := ComponentIndicator[i];
           until (ComponentNumber[h,0] >= 3)
          else
           repeat
             i := randomd(version,n) + 1;
           until StructActive[i] and (outd[i] >= 1);
(* then j *)
          j := RandomCurrentOutTie(version,i);
          ynij := true;
         end;  (* case 9 *)
      10:
        begin
(* First select i *)
        if MoreComponents then
          repeat
            repeat i := newranp(version,n,pout) until (outd[i] >= 1);
            h := ComponentIndicator[i];
           until (ComponentNumber[h,0] >= 3)
          else
           repeat
             i := newranp(version,n,pout);
           until StructActive[i] and (outd[i] >= 1);
(* then j *)
          j := RandomCurrentOutTie(version,i);
          ynij := true;
         end;  (* case 10 *)
      end; (* case Gibbs *)
    if ((Outd_yn(i) >= MaxDegree) and (not ynij) and (not cconditional)) then exit;
(* In the mixing case, i and j are necessarily in the same component. *)
(* Gibbs / MH step for this (i,j) *)
    case Gibbs of
    1: begin
       (* Here follows the algorithm for single Gibbs steps *)
       if ynij then tij := 1 else tij := 0;
       if (yn(j,i)) then tr := 1 else tr := 0;
       eu := 0.0;
       ChangeStat_VertPars(eu,tij,tr,i,j);
       if Cconditional then
         begin
(* random choice of other pair (h,k) *)
         OtherPair(h,k,i,j,false);
         ynhk := yn(h,k);
         if ((i <> h) and (((Outd_yn(h) >= MaxDegree) and (not ynhk))
           or ((Outd_yn(i) >= MaxDegree) and (not ynij)))) then exit;
         if (ynij <> ynhk) then
           begin
           ChangeTie(i,j);  (* necessary for calculating change statistics *)
           if ynhk then thk := 1 else thk := 0;
           if (yn(k,h)) then tr := 1 else tr := 0;
           ChangeStat_VertPars(eu,thk,tr,h,k);
           sp := expit(eu + NewNetworkEffects.ChangeLogOdds(DeltaStats));
           if (rando(version) < sp) then
             begin
             ChangeTie(h,k);
             UpdateStatistics;
             end (* if rando < sp *) else ChangeTie(i,j);
             (* (i,j) was changed earlier, now has been changed back again *)
           end; (* if yn(ij) <> yn(hk) *)
         end (* if Cconditional *) else
           begin
           sp := expit(eu + NewNetworkEffects.ChangeLogOdds(DeltaStats));
           if (rando(version) < sp) then
             begin
             ChangeTie(i,j);
             UpdateStatistics;
             end; (* if rando < sp *)
           end; (* if not cconditional *)
       end; (* case Gibbs = 1 *)
    2: begin
       (* Here follows the algorithm for dyadic Gibbs steps *)
       if Cconditional then begin
       (* randomly choose a second dyad *)
         repeat
           h := randomd(version,n) + 1;
           k := randomd(version,n) + 1;
         until ((h <> k) and
               (not (((h = i) and (k = j)) or ((h = j) and (k = i)))));
         (* change dyads (i,j) and (h,k) while keeping dyad count unchanged *)
         sum := one(yn(i,j)) + one(yn(j,i)) + one(yn(h,k)) + one(yn(k,h));
         case sum of
         0,4: exit;
         2: begin
             if (yn(i,j) <> yn(j,i)) then begin
             (* two asymmetric dyads;
                four possibilities for the two dyads (i,j), (h,k):
                ss, sc, cs, cc, where s = same and c = change *)
(* ss; as it were:
     p1 := expo(0) = 1.0;
     pro := 1.0; *)
(* cs *)
         eu1 := ErgmUtilityGain(i,j,halterr,outf,lf);
         ChangeTie(i,j);
         eu := eu1 + ErgmUtilityGain(j,i, halterr,outf,lf);
         pro := expo(eu);
         p2 := 1.0 + pro;
         ChangeTie(j,i);
(* cc *)
         eu1 := ErgmUtilityGain(h,k, halterr,outf,lf);
         ChangeTie(h,k);
         eu := eu1 + ErgmUtilityGain(k,h, halterr,outf,lf);
         p3 := p2 + pro*expo(eu);
         ChangeTie(k,h);
(* sc *)
         eu1 := ErgmUtilityGain(i,j,halterr,outf,lf);
         ChangeTie(i,j);
         eu := eu1 + ErgmUtilityGain(j,i, halterr,outf,lf);
         p4 := p3 + pro*expo(eu);
(* now the ties yn(ij), yn(ji), yn(hk), yn(kh) are in the state sccc
   determine the new state in 2 bisection steps : *)
         ran := rando(version)*p4;
         if (ran < p2) then begin
             ChangeTie(h,k);
             ChangeTie(k,h);
             if (ran < 1.0) then begin
               ChangeTie(j,i);
              end else begin
               ChangeTie(i,j);
               Inc(GibbsChange);
              end;
           end else (* ran >= p2 *) begin
            if (ran < p3) then ChangeTie(i,j) else ChangeTie(j,i);
            Inc(GibbsChange);
           end; (* if ran >= p2 *)
           end (* if yn(ij) <> yn(ji) *) else begin (* one null and one mutual dyad *)
            eu := ErgmUtilityGain(i,j,halterr,outf,lf);
            ChangeTie(i,j);
            eu := eu + ErgmUtilityGain(j,i, halterr,outf,lf);
            ChangeTie(j,i);
            eu := eu + ErgmUtilityGain(h,k, halterr,outf,lf);
            ChangeTie(h,k);
            eu := eu + ErgmUtilityGain(k,h, halterr,outf,lf);
            sp := eu/(1.0+eu);
            if (rando(version) < sp) then begin
              ChangeTie(k,h);
              Inc(GibbsChange);
             end else begin (* change everything back again *)
              ChangeTie(i,j);
              ChangeTie(j,i);
              ChangeTie(h,k);
             end;
            end; (* if one null and one mutual *)
          end; (* case sum = 2 *)
            1,3: begin
(* one asymmetric and one symmetric (null or mutual) dyad *)
              if (yn(i,j) <> yn(j,i)) then begin
                if (yn(i,j) = yn(h,k)) then begin
                  a1 := i; a2 := j;
                 end else begin
                  a1:= j; a2 := i; end;
                b1 := h; b2 := k;
               end else (* yn(ij) = yn(ji) *) begin
                if (yn(i,j) = yn(h,k)) then begin
                  a1 := h; a2 := k;
                 end else begin
                  a1 := k; a2 := h; end;
                b1 := i; b2 := j;
               end;
(* now (a1,a2) is asymmetric and (b1,b2) is symmetric,
    while yn(a1,a2) = yn(b1,b2) = yn(b2,b1)  .
   allowed new states for yn(a1,a2), yn(a2,a1), yn(b1,b2), yn(b2,b1)
   are ssss, ccss, sccs, scsc, where s = same, c = changed *)
(* ssss : p1 := 1.0 *)
(* ccss : *)
            eu1 := ErgmUtilityGain(a1,a2, halterr,outf,lf);
            ChangeTie(a1,a2);
            p2 := 1.0 + expo(eu1+ErgmUtilityGain(a2,a1, halterr,outf,lf));
            ChangeTie(a1,a2);
(* sccs : *)
            eu1 := ErgmUtilityGain(a2,a1, halterr,outf,lf);
            ChangeTie(a2,a1);
            p3 := p2 + expo(eu1+ErgmUtilityGain(b1,b2, halterr,outf,lf));
(* scsc : *)
            p4 := p3 + expo(eu1+ErgmUtilityGain(b2,b1, halterr,outf,lf));
(* now the ties yn(a1,a2), yn(a2,a1), yn(b1,b2), yn(b2,b1)
   are in the state scss.
   determine the new state in 2 bisection steps : *)
            ran := rando(version)*p4;
            if (ran < p2) then begin
              if (ran < 1.0) then begin
                ChangeTie(a2,a1);
               end else begin
                ChangeTie(a1,a2);
                Inc(GibbsChange);
               end;
             end else (* ran >= p2 *) begin
               if (ran < p3) then ChangeTie(b1,b2) else ChangeTie(b2,b1);
               Inc(GibbsChange);
              end; (* if ran >= p2 *)
             end;  (* case sum = 1,3 *)
            end; (* case sum  *)
        end else begin (* not Cconditional, still Gibbs = 2 *)
(* four possibilities for the dyad (yn(i,j), yn(j,i)):
   ss, sc, cs, cc, where s = same and c = change  *)
(* ss; as it were:
     pro:= expo(0) = 1.0;
     p1 := pro; *)
(* cs *)
         pro := expo(ErgmUtilityGain(i,j,halterr,outf,lf));
         p2 := 1.0 + pro;
         ChangeTie(i,j);
(* cc *)
         pro := pro*expo(ErgmUtilityGain(j,i, halterr,outf,lf));
         p3 := p2 + pro;
         ChangeTie(i,j);
(* sc *)
         pro := expo(ErgmUtilityGain(j,i, halterr,outf,lf)); (* without pro ! *)
         p4 := p3 + pro;
(* now the dyad is in the original state again.
   determine the new state in 2 bisection steps : *)
         ran := rando(version)*p4;
         if (ran < p2) then begin
           if (ran < 1.0) then (* no change *)
            else begin
              ChangeTie(i,j);
              Inc(GibbsChange);
             end;
          end else (* ran >= p2 *) begin
            ChangeTie(j,i);
            if (ran < p3) then ChangeTie(i,j);
            Inc(GibbsChange);
           end; (* if ran >= p2 *)
         end; (* if not Cconditional *)
       end; (* case Gibbs = 2 *)
    3: begin
(* Here follows the algorithm for triplet Gibbs steps *)
      repeat
        h := randomd(version,n) + 1
       until ((h <> i) and (h <> j));
      if (rando(version) < 0.5) then begin
        ii := i; hh := h; end
        else begin
         ii := h; hh := i; end;
(* The triplet chosen now is (i,j), (j,h), (ii,hh) *)
      if Cconditional then begin
        sum := one(yn(i,j)) + one(yn(j,h)) + one(yn(ii,hh));
        if ((sum <= 0) or (sum >= 3 )) then exit;
        tie1 := yn(i,j);
        tie2 := yn(j,h);
        tie3 := yn(ii,hh);
        (* of tie1, tie2, tie3, two are the same and one is different *)
        if (tie1 = tie2) then begin
          a1 := ii; a2 := hh;
          b1 := i;  b2 := j;
          c1 := j;  c2 := h;
         end else begin if (tie1 = tie3) then begin
            a1 := j;  a2 := h;
            b1 := i;  b2 := j;
            c1 := ii; c2 := hh;
           end else (* tie2 = tie3 *) begin
            a1 := i;  a2 := j;
            b1 := j;  b2 := h;
            c1 := ii; c2 := hh;
           end;
         end;
        (* now yn(a1,a2) differs from yn(b1,b2) = yn(c1,c2);
            this means that, to keep the density constant,
            the only allowed changes are interchanging
            yn(a1,a2) with either yn(b1,b2) or yn(c1,c2) *)
        eu := ErgmUtilityGain(a1,a2, halterr,outf,lf);
        ChangeTie(a1,a2);
        eu1 := expo(eu + ErgmUtilityGain(b1,b2, halterr,outf,lf));
        eu2 := expo(eu + ErgmUtilityGain(c1,c2, halterr,outf,lf));
        ran := rando(version)*(1.0+eu1+eu2);
        if (ran < 1.0) then begin (* change yn(a1,a2) back again *)
          ChangeTie(a1,a2);
         end else (* a change must be made *) begin
           if (ran < (1.0+eu1)) then begin
            (* interchange yn(a1,a2) with yn(b1,b2) *)
              ChangeTie(b1,b2);
            end else begin
            (* interchange yn(a1,a2) with yn(c1,c2) *)
              ChangeTie(c1,c2);
            end;
           inc(GibbsChange);
         end;  (* if ran >= 1.0 *)
    end  else (* not Cconditional *) begin
(* The triplet to be determined still is (i,j), (j,h), (ii,hh) ;
   if (ii,hh) = (i,h) this is a potential transitive triplet,
   if (ii,hh) = (h,i) this is a potential cyclical triplet.
  Not the 0/1 values of these 3 ties will be considered,
  but whether they are to be changed.
  The notation for the triplet is xxx
  where x can be c (change) or s (same = unchanged),
  and the three coordinates refer to (i,j), (j,h), (ii,hh) respectively.
  The values p1 to p8 are multiples of the cumulative probabilities
  of the 8 possible states of xxx.  *)
(* sss; as it were:
     p1 := 1.0;
     pro := 1.0; *)
(* css *)
     pro1 := expo(ErgmUtilityGain(i,j,halterr,outf,lf));
     p2 := 1.0 + pro1;
(* scs *)
     pro := expo(ErgmUtilityGain(j,h, halterr,outf,lf));
     p3 := p2 + pro;
(* ssc *)
     pro2 := expo(ErgmUtilityGain(ii,hh, halterr,outf,lf));
     p4 := p3 + pro2;
     ChangeTie(i,j); (* now in state css *)
(* ccs *)
     pro := pro1*expo(ErgmUtilityGain(j,h, halterr,outf,lf));
     p5 := p4 + pro;
(* csc *)
     pro := pro1*expo(ErgmUtilityGain(ii,hh, halterr,outf,lf));
     p6 := p5 + pro;
     ChangeTie(i,j);
     ChangeTie(ii,hh); (* now in state ssc *)
(* scc *)
     pro2 := pro2*expo(ErgmUtilityGain(j,h, halterr,outf,lf));
     p7 := p6 + pro2;
     ChangeTie(i,j); (* now in state csc *)
(* ccc *)
     pro := pro*expo(ErgmUtilityGain(j,h, halterr,outf,lf));
     p8 := p7 + pro;
(* Now the triplet is in state csc.
   Determine the new xxx in 3 bisection steps: *)
     ran := p8*rando(version);
     if (ran <= p4) then begin
       if (ran <= p2) then begin
         if (ran <= 1.0) then begin (* sss *)
           ChangeTie(i,j);
           ChangeTie(ii,hh);
          end else begin (* css *)
           ChangeTie(ii,hh);
           Inc(GibbsChange);
          end;
        end else (* ran > p2 *) begin
         if (ran <= p3) then begin (* scs *)
           ChangeTie(i,j);
           ChangeTie(j,h);
           ChangeTie(ii,hh);
          end else begin (* ssc *)
           ChangeTie(i,j);
          end;
         Inc(GibbsChange);
        end; (* if ran > p2 *)
        end else (* ran > p4 *) begin
       if (ran <= p6) then begin
         if (ran <= p5) then begin (* ccs *)
           ChangeTie(ii,hh);
           ChangeTie(j,h);
          end else begin (* csc *)
            ;
          end;
        end else (* ran > p6 *) begin
         if (ran <= p7) then begin (* scc *)
           ChangeTie(i,j);
           ChangeTie(j,h);
          end else begin (* ccc *)
           ChangeTie(j,h);
          end;
         end; (* if ran > p6 *)
        Inc(GibbsChange);
       end; (* if ran > p4 *)
      end; (* if not Cconditional *)
     end; (* case Gibbs = 3 *)
    4: begin
(* Here follows the algorithm for Metropolis Hastings steps, version A.
   Described in ERGM_algor.tex *)
       if ynij then tij := 1 else tij := 0;
       if yn(j,i) then tr := 1 else tr := 0;
       eu := 0.0;
       ChangeStat_VertPars(eu,tij,tr,i,j);
       if (not Cconditional) then
         begin
         if (tij >= 1) then
         eu := eu - lograt[outd_yn(i)] + NewNetworkEffects.ChangeLogOdds(DeltaStats)
          else eu := eu + lograt[outd_yn(i)+1] + NewNetworkEffects.ChangeLogOdds(DeltaStats);
         if (eu >= 0) then
           begin
           ChangeTie(i,j);
           UpdateStatistics;
          end else begin
           eu := expo(eu);
           if (rando(version) < eu) then
             begin
             ChangeTie(i,j);
             UpdateStatistics;
             end (* if rando < eu *);
           end; (* if eu < 0 *)
         end (* if not Cconditional *)
        else
         begin (* cconditional *)
         OtherPairA(i,j,tij,h,k,thk);
         if (h=k) then exit;
         if ((i <> h) and (((Outd_yn(h) >= MaxDegree) and (thk <= 0))
           or ((Outd_yn(i) >= MaxDegree) and (tij <= 0)))) then exit;
         if (tij <> thk) then
(* decide if y(i,j) and y(h,k) will be swapped *)
           begin
           ChangeTie(i,j);
(* The next statement must be put after ChangeTie(i,j),
   because of the possibility that (h,k) = (j,i). *)
           if yn(k,h) then tr := 1 else tr := 0;
           ChangeStat_VertPars(eu,thk,tr,h,k);
           if (tij >= 1) then
             eu := eu + lograt[outd_yn(h)+1] - lograt[outd_yn(i)]
                      + NewNetworkEffects.ChangeLogOdds(DeltaStats)
            else
             eu := eu + lograt[outd_yn(i)+1] - lograt[outd_yn(h)]
                   + NewNetworkEffects.ChangeLogOdds(DeltaStats);
           if (eu >= 0) then
             begin
             ChangeTie(h,k);
             UpdateStatistics;
             end
            else
             begin
             eu := expo(eu);
             if (rando(version) < eu) then
               begin
               ChangeTie(h,k);
               UpdateStatistics;
               end (* if rando < eu *)
              else
               begin
            (* change everything back again *)
                ChangeTie(i,j);
               end; (* if rando >= eu *)
             end; (* if eu < 0 *)
           end; (* if yn(ij) <> yn(hk) *)
         end; (* if Cconditional) *)
       end; (* case Gibbs = 4 *)
    5: begin
(* Here follows the algorithm for Metropolis Hastings steps, version B *)
       if ynij then tij := 1 else tij := 0;
       if (yn(j,i)) then tr := 1 else tr := 0;
       eu := 0.0;
       ChangeStat_VertPars(eu,tij,tr,i,j);
       if Cconditional then
         begin
         OtherPair(h,k,i,j,true);
         if (h = k) then exit;
         ynhk := yn(h,k);
         if ((i <> h) and (((Outd_yn(h) >= MaxDegree) and (not ynhk))
           or ((Outd_yn(i) >= MaxDegree) and (not ynij)))) then exit;
         if (ynij <> ynhk) then
           begin
           ChangeTie(i,j);  (* necessary for calculating change statistics *)
           if ynhk then thk := 1 else thk := 0;
           if (yn(k,h)) then tr := 1 else tr := 0;
           ChangeStat_VertPars(eu,thk,tr,h,k);
           eu := eu + NewNetworkEffects.ChangeLogOdds(DeltaStats);
           if (eu >= 0) then
             begin
             ChangeTie(h,k);
             UpdateStatistics(i,j,h,k,tij,thk);
             end
            else
             begin
             eu := expo(eu);
             if (rando(version) < eu) then
               begin
               ChangeTie(h,k);
               UpdateStatistics(i,j,h,k,tij,thk);
               end (* if rando < eu *)
              else ChangeTie(i,j);
              (* (i,j) was changed earlier, now has been changed back again *)
            end; (* if eu < 0 *)
          end; (* if yn(ij) <> yn(hk) *)
        end (* if Cconditional *)
       else
        begin
        eu := eu + NewNetworkEffects.ChangeLogOdds(DeltaStats);
        if (eu >= 0) then
          begin
          ChangeTie(i,j);
          UpdateStatistics(i,j,tij);
          end
         else
          begin
          eu := expo(eu);
          if (rando(version) < eu) then
            begin
            ChangeTie(i,j);
            UpdateStatistics(i,j,tij);
            end; (* if rando < eu *)
          end; (* if eu < 0 *)
        end; (* if not cconditional *)
       end; (* case Gibbs = 5 *)
    6: begin
(* Here follows the algorithm for Metropolis Hastings steps (type A)
      for undirected graphs;
     must be carried out only if symmetric;
    analogous to 4, except that the two symmetrically positioned elements
    both must be changed.
    This was in error in version up to 1.96. Corrected in 1.97.
    Corrected in 2.0j.
    If there are structurally determined values,
    this option may be used only when these values are symmetric.
    This is checked in ReviseDefinition. *)
       if ynij then
         begin
         if GibbsConnected then
           begin
           if (not OtherPathNetwork(i,j)) then exit;
(* then the network would become disconnected by deleting the tie i-j *)
           end;
         tij := 1;
         end
        else tij := 0;
       eu := 0.0;
       ChangeStat_VertPars(eu,tij,tij,i,j);
       if (not Cconditional) then
         begin
         if (tij >= 1) then
         eu := eu - lograt[outd_yn(i)] + NewNetworkEffects.ChangeLogOdds(DeltaStats)
          else eu := eu + lograt[outd_yn(i)+1] + NewNetworkEffects.ChangeLogOdds(DeltaStats);
         if (eu >= 0) then
           begin
           ChangeTie(i,j);
           ChangeTie(j,i);
           UpdateStatistics;
          end else begin
           eu := expo(eu);
           if (rando(version) < eu) then
             begin
             ChangeTie(i,j);
             ChangeTie(j,i);
             UpdateStatistics;
             end (* if rando < eu *);
           end; (* if eu < 0 *)
         end (* if not Cconditional *) else
           begin (* cconditional *)
           OtherPairA(i,j,tij,h,k,thk);
           if (h=k) then exit;
           if ((i <> h) and (((Outd_yn(h) >= MaxDegree) and (thk <= 0))
             or ((Outd_yn(i) >= MaxDegree) and (tij <= 0)))) then exit;
           if (tij <> thk) then
             begin
             ChangeTie(i,j);
             ChangeTie(j,i);
             ChangeStat_VertPars(eu,thk,thk,h,k);
             if (tij >= 1) then
               eu := eu + lograt[outd_yn(h)+1] - lograt[outd_yn(i)]
                      + NewNetworkEffects.ChangeLogOdds(DeltaStats)
              else
               eu := eu + lograt[outd_yn(i)+1] - lograt[outd_yn(h)]
                      + NewNetworkEffects.ChangeLogOdds(DeltaStats);
             if (eu >= 0) then
               begin
               ChangeTie(h,k);
               ChangeTie(k,h);
               if GibbsConnected then
               begin
               if (tij <= 0) then
                 begin
                 if (not PathNetwork(h,k)) then
(* then the network would become disconnected by deleting the tie h-k;
  change everything back again. *)
                   begin
                   ChangeTie(h,k);
                   ChangeTie(k,h);
                   ChangeTie(i,j);
                   ChangeTie(j,i);
                   exit;
                   end;
                 end
                else (* thk = 0 *)
                 begin
                 if (not PathNetwork(i,j)) then
(* then the network would become disconnected by deleting the tie i-j;
  change everything back again. *)
                   begin
                   ChangeTie(h,k);
                   ChangeTie(k,h);
                   ChangeTie(i,j);
                   ChangeTie(j,i);
                   exit;
                   end;
(* then the network would become disconnected by deleting the tie i-j *)
                 end;
               end;
               UpdateStatistics;
               end
              else
               begin
               eu := expo(eu);
               if (rando(version) < eu) then
                 begin
                 ChangeTie(h,k);
                 ChangeTie(k,h);
                 if GibbsConnected then
                 begin
                 if (tij <= 0) then
                   begin
                   if (not PathNetwork(h,k)) then
(* then the network would become disconnected by deleting the tie h-k;
  change everything back again. *)
                     begin
                     ChangeTie(h,k);
                     ChangeTie(k,h);
                     ChangeTie(i,j);
                     ChangeTie(j,i);
                     exit;
                     end;
                   end
                  else (* thk = 0 *)
                   begin
                   if (not PathNetwork(i,j)) then
(* then the network would become disconnected by deleting the tie i-j;
  change everything back again. *)
                     begin
                     ChangeTie(h,k);
                     ChangeTie(k,h);
                     ChangeTie(i,j);
                     ChangeTie(j,i);
                     exit;
                     end;
(* then the network would become disconnected by deleting the tie i-j *)
                   end;
                 end;
                 UpdateStatistics;
                 end (* if rando < eu *)
                else
                 begin
              (* change everything back again *)
                 ChangeTie(i,j);
                 ChangeTie(j,i);
                 end; (* if rando >= eu *)
               end; (* if eu < 0 *)
             end (* if yn(ij) <> yn(hk) = yn(kh) *);
           end; (* if Cconditional) *)
        end; (* case Gibbs = 6 *)
    7: begin
(* Here follows the algorithm for Metropolis Hastings steps
      for antisymmetric graphs;
     must b carried out only if asymmetric;
    If there are structurally determined values,
    this option may be used only when these values are symmetric.
    This is checked in ReviseDefinition. *)
       if ynij then tij := 1 else tij := 0;
       tr := 1-tij;
       eu := 0.0;
       ChangeStat_VertPars(eu,tij,tr,i,j);
       ChangeStat_VertPars(eu,tij,tij,j,i);  // @@@ check tij, tij
       eu := eu + NewNetworkEffects.ChangeLogOdds(DeltaStats);
       if (eu >= 0) then
         begin
         ChangeTie(i,j);
         ChangeTie(j,i);
         UpdateStatistics;
        end else begin
         eu := expo(eu);
         if (rando(version) < eu) then
           begin
           ChangeTie(i,j);
           ChangeTie(j,i);
           UpdateStatistics;
           end (* if rando < eu *);
         end; (* if eu < 0 *)
        end; (* case Gibbs = 7 *)
    8: begin
(* Algorithm for Metropolis Hastings steps, conditioning
   on in - and outdegrees *)
       if (rando(version) < 0.5) then begin
(* switch alternating rectangle: swap y_ih, y_jk with y_ik, y_jh
   if y_ih = y_jk <> y_ik = y_jh and MH probabilistic condition is satisfied*)
         repeat
           h := randomd(version,n) + 1;
           k := randomd(version,n) + 1;
         until ((h <> k) and (h <> i) and (h <> j) and (k <> i) and (k <> j));
         yih := yn(i,h);
         if ((yn(j,k) = yih) and (yn(i,k) <> yih) and (yn(j,h) <> yih)) then begin
           eu := ErgmUtilityGain(i,h, halterr,outf,lf);
           ChangeTie(i,h);
           eu := eu + ErgmUtilityGain(i,k, halterr,outf,lf);
           ChangeTie(i,k);
           eu := eu + ErgmUtilityGain(j,h, halterr,outf,lf);
           ChangeTie(j,h);
           eu := eu + ErgmUtilityGain(j,k, halterr,outf,lf);
           if (eu >= 0) then begin
             inc(GibbsChange);
             ChangeTie(j,k);
            end else begin
             eu := expo(eu);
             if (rando(version) < eu) then begin
               inc(GibbsChange);
               ChangeTie(j,k);
             end (* if rando < eu *) else begin
           (* change everything back again *)
               ChangeTie(i,h);
               ChangeTie(i,k);
               ChangeTie(j,h);
              end; (* if rando >= eu *)
            end; (* if eu < 0 *)
         end; (* else configuration i,j,h,k, allows no change *)
        end (* if rando(version < 0.5) *) else begin
(* switch alternating triad *)
         repeat
           h := randomd(version,n) + 1
         until ((h <> i) and (h <> j));
         yih := yn(i,h);
         if ((yn(h,j) = yih) and (yn(j,i) = yih) and (yn(i,j) <> yih)
          and (yn(j,h) <> yih) and (yn(h,i) <> yih)) then begin
           eu := ErgmUtilityGain(i,j,halterr,outf,lf);
           ChangeTie(i,j);
           eu := eu + ErgmUtilityGain(j,h, halterr,outf,lf);
           ChangeTie(j,h);
           eu := eu + ErgmUtilityGain(h,i, halterr,outf,lf);
           ChangeTie(h,i);
           eu := eu + ErgmUtilityGain(j,i, halterr,outf,lf);
           ChangeTie(j,i);
           eu := eu + ErgmUtilityGain(h,j, halterr,outf,lf);
           ChangeTie(h,j);
           eu := eu + ErgmUtilityGain(i,h, halterr,outf,lf);
           if (eu >= 0) then begin
             inc(GibbsChange);
             ChangeTie(i,h);
            end else begin
             eu := expo(eu);
             if (rando(version) < eu) then begin
               inc(GibbsChange);
               ChangeTie(i,h);
             end (* if rando < eu *) else begin
           (* change everything back again *)
               ChangeTie(i,j);
               ChangeTie(j,h);
               ChangeTie(h,i);
               ChangeTie(j,i);
               ChangeTie(h,j);
              end; (* if rando >= eu *)
             end; (* if eu < 0 *)
            end; (* else configuration i,j,h allows no change *)
        end;
       end; (* case Gibbs = 8 *)
    9, 10: begin
(* Metropolis Hastings steps, fixed out-degrees *)
(* (i,j) has been chosen so that y(i,j) = 1 *)
       if yn(j,i) then tr := 1 else tr := 0;
       eu := 0.0;
       ChangeStat_VertPars(eu,1,tr,i,j);
       if (Gibbs <= 9) or MoreComponents then
         begin
         repeat k := RandomVertexInComponent(i,false)
           until  (k <> j) and (not(Struct(i,k,1)));
         end
        else (* Gibbs = 10 and not MoreComponents *)
         begin
         repeat k := newranp(version,n,pin) until (k <> j);
         if ((k = i) or Struct(i,k,1)) then exit;
           (* this is not included in the "until" condition
              to keep the proposal probabilities simple *)
         end;
       if yn(i,k) then exit;
(* now y(i,j) = 1, y(i,k) = 0 *)
(* decide if y(i,j) and y(i,k) will be swapped *)
       ChangeTie(i,j);
       if yn(k,i) then tr := 1 else tr := 0;
       ChangeStat_VertPars(eu,0,tr,i,k);
       eu := eu + NewNetworkEffects.ChangeLogOdds(DeltaStats);
       if (gibbs >= 10) and (not MoreComponents) then eu := eu + lograt[j] - lograt[k];
       if (eu >= 0) then
         begin
         ChangeTie(i,k);
         UpdateStatistics;
         end
        else
         begin
         eu := expo(eu);
         if (rando(version) < eu) then
           begin
           ChangeTie(i,k);
           UpdateStatistics;
           end (* if rando < eu *)
          else
           begin
        (* change everything back again *)
            ChangeTie(i,j);
           end; (* if rando >= eu *)
         end; (* if eu < 0 *)
       end; (* case Gibbs = 9, 10 *)
    end; (* case Gibbs *)
    if halterr then begin
      erstr := 'error in ErgmUtilityGain';
      error := true;
    end;
    if MadeChangeTieError then begin
      erstr := ChangeTieErrorText;
      error := true;
     end;
(* the following may be useful for checking
   when there are problms with the running statistics *)
(* write(outf,'i,j  ',i:4,j:4,tij:3);
if cconditional then writeln(outf,h:6,k:4,thk:3) else writeln(outf);
for i:= 1 to n do begin
  for j:=1 to n do write(outf,ynv(i,j):2);
  writeln(outf); end;
  writeln(outf);      *)
//checkrunningstatistics(outf);
end; (* ModelSpecification.ErgmRunStep *)


function ModelSpecification.ErgmUtilityGain(i,j: integer; var halterr: Boolean;
                      var outf,lf: TFileStream): real;
(*  f(x(i \leadsto j)) - f(x); for the Ergm case;
    For the symmetric = undirected case,
    this must be the gain for changing x(i,j) as well as x(j,i).
    This procedure is used only in the non-opimized Gibbs cases
    and in PseudosDeviance used in MLEstimateParameter. *)
var u: real;
    h: integer;
    t,tr: shortint;
    s: pvec;
begin
  if yn(i,j) then t := 1 else t := 0;
  if yn(j,i) then tr := 1 else tr := 0;
  for h:=ActiveNetworkEffects_f downto 1 do s[h] := 0.0;
  ERGMChangeStatistics(i,j,s,t,tr,NewNetworkEffects);
  if VertexPar then
    begin
    if (t >= 1) then u := - vertex_a[i] - vertex_b[j]
     else u := vertex_a[i] + vertex_b[j];
    end;
  ergmutilitygain := u + NewNetworkEffects.ChangeLogOdds(s);
end; (* ErgmUtilityGain *)


procedure ModelSpecification.InitialiseVertexArrays;
(* used only in the VertexPar option *)
(* Note that this procedure uses ynm(i,j,1). *)
var i,j: integer;
  suma, sumb: real;
begin
  setlength(targeta,n+1);
  setlength(facta,n+1);
  setlength(avera,n+1);
  setlength(targetb,n+1);
  setlength(factb,n+1);
  setlength(averb,n+1);
  for i:=1 to n do begin
    targeta[i] := 0.0; targetb[i] := 0.0;
    facta[i] := 0.0; factb[i] := 0.0;
   end;
  for i:=1 to n do for j:=1 to n do
   if ((i <> j) and ynm(i,j,1)) then begin
     targeta[i] := targeta[i] + 1.0;
     targetb[j] := targetb[j] + 1.0;
    end; (* if *)
(* now targeta and targetb are the out- and in-degrees *)
  for i:=1 to n do begin
    if ((targeta[i] > 0.5) and (targeta[i] < n - 1.5))
     then facta[i] := 5.0*(n-1.0)/(targeta[i]*(n-1.0-targeta[i]))
     else facta[i] := 0.1;
    if ((targetb[i] > 0.5) and (targetb[i] < n - 1.5))
     then factb[i] := 5.0*(n-1.0)/(targetb[i]*(n-1.0-targetb[i]))
     else factb[i] := 0.1;
    if (InitialStandard or VertexParNew) then begin
      vertex_a[i] := 2.0*(2.0*targeta[i]-n+1.0)/(n-1.0);
      vertex_b[i] := 2.0*(2.0*targetb[i]-n+1.0)/(n-1.0);
      VertexParNew := false;
     end;
   end; (* for  *)
(* redistribute averages for vertex_a and vertex_b *)
  suma := 0.0; sumb := 0.0;
  for i:=1 to n do begin
    suma := suma + vertex_a[i];
    sumb := sumb + vertex_b[i];
   end;
  suma := suma/n; sumb := sumb/n;
  for i:=1 to n do begin
    vertex_a[i] := vertex_a[i] - 0.5*(suma-sumb);
    vertex_b[i] := vertex_b[i] + 0.5*(suma-sumb);
   end;
end; (* ModelSpecification.InitialiseVertexArrays *)


procedure ModelSpecification.VertexStep(ga: real; var outf: TFileStream);
(* used only in the VertexPar option *)
var i: integer;
x: string;
begin
 for i:=n downto 1 do begin
   vertex_a[i] := vertex_a[i] - ga*facta[i]*(Outd_yn(i)-targeta[i]);
//   vertex_a[i] := vertex_a[i] - ga*facta[i]*sgn(Outd_yn(i)-targeta[i]);
   if (abs(vertex_a[i]) > 90.0) then
    if (vertex_a[i] > 0.0) then vertex_a[i] := 90.0
     else vertex_a[i] := -90.0;
   avera[i] := avera[i] + vertex_a[i];
   vertex_b[i] := vertex_b[i] - ga*factb[i]*(Ind_yn(i)-targetb[i]);
//   vertex_b[i] := vertex_b[i] - ga*factb[i]*sgn(Ind_yn(i)-targetb[i]);
   if (abs(vertex_b[i]) > 90.0) then
    if (vertex_b[i] > 0.0) then vertex_b[i] := 90.0
     else vertex_b[i] := -90.0;
   averb[i] := averb[i] + vertex_b[i];
   if (Outd_yn(i) < 0) then
     x := format('%s%d%s%8d%s',['warning: outd[',i,']',Outd_yn(i),' !!!!!!!!!!!!!!!' ])+eol;
     outf.Write(pchar(x)^,length(x));
  end; (* for i *)
end; (* ModelSpecification.VertexStep *)


procedure ModelSpecification.ReportVertexParameters(var outf: TFileStream);
(* used only in the VertexPar option *)
var i: integer;
x: string;
begin
   x := 'Incidental vertex parameters :'+eol;
   outf.Write(pchar(x)^,length(x));
   x := 'for outdegrees :'+eol;
   outf.Write(pchar(x)^,length(x));
   for i:=1 to n do begin
     x := format('%s%7.3f',[' ',vertex_a[i]]);
     outf.Write(pchar(x)^,length(x));
     if ((i mod 10) = 0) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end;
   if ((i mod 10) <> 0) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
   x := 'for in-degrees :'+eol;
   outf.Write(pchar(x)^,length(x));
   for i:=1 to n do begin
     x:= format('%s%7.3f',[' ',vertex_b[i]]);
     outf.Write(pchar(x)^,length(x));
     if ((i mod 10) = 0) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end;
   if ((i mod 10) <> 0) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
   outf.Write(pchar(eol)^,length(eol));//writeln(outf);
end; (*  ModelSpecification.ReportVertexParameters *)


procedure ModelSpecification.InitialiseVertexAverage;
(* used only in the VertexPar option *)
var i: integer;
begin
  for i:=1 to n do begin
    avera[i] := vertex_a[i];
    averb[i] := vertex_b[i];
   end;
end;

procedure ModelSpecification.VertexAverage(nai: Cardinal);
(* used only in the VertexPar option *)
var i: integer;
begin
  for i:=1 to n do begin
    vertex_a[i] := avera[i]/(nai+1.0);
    vertex_b[i] := averb[i]/(nai+1.0);
   end; (* i *)
end; (* ModelSpecification.VertexAverage *)


function ModelSpecification.PseudoDeviance(theta: pvec; var halterr: Boolean;
                      var outf,lf: TFileStream): real;
(* Pseudo-deviance for observations=1 case *)
var i,j: integer; ps: real;
begin
 SetAlpha(theta);
 if halterr then begin
   PseudoDeviance := 0.0;
   exit;
  end;
 ps := 0.0;
 if (symmetric or asymmetric) then begin
   for i:=2 to n do for j:= i-1 downto 1 do begin
     ps := ps + 2.0*log(1.0+expo(ErgmUtilityGain(i,j,halterr,outf,lf)));
    end;
  end else begin
   for i:=1 to n do for j:=1 to n do
    if not (i = j) then
       ps := ps + 2.0*log(1.0 + expo(ErgmUtilityGain(i,j,halterr,outf,lf)));
  end;
 PseudoDeviance := ps;
end;


