unit s_matrices;
(*Entire unit by MS: convenient, basic matrix routines.*)

(*--------------------------------------------------------------------------------------------------*)

interface

uses  classes, rangen, slib, Sysutils, s_constants;

const
  tin = 1e-16; (*Serves as threshold.*)
  sequences = 1; (*The number of sequences/runs to be executed.*)

type
  R_Ar = array of real;
  rrarray = array of array of real;
  rrrarray = array of array of array of real;
  barray = array of byte;
  iarray = array of integer;
  iiarray = array of array of integer;
  boarray = array of boolean;
  boboarray = array of array of boolean;
  blackarray = array[0..20] of real;

var
  dim, dim1, dim2, lag: byte; (*Indicate the dimensions of matrices used in the matrix routines below.*)
  ldim, ldim1, ldim2, run: integer;
  cdim: cardinal;
  determinant, real1, real2, scale: real;
  boolean1, boolean2: boolean;
  filename, input, input1, input2, input3, output: shortstring;
  ascii, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11: textfile;
  matrix1, matrix2: pmatrix;
  covI, covII, covIII, matrix3, matrix4, matrix5: rrarray;
  matrix6, matrix7: rrrarray;
  vector1, vector2, vector3, vector4, vector10: pvec; (*Used in matrix routines below.*)
  vector5, vector6, vector7, vector8: R_Ar;
  bvector: boarray;
  bmatrix: boboarray;
  blackbox: blackarray; (*A blackbox in both the traditional and modern sense: basic information on the behavior of the Markov chain.*)
  threedim: rrrarray;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixProd(var dim1, dim2: byte; var result, matrix1, matrix2: pmatrix); overload;
(*Computes <result = matrix1 * matrix2>
where
<matrix1> is of order <dim1 x dim2>
<matrix2> is of order <dim2 x dim2>.*)
procedure MatrixProd(var product: rrarray; var ldim: integer; var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <product = matrix3 * matrix4>
where
<matrix3> is of order <ldim x dim>
<matrix4> is of order <dim x dim>.*)
procedure MatrixProd(const dim1, dim2: byte; var product: rrarray; const matrix1: pmatrix; const matrix4: rrarray); overload;
(*Computes <product = matrix1 * matrix4>
where
<matrix1> is of order <dim1 x dim2>
<matrix4> is of order <dim2 x dim2>.*)
procedure MatrixProd(const dim1, dim2: byte; var product: rrarray; const matrix3: rrarray; const matrix1: pmatrix); overload;
(*Computes <product = matrix3 * matrix1>
where
<matrix3> is of order <dim1 x dim2>
<matrix1> is of order <dim2 x dim2>.*)
procedure MatrixProd(const dim1, dim2: byte; var product: rrarray; const matrix3, matrix4: rrarray); overload;
(*Computes <product = matrix3 * matrix4>
where
<matrix3> is of order <dim1 x dim2>
<matrix4> is of order <dim2 x dim2>.*)
procedure MatrixProd(var product: rrarray; var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <product = matrix3 * matrix4>
where
<matrix3> is of order <dim x dim>
<matrix4> is of order <dim x dim>.*)

procedure MatrixP(const dim1, dim2, dim4: byte; var result: pmatrix; const matrix1, matrix2: pmatrix); overload;
(*General matrix product: <result = matrix1 * matrix2>
where
<matrix1> is of order <dim1 x dim2>
<matrix2> is of order <dim3 x dim4>
with the restriction <dim2 = dim3>;
note that <result> is of order <dim1 x dim4>.*)
procedure MatrixP(const dim1, dim2, dim4: byte; var result: rrarray; const matrix3, matrix4: rrarray); overload;
(*General matrix product: <result = matrix3 * matrix4>
where
<matrix3> is of order <dim1 x dim2>
<matrix4> is of order <dim3 x dim4>
with the restriction <dim2 = dim3>;
note that <result> is of order <dim1 x dim4>.*)

function VectorIProd(var dim: byte; var vector1: pvec): real; overload;
function VectorIProd(var dim: byte; var vector1, vector2: pvec): real; overload;
function VectorIProd(var dim: byte; var vector5, vector6: R_Ar): real; overload;
function VectorIProd(var ldim: integer; var vector5: R_Ar; var realvector: array of real): real; overload;
(*Evaluate vector in-product.*)

procedure VectorOProd(var dim: byte; var matrix1: pmatrix; var vector1: pvec); overload;
procedure VectorOProd(var dim: byte; var matrix1: pmatrix; var vector1, vector2: pvec); overload;
procedure VectorOProd(var dim: byte; var matrix3: rrarray; var vector1: pvec; var vector6: R_Ar); overload;
procedure VectorOProd(var dim: byte; var matrix3: rrarray; var vector5, vector6: R_Ar); overload;
procedure VectorOProd(var ldim: integer; var dim: byte; var matrix3: rrarray; var vector5, vector6: R_Ar); overload;
(*Evaluates the outer product <matrix1 = vector1 * vector2'>.*)

procedure VectorOSProd(var dim: byte; dim2: integer; var matrix3: rrarray; var vector1: pvec; var vector6: rrarray);
(*TS: copy of the third version of VectorOProd. summing over the
      second index of vector6 for index values 1..dim2 *)

procedure MatrixVecProd(const dim1, dim2: byte; var vector5: R_Ar; const matrix3: rrarray; const vector6: R_Ar); overload;
(*Matrix <matrix3> of dimension <dim1 x dim2> is postmultiplied by vector <vector6> of dimension <dim2>:
<vector5 = matrix3 x vector6>.
The resulting vector is of dimension <dim1>.*)
procedure MatrixVecProd(const dim: byte; var vector5: R_Ar; var matrix3: rrarray; var vector6: R_Ar); overload;
(*Matrix <matrix3> of dimension <dim x dim> is postmultiplied by vector <vector6> of dimension <dim>:
<vector5 = matrix3 x vector6>.*)
 
procedure Matrix_AdiagB_Prod(const dim: byte; var result: pmatrix; const matrix3, matrix4: rrarray); overload;
procedure Matrix_AdiagB_Prod(const dim: byte; var result: pmatrix; const matrix1: pmatrix; const matrix4: rrarray); overload;
procedure Matrix_AdiagB_Prod(const dim: byte; var result: rrarray; const matrix3, matrix4: rrarray); overload;
(*Computes <result = matrix3 * matrix4>
where <matrix3> is a NON-DIAGONAL matrix and <matrix4> a DIAGONAL matrix,
and both matrices are of order <dim x dim>.*)

procedure MatrixTProductI(const dim: byte; var matrix1, result: pmatrix);
(*Computes <matrix1 x transpose(matrix1)>, where <matrix1> is assumed to be square (not necessarily symmetric).*)

procedure MatrixTProductII(var dim: byte; var matrix1, result: pmatrix);
(*Computes <transpose(matrix1) x matrix1>, where <matrix1> is assumed to be square (not necessarily symmetric).*)

{
procedure MatrixTripleProd(var dim1: byte; var matrix1, matrix2: pmatrix; var matrix3: pmatrix);
(*Matrix product <matrix3 := matrix1 * matrix2 * matrix1> where the matrices have dimension <dim1 x dim1>.*)
}

function QuadraticForm(const dim: byte; const vector: R_Ar; const matrix: rrarray): real;
(*Evaluate quadratic form <transpose(vector) * matrix * vector>.*)

procedure KroneckerProd(var dim1, dim2, dim3, dim4: byte; var matrix3: rrarray; var matrix1, matrix2: pmatrix); overload;
(*Evaluate Kronecker product <matrix3 := matrix1 x matrix2>,
where
  <matrix1> is of dimension <dim1 x dim2>
and
  <matrix2> is of dimension <dim3 x dim4>.*)
procedure KroneckerProd(var dim1, dim2, dim3, dim4: byte; var matrix3: rrarray; var matrix4: rrarray; var matrix2: pmatrix); overload;
(*Evaluate Kronecker product <matrix3 := matrix4 x matrix2>,
where
  <matrix4> is of dimension <dim1 x dim2>
and
  <matrix2> is of dimension <dim3 x dim4>.*)
procedure KroneckerProd(var dim1, dim2, dim3, dim4: byte; var matrix3: rrarray; var matrix4: rrarray; var matrix5: rrarray); overload;
(*Evaluate Kronecker product <matrix3 := matrix4 x matrix5>,
where
  <matrix4> is of dimension <dim1 x dim2>
and
  <matrix5> is of dimension <dim3 x dim4>.*)

procedure HadamardProd(var dim1, dim2: byte; var result, matrix1, matrix2: pmatrix);
(*Evaluates Hadamard product <result := matrix1 x matrix2.>*)

procedure VecOperator(var dim: byte; var vector5: R_Ar; var matrix1: pmatrix); overload;
procedure VecOperator(var dim: byte; var vector5: R_Ar; var matrix3: rrarray); overload;
(*The <vec> operator transforms a given matrix into a vector by stacking the columns of the matrix one underneath the other.*)

procedure CommutationMatrix(const dim: byte; var permutation: rrarray);
(*Construct the commutation matrix (Magnus and Neudecker, 1988, pp. 46 - 48),
which is a (symmetric) permutation matrix that transforms <vec A> into <vec A'>.
See the paper on "Commutation matrix" in the random effects files.
The code presumes that matrix <A> is square, but this restriction can be removed.*)

procedure DuplicationMatrix(var dim: byte; var duplication: rrarray);
(*Construct the duplication matrix (Magnus and Neudecker, 1988, pp. 48 - 50),
which transforms <v(A)> into <vec(A)> by <D v(A) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
<A> is symmetric,
and <D> is the duplication matrix.
See the paper on "Duplication matrix" in the random effects files.*)

procedure EliminationMatrixI(const dim: byte; var elimination: rrarray);
(*Construct the elimination matrix <D> (Magnus, 1988, pp. 76 - 92), "Linear structures" satisfying <D v(A) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
<A> is square but not symmetric (otherwise: duplication matrix);
must be called once during run-time.*)

procedure EliminationMatrixII(const dim: byte; var elimination: rrarray);
(*Construct the elimination matrix <D> satisfying <D v(A) + vec(I) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
and <A> is square and subject to the following linear constraints:
- the above-diagonal elements are 0
- the elements on the main diagonal are 1;
<vec(I)> is the <vec> of the identity matrix <I>;
procedure must be called once during run-time.*)

procedure EliminationMatrixIII(const dim: byte; var elimination: rrarray);
(*Construct the elimination matrix <D> satisfying <D v(A) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
and <A> is square and diagonal;
procedure must be called once during run-time.*)

procedure MatrixSqrt(var dim: byte; var matrix1: pmatrix); overload;
procedure MatrixSqrt(var actors: integer; var dim: byte; var matrix3: rrarray); overload;
(*Coordinate-wise square-root of real valued matrix.*)

procedure MatrixSum(const dim1, dim2: byte; var result: pmatrix; const matrix1, matrix2: pmatrix);
(*Computes <result = matrix1 + matrix2.>
 where the matrices are of order <dim1 x dim2>.*)

procedure MatrixAdd(var dim1, dim2: byte; var matrix1, matrix2: pmatrix); overload;
(*Computes <matrix1 = matrix1 + matrix2>
 where the matrices are of order <dim1 x dim2>.*)
procedure MatrixAdd(var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = matrix3 + matrix4>
 where the matrices are of order <dim x dim>.*)
procedure MatrixAdd(var ldim: integer; var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = matrix3 + matrix4>
 where the matrices are of order <ldim x dim>.*)

procedure MatrixMinus(var dim1, dim2: byte; var result, matrix1, matrix2: pmatrix); overload;
(*Computes <result = matrix1 - matrix2>
where both <matrix1> and <matrix2> are of order <dim1 x dim2>.*)
procedure MatrixMinus(var dim1, dim2: byte; var matrix3, matrix4, matrix5: rrarray); overload;
(*Computes <matrix3 = matrix4 - matrix5>
where both <matrix4> and <matrix5> are of order <dim1 x dim2>.*)
procedure MatrixMinus(var ldim: integer; var dim: byte; var matrix3, matrix4, matrix5: rrarray); overload;
(*Computes <matrix3 = matrix4 - matrix5>
where both <matrix4> and <matrix5> are of order <ldim x dim>.*)

procedure VectorAdd(var dim: byte; var vector1, vector2: pvec); overload;
(*Add <vector1 := vector1 + vector2>.*)
procedure VectorAdd(var dim: byte; var vector5, vector6: R_Ar); overload;
procedure VectorAdd(var ldim: integer; var vector5, vector6: R_Ar); overload;
(*Add <vector5 := vector5 + vector6>.*)

procedure VectorMinus(var dim: byte; var vector1, vector2: pvec); overload;
(*Compute <vector1 := vector1 - vector2>.*)
procedure VectorMinus(var ldim: integer; var vector5, vector6: R_Ar); overload;
(*Compute <vector5 := vector5 - vector6>.*)

procedure MatrixT(const dim1, dim2: byte; var result, matrix1: pmatrix); overload;
(*Computes <result = transpose(matrix1)>, where <matrix1> is of order <dim1 x dim2>
and therefore <result> is or order <dim2 x dim1>.*)
procedure MatrixT(const dim: byte; var result: pmatrix; const matrix1: pmatrix); overload;
(*Computes <result = transpose(matrix1)>, where <matrix1> is of order <dim x dim>
and therefore <result> is or order <dim x dim>.*)
procedure MatrixT(const dim1, dim2: byte; var matrix3: rrarray; const matrix1: pmatrix); overload;
(*Computes <matrix3 = transpose(matrix1)>, where <matrix1> is of order <dim1 x dim2>
and therefore <matrix3> is of order <dim2 x dim1>.*)
procedure MatrixT(const dim1, dim2: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = transpose(matrix4)>, where <matrix4> is of order <dim1 x dim2>
and therefore <matrix3> is of order <dim2 x dim1>.*)
procedure MatrixT(const dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = transpose(matrix4)> where <matrix4> is of order <dim x dim>
and therefore <matrix4> is of order <dim x dim>.*)

procedure MakeDynamic(var dim: byte; var matrix3: rrarray; var matrix1: pmatrix); overload;
(*Transform static array into dynamic array.*)

procedure MakeStaticI(var dim: byte; var matrix1: pmatrix; var matrix3: rrarray); overload;
procedure MakeStaticII(const dim: byte; var matrix1: pmatrix; const matrix3: rrarray); overload;
(*Transform dynamic array into static array.*)

procedure EvaluateCMarginal(const dim: byte; const ldim: integer; var matrix3: rrarray);
(*Evaluate column marginals (column sums).*)

function MatrixTrace(var dim: byte; var matrix1: pmatrix): real;
(*Computes the trace of the square matrix <matrix1>*)

procedure MatrixI(var dim: byte; var determinant: real; var isigma, sigma, cholfac: pmatrix);
(*Invert variance-covariance matrix.*)

function CholeskyI(const dim: byte; const c: pmatrix; var ic: pmatrix): boolean;
(*Inversion of the Cholesky factor; exploits the structure of the Cholesky factor;
see: Finn (1974, p. 42), "A General Model for Multivariate Analysis".*)

procedure ScalarDiv(const dim: byte; var matrix1: pmatrix; const ldim: integer); overload;
(*Divide static array by scalar.*)
procedure ScalarDiv(const dim1: byte; var matrix1: pmatrix; const dim2: byte); overload;
procedure ScalarDiv(const dim1, dim2: byte; var matrix3: rrarray; const ldim2: integer); overload;
procedure ScalarDiv(const dim: byte; var matrix3: rrarray; const ldim2: integer); overload;
procedure ScalarDiv(const ldim1: integer; const dim: byte; var matrix3: rrarray; const ldim2: integer); overload;
procedure ScalarDiv(const ldim1, ldim2: integer; var matrix3: rrarray; const ldim: integer); overload;
procedure ScalarDiv(const dim: byte; var vector1: pvec; const ldim: integer); overload;
procedure ScalarDiv(const dim: byte; var vector5: R_Ar; const ldim: integer); overload;
procedure ScalarDiv(const ldim1: integer; var vector5: R_Ar; const ldim2: integer); overload;
(*Divide dynamic vector by scalar.*)

procedure ScalarProd(var dim: byte; var matrix1: pmatrix; var ldim: integer); overload;
procedure ScalarProd(var dim: byte; var result, matrix1: pmatrix; var ldim: integer); overload;
procedure ScalarProd(const dim: byte; var result: pmatrix; const matrix1: pmatrix; const real1: real); overload;
(*Multiply static array by scalar.*)
procedure ScalarProd(var ldim1: integer; var dim: byte; var matrix3: rrarray; var ldim2: integer); overload;
(*Multiply dynamic array by scalar.*)
procedure ScalarProd(const dim: byte; var matrix3: rrarray; const ldim: integer; const matrix4: rrarray); overload;
(*Multiply dynamic array by scalar: <matrix3 = ldim * matrix4>.*)
procedure ScalarProd(const dim1, dim2: byte; var matrix3: rrarray; const ldim: integer); overload;
(*Multiply dynamic array by scalar: <matrix3 = ldim * matrix3>.*)
procedure ScalarProd(const dim: byte; var matrix3: rrarray; const ldim: integer); overload;
(*Multiply dynamic array by scalar: <matrix3 = ldim * matrix3>.*)
procedure ScalarProd(var dim: byte; var vector1: pvec; var ldim: integer); overload;
(*Multiply static vector by scalar.*)

procedure MatrixVector(var dim: byte; var vector5: R_Ar; var matrix1: pmatrix);
(*Transform matrix <matrix1> into <vec(matrix1)>, where <matrix1> is assumed to be symmetric.*)

procedure VectorMatrix(var dim: byte; var matrix1: pmatrix; var vector5: R_Ar);
(*Invert the transformation of matrix <matrix1> transformed into <vec(matrix1)>, where <matrix1> is assumed to be symmetric.*)

procedure VectorDelete(var dim: byte; var exclusion: pBVec; var vector1, vector2: pvec);
(*Delete coordinate <dim1> of <vector1> of dimension <dim>
and store the result in the vector <vector1> of dimension <dim - 1>.*)

procedure MatrixDelete(var dim: byte; var exclusion: pBVec; var result, matrix1: pmatrix);
(*Delete row <dim1> and column <dim1> of matrix <matrix1> with dimension <dim x dim>.
On output, the original matrix is unchanged and the result is stored in the matrix
<result> of dimension <dim - 1 x dim - 1>.*)

procedure FilterRows(var matrix3: rrarray; var ldim: integer; var dim1, dim2: byte; var matrix4: rrarray); overload;
(*Write the rows of <matrix4> that are a multiple of <dim2> to <matrix3>.*)
procedure FilterRows(var matrix6: rrrarray; var ldim1, ldim2: integer; var dim1, dim2: byte; var matrix7: rrrarray); overload;
(*Write the rows of <matrix7> that are a multiple of <dim2> to <matrix6>.*)

procedure SetMatrix(const dim1, dim2: byte; var matrix1: pmatrix; const matrix2: pmatrix); overload;
procedure SetMatrix(var ldim1, ldim2: integer; var matrix1, matrix2: pmatrix); overload;
(*Set <matrix1 := matrix2>.*)
procedure SetMatrix(var dim: byte; var matrix1: pmatrix; var matrix3: rrarray); overload;
(*Set <matrix1 := matrix3>, where <matrix1> is a static array while <matrix3> is a dynamic array.*)
procedure SetMatrix(var dim: byte; var matrix3, matrix4: rrarray); overload;
procedure SetMatrix(const ldim: integer; const dim: byte; var matrix3: rrarray; const matrix4: rrarray); overload;
procedure SetMatrix(var ldim1, ldim2: integer; var matrix3, matrix4: rrarray); overload;
(*Set <matrix3 := matrix4>.*)

procedure SetDiagMatrix(const dim: byte; var matrix1: pmatrix; const real1: real); overload;
(*Creates diagonal matrix of order <dim x dim> with value <real1> on the main diagonal.*)

procedure SetIdentityMatrix(const dim: byte; var matrix1: pmatrix); overload;
procedure SetIdentityMatrix(var matrix3: rrarray; var dim: byte); overload;
(*Creates identity matrix of order <dim x dim>.*)

procedure SetMatrixValue(const dim: byte; var matrix1: pmatrix; const real1: real); overload;
procedure SetMatrixValue(var dim: byte; var matrix3: rrarray; var real1: real); overload;
procedure SetMatrixValue(var dim1, dim2: byte; var matrix3: rrarray; var real1: real); overload;
procedure SetMatrixValue(var ldim: integer; var dim: byte; var matrix3: rrarray; var real1: real); overload;
(*Sets matrix of dimension <dim x dim> with elements equal to <real1>.*)

procedure SetVectorValue(const dim: byte; var vector1: pvec; const real1: real);
(*Sets <matrix1> to the matrix of dimension <dim x dim> with elements equal to <real1>.*)

procedure SetVector(const dim: byte; var vector1: pvec; const vector2: pvec);
(*Set <vector1 := vector2>.*)

procedure MatrixStore(const ldim1: integer; const dim: byte; var autoRE: rrrarray; const ldim2: integer; const matrix3: rrarray); overload;
procedure MatrixStore(const dim: byte; var autoCovRE: rrrarray; const ldim: integer; const matrix1: pmatrix); overload;
(*Store matrix.*)

procedure VectorStore(const dim: byte; var autotheta: rrarray; const ldim: integer; const vector1: pvec); overload;
procedure VectorStore(const dim: byte; var autotheta: rrarray; const ldim: integer; const vector5: R_Ar); overload;
(*Store vector.*)

procedure Rescale(const dim: byte; var factor: real; var matrix1: pmatrix); overload;
(*Rescale matrix by factor <factor>.*)
procedure Rescale(const dim: byte; const factor: real; const matrix1: pmatrix; var matrix2: pmatrix); overload;
(*Rescale matrix by factor <factor> without changing the original matrix.*)

procedure MatrixNorm(var dim1: byte; var matrix1: pmatrix);
(*Makes the lengths of the column vectors of a <dim1 x dim1> matrix unity by dividing each column <c[i]> by <square root[c(i)'c(i}]>.*)

procedure CorrelationMatrix(const dim: byte; var matrix1: pmatrix);
(*Transforms below-diagonal elements of variance-covariance matrix into correlations.*)

procedure TransformCov(var dim: byte; var sigma: pmatrix);
(*The aim is to preserve the correlations in the given covariance matrix
but to change the scale. Observe that correlations are scale-invariant.*)

function CholeskyDecomposition(var matrix1: pmatrix; const dim: byte; var determinant: real): boolean;
(*Cholesky decomposition ("taking the square root") of symmetric, positive definite matrix,
<Sigma = C transpose(C)>, where <C> is the lower-triangular matrix solving the equation.
See random effects paper: latent attachements; Moonan (1957), JASA; Burden & Faires (1997, pp. 410-411), "Numerical Analysis", Algorithm 6.6; Devroye (1986, p. 564).*)

function Cholesky(const dim: byte; var cholfac: pmatrix; const sigma: pmatrix; var determinant: real): boolean;
(*Cholesky decomposition of positive definite variance-covariance-matrix.*)

procedure OpenFile(var ascii: textfile; var filename: shortstring);
(*Assign file name to textfile variable and rewrite file.*)

procedure OpenMFile(const boolean1: boolean; var filename: shortstring; var ascii: textfile);
(*If boolean condition is satisfied, prepare file for writing assignments.*)

procedure UpdateFrequencyDistribution(var real1, real2: real; var matrix4: rrarray);
(*Update frequency distribution by adding one data point.*)

procedure ReadFile(var ascii: textfile; const filename: shortstring); overload;
procedure ReadFile(var ascii: textfile; const filename: shortstring; var boolean1: boolean); overload;
(*Open and reset file.*)

procedure AppendToFile(var ascii: textfile; var filename: shortstring);
(*Assign file name to textfile variable and append to file.*)

procedure WriteReal(var ascii: textfile; var real1: real); overload;
procedure WriteReal(var ascii: textfile; const real1: extended); overload;
procedure WriteReal(var ascii: textfile; var real1, real2: real); overload;
procedure WriteReal(var boolean1: boolean; var ascii: textfile; var real1: real); overload;
procedure WriteReal(var boolean1: boolean; var ascii: textfile; var real1, real2: real); overload;
(*Writes real number to file to produce time series.*)

procedure MatrixWrite(var ascii: textfile; const dim1, dim2: byte; const matrix1: pmatrix); overload;
procedure MatrixWrite(var ascii: textfile; var dim: byte; var ldim: integer; var matrix3: rrarray); overload;
procedure MatrixWrite(var ldim: integer; var dim: byte; var ascii: textfile; var matrix3: rrarray); overload;
procedure MatrixWrite(var ascii: textfile; const dim1, dim2: byte; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
procedure MatrixWrite(const ldim: integer; const dim: byte; var ascii: textfile; const matrix3: rrarray; const ldim1, ldim2: integer); overload;
procedure MatrixWrite(var ascii: TFileStream; const dim1, dim2: byte; const matrix1: pmatrix); overload;
procedure MatrixWrite(var ascii: TFileStream; const dim1, dim2: byte; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
(*Writes real matrix of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)

procedure MatrixWriteLight(var ascii: textfile; const dim1, dim2: byte; const matrix1: pmatrix); overload;
procedure MatrixWriteLight(var ascii: TFileStream; const dim1, dim2: byte; const matrix1: pmatrix); overload;

procedure MatrixWritePlot(var boolean1: boolean; var ldim: integer; var dim: byte;  var ascii: textfile; var matrix3: rrarray); overload;
procedure MatrixWritePlot(var ldim: integer; var dim: byte;  var ascii: textfile; var matrix3: rrarray); overload;
(*Writes real matrix of dimension <dim1 x dim2> to ASCII file indexed by <ascii> on one single line.*)

procedure MatrixWriteCov(const dim: byte; var ascii: textfile; var matrix1: pmatrix); overload;
(*Write the vector of non-redundant elements of variance-covariance matrix <matrix1> to the file,
where the vector was obtained by applying the vec operator to <matrix1> and then deleting the above-diagonal elements.*)

procedure MatrixWriteTriangle(const dim: byte; var ascii: textfile; const matrix1: pmatrix); overload;
procedure MatrixWriteTriangle(const dim: byte; var ascii: textfile; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
procedure MatrixWriteTriangle(const dim: byte; var ascii: TFileStream; const matrix1: pmatrix); overload;
procedure MatrixWriteTriangle(const dim: byte; var ascii: TFileStream; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
(*Writes lower triangle (including diagonal) of symmetric matrix to file.*)

procedure VectorWrite(var ascii: textfile; var dim: byte; var vector1: pvec);
(*Writes estimates of fixed effects (parameters) to file to produce trace plot.*)

procedure MatrixRead(var ascii: textfile; var filename: shortstring; var matrix3: rrarray; var ldim: integer; var dim: byte);
(*Read matrix <matrix3> of dimension <ldim x dim> from input file <ascii> with name <filename>.*)

function MatrixURead(const filename: string; const dim: byte; var matrix3: rrarray): boolean;
(*Read matrix with an unknown number of rows and a known number of columns.*)

procedure Get_Cov_Vector(const dim: byte; var sigma: pmatrix; const vector1: pvec);
(*Extract the variance-covariance matrix <sigma> from the vector of non-redundant elements of <sigma>,
where the vector was obtained by applying the vec operator to <sigma> and then deleting the above-diagonal elements.*)

function NonSingular(var dim: byte; var sigma: pmatrix): boolean;
(*Test whether system is non-singular.*)

function CauchySchwartz(var dim: byte; var sigma: pmatrix): boolean;
(*Test whether the variance-covariance matrix satisfies the Cauchy-Schwartz inequality.*)

function Multiple(const ldim: integer; const dim: byte): boolean;
(*Tests whether <ldim> is a multiple of <dim>, that is, whether
<ldim = dim x i> holds for some integer <i>.*)

procedure SetArray(var vector6: R_Ar; const dim: byte); overload;
procedure SetArray(var vector6: R_Ar; const ldim: integer); overload;
procedure SetArray(var matrix3: rrarray; const dim: byte); overload;
procedure SetArray(var matrix3: rrarray; const dim1, dim2: byte); overload;
procedure SetArray(var matrix3: rrarray; const ldim: integer; const dim: byte); overload;
procedure SetArray(var matrix3: rrarray; const dim: byte; const ldim: integer); overload;
procedure SetArray(var matrix3: rrarray; const ldim1, ldim2: integer); overload;
procedure SetArray(var threedim: rrrarray; const dim: byte); overload;
procedure SetArray(var threedim: rrrarray; const ldim: integer; const dim: byte); overload;
procedure SetArray(var threedim: rrrarray; const dim: byte; const ldim: integer); overload;
procedure SetArray(var threedim: rrrarray; const ldim1: integer; const dim: byte; const ldim2: integer); overload;
procedure SetArray(var booleans: boarray; const ldim: integer); overload;
(*The first argument refers to the array that is to be initialised.
The other arguments refer to the first, second, and so on, dimension.
If only one (additional) argument is inputted when initialising an n > 1 dimensional array,
then it is assumed that the dimension have the same length.*)

procedure SetZero(var matrix: rrarray; const d1, d2: integer);
(* sets matrix[i,j] := 0.0 for 0<=1<=d1, 0<=j<=d2 *)

(*--------------------------------------------------------------------------------------------------*)

implementation

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixProd(var dim1, dim2: byte; var result, matrix1, matrix2: pmatrix); overload;
(*Computes <result = matrix1 * matrix2>
where
<matrix1> is of order <dim1 x dim2>
<matrix2> is of order <dim2 x dim2>.*)
var i, j, k: integer;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    result[i,j] := 0;
    for k := 1 to dim2 do result[i,j] := result[i,j] + (matrix1[i,k] * matrix2[k,j]);
    end;
  end;
end;

procedure MatrixProd(var product: rrarray; var ldim: integer; var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <product = matrix3 * matrix4>
where
<matrix3> is of order <ldim x dim>
<matrix4> is of order <dim x dim>.*)
var i, j, k: integer;
begin
SetArray(product,ldim,dim);
for i := 1 to ldim do
  begin
  for j := 1 to dim do
    begin
    product[i,j] := 0;
    for k := 1 to dim do product[i,j] := product[i,j] + (matrix3[i,k] * matrix4[k,j]);
    end;
  end;
end;

procedure MatrixProd(const dim1, dim2: byte; var product: rrarray; const matrix1: pmatrix; const matrix4: rrarray); overload;
(*Computes <product = matrix1 * matrix4>
where
<matrix1> is of order <dim1 x dim2>
<matrix4> is of order <dim2 x dim2>.*)
var i, j, k: integer;
begin
SetArray(product,dim1,dim2);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    for k := 1 to dim2 do product[i,j] := product[i,j] + (matrix1[i,k] * matrix4[k,j]);
    end;
  end;
end;

procedure MatrixProd(const dim1, dim2: byte; var product: rrarray; const matrix3: rrarray; const matrix1: pmatrix); overload;
(*Computes <product = matrix3 * matrix1>
where
<matrix3> is of order <dim1 x dim2>
<matrix1> is of order <dim2 x dim2>.*)
var i, j, k: integer;
begin
SetArray(product,dim1,dim2);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    for k := 1 to dim2 do product[i,j] := product[i,j] + (matrix3[i,k] * matrix1[k,j]);
    end;
  end;
end;

procedure MatrixProd(const dim1, dim2: byte; var product: rrarray; const matrix3, matrix4: rrarray); overload;
(*Computes <product = matrix3 * matrix4>
where
<matrix3> is of order <dim1 x dim2>
<matrix4> is of order <dim2 x dim2>.*)
var i, j, k: integer;
begin
SetArray(product,dim1,dim2);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    for k := 1 to dim2 do product[i,j] := product[i,j] + (matrix3[i,k] * matrix4[k,j]);
    end;
  end;
end;

procedure MatrixProd(var product: rrarray; var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <product = matrix3 * matrix4>
where
<matrix3> is of order <dim x dim>
<matrix4> is of order <dim x dim>.*)
var i, j, k: integer;
begin
SetArray(product,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    product[i,j] := 0;
    for k := 1 to dim do product[i,j] := product[i,j] + (matrix3[i,k] * matrix4[k,j]);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixP(const dim1, dim2, dim4: byte; var result: pmatrix; const matrix1, matrix2: pmatrix); overload;
(*General matrix product: <result = matrix1 * matrix2>
where
<matrix1> is of order <dim1 x dim2>
<matrix2> is of order <dim3 x dim4>
with the restriction <dim2 = dim3>;
note that <result> is of order <dim1 x dim4>.*)
var i, j, k: integer;
begin
for i := 1 to dim1 do (*For each row of <matrix1>...*)
  begin
  for j := 1 to dim4 do (*...and each column of <matrix2>: evaluate the inner product.*)
    begin
    result[i,j] := 0;
    for k := 1 to dim2 do result[i,j] := result[i,j] + (matrix1[i,k] * matrix2[k,j]); (*The vector inner product corresponds to <dim2> scalar products.*)
    end;
  end;
end;

procedure MatrixP(const dim1, dim2, dim4: byte; var result: rrarray; const matrix3, matrix4: rrarray); overload;
(*General matrix product: <result = matrix3 * matrix4>
where
<matrix3> is of order <dim1 x dim2>
<matrix4> is of order <dim3 x dim4>
with the restriction <dim2 = dim3>;
note that <result> is of order <dim1 x dim4>.*)
var i, j, k: integer;
begin
SetArray(result,dim1,dim4);
for i := 1 to dim1 do (*For each row of <matrix3>...*)
  begin
  for j := 1 to dim4 do (*...and each column of <matrix4>: evaluate the inner product.*)
    begin
    for k := 1 to dim2 do result[i,j] := result[i,j] + (matrix3[i,k] * matrix4[k,j]); (*The vector inner product corresponds to <dim2> scalar products.*)
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function VectorIProd(var dim: byte; var vector1: pvec): real; overload;
(*Evaluate vector in-product.*)
var
  i: byte;
  sum: real;
begin
sum := 0;
for i := 1 to dim do sum := sum + sqr(vector1[i]);
VectorIProd := sum;
end;

function VectorIProd(var dim: byte; var vector1, vector2: pvec): real; overload;
(*Evaluate vector in-product.*)
var
  i: byte;
  sum: real;
begin
sum := 0;
for i := 1 to dim do sum := sum + (vector1[i] * vector2[i]);
VectorIProd := sum;
end;

function VectorIProd(var dim: byte; var vector5, vector6: R_Ar): real; overload;
(*Evaluate vector in-product.*)
var
  i: byte;
  sum: real;
begin
sum := 0;
for i := 1 to dim do sum := sum + (vector5[i] * vector6[i]);
VectorIProd := sum;
end;

function VectorIProd(var ldim: integer; var vector5: R_Ar; var realvector: array of real): real; overload;
(*Evaluate vector in-product.*)
var
  i: integer;
  sum: real;
begin
sum := 0;
for i := 1 to ldim do sum := sum + (vector5[i] * realvector[i]);
VectorIProd := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorOProd(var dim: byte; var matrix1: pmatrix; var vector1: pvec); overload;
(*Evaluates the outer product of some vector <vector1>, that is, <matrix1 = vector1 * vector1'>.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix1[i,j] := vector1[i] * vector1[j];
  end;
end;

procedure VectorOProd(var dim: byte; var matrix1: pmatrix; var vector1, vector2: pvec); overload;
(*Evaluates the outer product <matrix1 = vector1 * vector2'>.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix1[i,j] := vector1[i] * vector2[j];
  end;
end;

procedure VectorOProd(var dim: byte; var matrix3: rrarray; var vector1: pvec; var vector6: R_Ar); overload;
(*Evaluates the outer product <matrix1 = vector1 * vector2'>.*)
var i, j: byte;
begin
SetArray(matrix3,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix3[i,j] := vector1[i] * vector6[j];
  end;
end;

procedure VectorOProd(var dim: byte; var matrix3: rrarray; var vector5, vector6: R_Ar); overload;
(*Evaluates the outer product <matrix1 = vector1 * vector2'>.*)
var i, j: byte;
begin
SetArray(matrix3,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix3[i,j] := vector5[i] * vector6[j];
  end;
end;

procedure VectorOProd(var ldim: integer; var dim: byte; var matrix3: rrarray; var vector5, vector6: R_Ar); overload;
(*Evaluates the outer product <matrix1 = vector1 * vector2'>.*)
var i, j: integer;
begin
SetArray(matrix3,ldim,dim);
for i := 1 to ldim do
  begin
  for j := 1 to dim do matrix3[i,j] := vector5[i] * vector6[j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorOSProd(var dim: byte; dim2: integer; var matrix3: rrarray; var vector1: pvec; var vector6: rrarray);
(*TS: copy of the third version of VectorOProd. summing over the
      second index of vector6 for index values 1..dim2 *)
(*Evaluates the outer product <matrix1 = vector1 * vector6' * E>,
  where E is the dim2 by dim2 matrix with all elements equal to 1.*)
var i, j, h: byte;
    s: real;
begin
SetArray(matrix3,dim);
for i := 1 to dim do for j := 1 to dim do
  begin
  s := 0.0;
  for h:=dim2 downto 1 do s := s + vector1[i] * vector6[j, h];
  matrix3[i,j] := s;
  end;
end;


(*--------------------------------------------------------------------------------------------------*)

procedure MatrixVecProd(const dim1, dim2: byte; var vector5: R_Ar; const matrix3: rrarray; const vector6: R_Ar); overload;
(*Matrix <matrix3> of dimension <dim1 x dim2> is postmultiplied by vector <vector6> of dimension <dim2>:
<vector5 = matrix3 x vector6>.
The resulting vector is of dimension <dim1>.*)
var i, j: byte;
begin
SetArray(vector5,dim1); (*The matrix to be postmultiplied is of order <dim1 x dim2>, so the resulting vector is of dimension <dim1>, the number of rows.*)
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    vector5[i] := vector5[i] + (matrix3[i,j] * vector6[j]);
    end;
  end;
end;

procedure MatrixVecProd(const dim: byte; var vector5: R_Ar; var matrix3: rrarray; var vector6: R_Ar); overload;
(*Matrix <matrix3> of dimension <dim x dim> is postmultiplied by vector <vector6> of dimension <dim>:
<vector5 = matrix3 x vector6>.*)
var i, j: byte;
begin
SetArray(vector5,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    vector5[i] := vector5[i] + (matrix3[i,j] * vector6[j]);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Matrix_AdiagB_Prod(const dim: byte; var result: pmatrix; const matrix3, matrix4: rrarray); overload;
(*Computes <result = matrix3 * matrix4>
where <matrix3> is a NON-DIAGONAL matrix and <matrix4> a DIAGONAL matrix,
and both matrices are of order <dim x dim>.*)
var i, j: integer;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    result[i,j] := matrix3[i,j] * matrix4[j,j];
    end;
  end;
end;

procedure Matrix_AdiagB_Prod(const dim: byte; var result: pmatrix; const matrix1: pmatrix; const matrix4: rrarray); overload;
(*Computes <result = matrix1 * matrix4>
where <matrix1> is a NON-DIAGONAL matrix and <matrix4> a DIAGONAL matrix,
and both matrices are of order <dim x dim>.*)
var i, j: integer;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    result[i,j] := matrix1[i,j] * matrix4[j,j];
    end;
  end;
end;

procedure Matrix_AdiagB_Prod(const dim: byte; var result: rrarray; const matrix3, matrix4: rrarray); overload;
(*Computes <result = matrix3 * matrix4>
where <matrix3> is a NON-DIAGONAL matrix and <matrix4> a DIAGONAL matrix,
and both matrices are of order <dim x dim>.*)
var i, j: integer;
begin
SetArray(result,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    result[i,j] := matrix3[i,j] * matrix4[j,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixTProductI(const dim: byte; var matrix1, result: pmatrix);
(*Computes <matrix1 x transpose(matrix1)>, where <matrix1> is assumed to be square (not necessarily symmetric).*)
var
  i, j, h: byte;
  sum: real;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do (*Compute element (i,j).*)
    begin
    sum := 0;
    for h := 1 to dim do sum := sum + (matrix1[i,h] * matrix1[j,h]);
    result[i,j] := sum;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixTProductII(var dim: byte; var matrix1, result: pmatrix);
(*Computes <transpose(matrix1) x matrix1>, where <matrix1> is assumed to be square (not necessarily symmetric).*)
var
  i, j, h: byte;
  sum: real;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do (*Compute element (i,j).*)
    begin
    sum := 0;
    (*Note: if <B = A'>, then element (i,j) of <A'A> is <sum(h) b(i,h) a(h,j) = sum(h) a(h,i) a(h,j)>.*)
    for h := 1 to dim do sum := sum + (matrix1[h,i] * matrix1[h,j]);
    result[i,j] := sum;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

{
procedure MatrixTripleProd(var dim1: byte; var matrix1, matrix2: pmatrix; var matrix3: pmatrix);
(*Matrix product <matrix3 := matrix1 * matrix2 * matrix1> where the matrices have dimension <dim1 x dim1>.*)
var i,j,h,k: byte;
temp: real;
begin
for i:=dim1 downto 1 do
  begin
  for j:=i downto 1 do
    begin
    temp := 0.0;
    for h:=dim1 downto 1 do
      begin
      for k:=dim1 downto 1 do
        begin
        temp := temp + (matrix1[i,h]* matrix2[h,k] * matrix1[j,k]);
        matrix3[i,j] := temp;
        if (j < i) then matrix3[j,i] := temp;
        end;
      end;
    end;
  end;
end;
}

(*--------------------------------------------------------------------------------------------------*)

function QuadraticForm(const dim: byte; const vector: R_Ar; const matrix: rrarray): real;
(*Evaluate quadratic form <transpose(vector) * matrix * vector>.*)
var
  i, j: byte;
  sum: real;
begin
sum := 0;
for i := 1 to dim do
  begin
  for j := 1 to dim do sum := sum + (vector[i] * matrix[i,j] * vector[j]);
  end;
QuadraticForm := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure KroneckerProd(var dim1, dim2, dim3, dim4: byte; var matrix3: rrarray; var matrix1, matrix2: pmatrix); overload;
(*Evaluate Kronecker product <matrix3 := matrix1 x matrix2>,
where
  <matrix1> is of dimension <dim1 x dim2>
and
  <matrix2> is of dimension <dim3 x dim4>.*)
var i, i1, i2, j, j1, j2, dim_1, dim_2: byte;
begin
dim_1 := dim1 * dim3; (*Vertical dimension.*)
dim_2 := dim2 * dim4; (*Horizontal dimension.*)
SetArray(matrix3,dim_1,dim_2);
i := 0;
for i1 := 1 to dim1 do (*Row of <matrix1>.*)
  begin
  for i2 := 1 to dim3 do (*Row of <matrix2>.*)
    begin
    Inc(i); (*Row of the product matrix <matrix3>.*)
    j := 0; (*Must be reset for each row.*)
    for j1 := 1 to dim2 do (*Column of <matrix1>.*)
      begin
      for j2 := 1 to dim4 do (*Column of <matrix2>.*)
        begin
        Inc(j); (*Column of the product matrix <matrix3>.*)
        matrix3[i,j] := matrix1[i1,j1] * matrix2[i2,j2];
        end;
      end;
    end;
  end;
end;

procedure KroneckerProd(var dim1, dim2, dim3, dim4: byte; var matrix3: rrarray; var matrix4: rrarray; var matrix2: pmatrix); overload;
(*Evaluate Kronecker product <matrix3 := matrix4 x matrix2>,
where
  <matrix4> is of dimension <dim1 x dim2>
and
  <matrix2> is of dimension <dim3 x dim4>.*)
var i, i1, i2, j, j1, j2, dim_1, dim_2: byte;
begin
dim_1 := dim1 * dim3; (*Vertical dimension.*)
dim_2 := dim2 * dim4; (*Horizontal dimension.*)
SetArray(matrix3,dim_1,dim_2);
i := 0;
for i1 := 1 to dim1 do (*Row of <matrix4>.*)
  begin
  for i2 := 1 to dim3 do (*Row of <matrix2>.*)
    begin
    Inc(i); (*Row of the product matrix <matrix3>.*)
    j := 0; (*Must be reset for each row.*)
    for j1 := 1 to dim2 do (*Column of <matrix4>.*)
      begin
      for j2 := 1 to dim4 do (*Column of <matrix2>.*)
        begin
        Inc(j); (*Column of the product matrix <matrix3>.*)
        matrix3[i,j] := matrix4[i1,j1] * matrix2[i2,j2];
        end;
      end;
    end;
  end;
end;

procedure KroneckerProd(var dim1, dim2, dim3, dim4: byte; var matrix3: rrarray; var matrix4: rrarray; var matrix5: rrarray); overload;
(*Evaluate Kronecker product <matrix3 := matrix4 x matrix5>,
where
  <matrix4> is of dimension <dim1 x dim2>
and
  <matrix5> is of dimension <dim3 x dim4>.*)
var i, i1, i2, j, j1, j2, dim_1, dim_2: byte;
begin
dim_1 := dim1 * dim3; (*Vertical dimension.*)
dim_2 := dim2 * dim4; (*Horizontal dimension.*)
SetArray(matrix3,dim_1,dim_2);
i := 0;
for i1 := 1 to dim1 do (*Row of <matrix4>.*)
  begin
  for i2 := 1 to dim3 do (*Row of <matrix5>.*)
    begin
    Inc(i); (*Row of the product matrix <matrix3>.*)
    j := 0; (*Must be reset for each row.*)
    for j1 := 1 to dim2 do (*Column of <matrix4>.*)
      begin
      for j2 := 1 to dim4 do (*Column of <matrix5>.*)
        begin
        Inc(j); (*Column of the product matrix <matrix3>.*)
        matrix3[i,j] := matrix4[i1,j1] * matrix5[i2,j2];
        end;
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure HadamardProd(var dim1, dim2: byte; var result, matrix1, matrix2: pmatrix);
(*Evaluates Hadamard product <result := matrix1 x matrix2.>*)
var i, j: byte;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do result[i,j] := matrix1[i,j] * matrix2[i,j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VecOperator(var dim: byte; var vector5: R_Ar; var matrix1: pmatrix); overload;
(*The <vec> operator transforms a given matrix into a vector by stacking the columns of the matrix one underneath the other.*)
var i, j, count: integer;
begin
i := sqr(dim); (*Length of vector = number of elements of the matrix.*)
SetArray(vector5,i);
count := 0;
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    Inc(count);
    vector5[count] := matrix1[j,i];
    end;
  end;
end;

procedure VecOperator(var dim: byte; var vector5: R_Ar; var matrix3: rrarray); overload;
(*The <vec> operator transforms a given matrix into a vector by stacking the columns of the matrix one underneath the other.*)
var i, j, count: integer;
begin
i := sqr(dim); (*Length of vector = number of elements of the matrix.*)
SetArray(vector5,i);
count := 0;
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    Inc(count);
    vector5[count] := matrix3[j,i];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CommutationMatrix(const dim: byte; var permutation: rrarray);
(*Construct the commutation matrix (Magnus and Neudecker, 1988, pp. 46 - 48),
which is a symmetric permutation matrix that transforms <vec A> into <vec A'>.
See the paper on "Commutation matrix" in the random effects files.
The code presumes that matrix <A> is square, but this restriction can be removed.*)
var present, target, dimension, i, j: byte;
begin
dimension := sqr(dim);
SetArray(permutation,dimension);
target := 0;
for i := 1 to dim do (*Row of <A>.*)
  begin
  for j := 1 to dim do (*Column of <A>*)
    begin
    (*Remark: for each element <a[i,j]>, determine its <present> and its <target> position,
    and construct the permutation matrix such that it moves <a[i,j]> to its target position:*)
    Inc(target); (*Position in the target vector <vec A'>, obtained by stacking the rows of <A> (not <A'>) one underneath the other.*)
    present := ((j - 1) * dim) + i;
    (*Position in the present vector <vec A>:
    <j - 1> columns preceed the <j>-th column, hence <(j - 1) * dim> elements;
    further, element <a[i,j]> is, within the <j>-th column, on position <i>,
    hence its position in <vec A> is <(j - 1) * dim + i>.*)
    permutation[target,present] := 1; (*Take <a[i,j]> (which column <present> does) and transfer it to position <target> (which row <target> does).*)
    (*As the nil operator was applied to <permutation>, the matrix elements are by default zero,
    and no further (row) elements of <permutation> need to be manipulated.*)
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure DuplicationMatrix(var dim: byte; var duplication: rrarray);
(*Construct the duplication matrix (Magnus and Neudecker, 1988, pp. 48 - 50),
which transforms <v(A)> into <vec(A)> by <D v(A) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
<A> is symmetric,
and <D> is the duplication matrix.
See the paper on "Duplication matrix" in the random effects files.*)
var i, ii, j, jj, h, column, count, dim_1, dim_2, vec_position, redundant, row: byte;
begin
dim_1 := sqr(dim);
dim_2 := (dim * (dim + 1)) div 2;
SetArray(duplication,dim_1,dim_2);
row := 0;
count := 0;
(*<vec(A)> is obtained by stacking the columns of <A> one underneath the other.*)
for j := 1 to dim do (*Column <j>.*)
  begin
  for i := 1 to dim do (*Within column <j>: element <i>.*)
    begin
    (*For each element <a[i,j]>:*)
    Inc(row); (*Row number of <duplication> matrix: present row <(j - 1) * dim + i> is responsible for getting <a[i,j]> into <vec(A)>.*)
    if (j > i) then (*Above-diagonal element: not contained in <v(A)>; take <a[j,i]> instead:*)
      begin
      (*The present row is responsible for picking the above-diagonal element <a[i,j]> from <v(A)>,
      and transferring it to <vec(A)>. Since the above diagonal element <a[i,j]> is not contained in <v(A)>,
      but <a[j,i]> is, we pick <a[j,i]> from <v(A)> in position <(i - 1) * dim + j - #redudandant elements> in <vec(A)> preceding <a[j,i]>.*)
      Inc(count); (*Count the number of redundant elements of <vec(A)> up to and including <a[j,i]>.*)
      (*Change notation: <a[j,i]> is in the <j>-th row and <i>-th column: let <a[ii,jj] = a[j,i]>.*)
      ii := j;
      jj := i;
      vec_position := ((jj - 1) *  dim) + ii; (*Column <jj> is preceded in <vec(A)> by <jj - 1> columns, hence <(jj - 1) * dim> elements, thus the position is <(jj - 1) * dim + ii>.*)
      redundant := 0;
      for h := 1 to jj do redundant := redundant + (h - 1); (*Number of redundant elements up to and including column <jj>.*)
      column := vec_position - redundant; (*Position of <a[ii,jj] = a[j,i]> in <v(A)>.*)
      end
    else  (*Below-diagonal / diagonal element: contained in <v(A)>:*)
      begin
      vec_position := ((j - 1) * dim) + i; (*Column <j> is preceded in <vec(A)> by <j - 1> columns, hence <(j - 1) * dim> elements, thus the position is <(j - 1) * dim + i>.*)
      column := vec_position - count; (*Position of <a[i,j]> in <v(A)>.*)
      end;
    duplication[row,column] := 1;
    (*As the nil operator was applied to <duplication>, the matrix elements are by default zero,
    and no further (row) elements of <duplication> need to be manipulated.*)
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EliminationMatrixI(const dim: byte; var elimination: rrarray);
(*Construct the elimination matrix <D> (Magnus, 1988, pp. 76 - 92), "Linear structures" satisfying <D v(A) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
<A> is square but not symmetric (otherwise: duplication matrix);
must be called once during run-time.*)
var i, j, column, count, dim_1, dim_2, vec_position, row: byte;
begin
dim_1 := sqr(dim); (*The dimension of <vec(A)>.*)
dim_2 := (dim * (dim + 1)) div 2; (*The total number of below-diagonal plus diagonal elements: the dimension of <v(A)>.*)
SetArray(elimination,dim_1,dim_2);
row := 0;
count := 0;
(*<vec(A)> is obtained by stacking the columns of <A> one underneath the other.*)
for j := 1 to dim do (*Column <j>.*)
  begin
  for i := 1 to dim do (*Within column <j>: element <i>.*)
    begin
    (*For each element <a[i,j]>:*)
    Inc(row); (*Row number of <elimination> matrix: present row <(j - 1) * dim + i> is responsible for getting <a[i,j]> into <vec(A)>.*)
    if (j > i) then (*Above-diagonal element.*)
      begin
      Inc(count); (*Do nothing else: the present row is by default null, and hence deletes the above-diagonal element of <A>.*)
      end
    else  (*Below-diagonal / diagonal element: contained in <v(A)>:*)
      begin
      vec_position := ((j - 1) * dim) + i; (*Column <j> is preceded in <vec(A)> by <j - 1> columns, hence <(j - 1) * dim> elements, thus the position is <(j - 1) * dim + i>.*)
      column := vec_position - count; (*Position of <a[i,j]> in <v(A)>; <count> is the number of redundant elements up to date.*)
      elimination[row,column] := 1;
      (*As the nil operator was applied to <elimination>, the matrix elements are by default zero,
      and no further (row) elements of <elimination> need to be manipulated.*)
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EliminationMatrixII(const dim: byte; var elimination: rrarray);
(*Construct the elimination matrix <D> satisfying <D v(A) + vec(I) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
and <A> is square and subject to the following linear constraints:
- the above-diagonal elements are 0
- the elements on the main diagonal are 1;
<vec(I)> is the <vec> of the identity matrix <I>;
procedure must be called once during run-time.*)
var i, j, column, count, dim_1, dim_2, vec_position, row: byte;
begin
dim_1 := sqr(dim); (*The dimension of <vec(A)>.*)
dim_2 := (dim * (dim - 1)) div 2; (*The total number of below-diagonal elements: the dimension of <v(A)>.*)
SetArray(elimination,dim_1,dim_2);
row := 0;
count := 0;
(*<vec(A)> is obtained by stacking the columns of <A> one underneath the other.*)
for j := 1 to dim do (*Column <j>.*)
  begin
  for i := 1 to dim do (*Within column <j>: element <i>.*)
    begin
    (*For each element <a[i,j]>:*)
    Inc(row); (*Row number of <elimination> matrix: present row <(j - 1) * dim + i> is responsible for getting <a[i,j]> into <vec(A)>.*)
    if (j >= i) then (*Above-diagonal or diagonal element.*)
      begin
      Inc(count); (*Do nothing else: the present row is by default null, and hence makes the element on the main diagonal or above it null.*)
      end
    else  (*Below-diagonal element: contained in <v(A)>:*)
      begin
      vec_position := ((j - 1) * dim) + i; (*Column <j> is preceded in <vec(A)> by <j - 1> columns, hence <(j - 1) * dim> elements, thus the position is <(j - 1) * dim + i>.*)
      column := vec_position - count; (*Position of <a[i,j]> in <v(A)>; <count> is the number of redundant elements of <A> up to date.*)
      elimination[row,column] := 1;
      (*As the nil operator was applied to <elimination>, the matrix elements are by default zero,
      and no further (row) elements of <elimination> need to be manipulated.*)
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EliminationMatrixIII(const dim: byte; var elimination: rrarray);
(*Construct the elimination matrix <D> satisfying <D v(A) = vec(A)>,
where <v(A)> is the vector of non-redundant elements of <A>,
and <A> is square and diagonal;
procedure must be called once during run-time.*)
var i, j, column, count, dim_1, dim_2, vec_position, row: byte;
begin
dim_1 := sqr(dim); (*The dimension of <vec(A)>.*)
dim_2 := dim; (*The total number of diagonal elements: the dimension of <v(A)>.*)
SetArray(elimination,dim_1,dim_2);
row := 0;
count := 0;
(*<vec(A)> is obtained by stacking the columns of <A> one underneath the other.*)
for j := 1 to dim do (*Column <j>.*)
  begin
  for i := 1 to dim do (*Within column <j>: element <i>.*)
    begin
    (*For each element <a[i,j]>:*)
    Inc(row); (*Row number of <elimination> matrix: present row <(j - 1) * dim + i> is responsible for getting <a[i,j]> into <vec(A)>.*)
    if (j <> i) then (*Non-diagonal element.*)
      begin
      Inc(count); (*Do nothing else: the present row is by default null, and hence deletes the above-diagonal element of <A>.*)
      end
    else  (*Diagonal element: contained in <v(A)>:*)
      begin
      vec_position := ((j - 1) * dim) + i; (*Column <j> is preceded in <vec(A)> by <j - 1> columns, hence <(j - 1) * dim> elements, thus the position is <(j - 1) * dim + i>.*)
      column := vec_position - count; (*Position of <a[i,j]> in <v(A)>; <count> is the number of redundant elements of <A> up to date.*)
      elimination[row,column] := 1;
      (*As the nil operator was applied to <elimination>, the matrix elements are by default zero,
      and no further (row) elements of <elimination> need to be manipulated.*)
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixSqrt(var dim: byte; var matrix1: pmatrix); overload;
(*Coordinate-wise square-root of symmetric matrix.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix1[i,j] := sqrt(matrix1[i,j]);
  end;
end;

procedure MatrixSqrt(var actors: integer; var dim: byte; var matrix3: rrarray); overload;
(*Coordinate-wise square-root of symmetric matrix.*)
var i, j: integer;
begin
for i := 1 to actors do
  begin
  for j := 1 to dim do matrix3[i,j] := sqrt(matrix3[i,j]);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixSum(const dim1, dim2: byte; var result: pmatrix; const matrix1, matrix2: pmatrix);
(*Computes <result = matrix1 + matrix2.>
 where the matrices are of order <dim1 x dim2>.*)
var i, h: byte;
begin
for i := 1 to dim1 do
  begin
  for h := 1 to dim2 do result[i,h] := matrix1[i,h] + matrix2[i,h];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixAdd(var dim1, dim2: byte; var matrix1, matrix2: pmatrix); overload;
(*Computes <matrix1 = matrix1 + matrix2>
 where the matrices are of order <dim1 x dim2>.*)
var i, h: byte;
begin
for i := 1 to dim1 do
  begin
  for h := 1 to dim2 do matrix1[i,h] := matrix1[i,h] + matrix2[i,h];
  end;
end;

procedure MatrixAdd(var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = matrix3 + matrix4>
 where the matrices are of order <dim x dim>.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do matrix3[i,h] := matrix3[i,h] + matrix4[i,h];
  end;
end;

procedure MatrixAdd(var ldim: integer; var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = matrix3 + matrix4>
 where the matrices are of order <ldim x dim>.*)
var
  i: integer;
  h: byte;
begin
for i := 1 to ldim do
  begin
  for h := 1 to dim do matrix3[i,h] := matrix3[i,h] + matrix4[i,h];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixMinus(var dim1, dim2: byte; var result, matrix1, matrix2: pmatrix); overload;
(*Computes <result = matrix1 - matrix2>
where both <matrix1> and <matrix2> are of order <dim1 x dim2>.*)
var i, j: integer;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    result[i,j] := matrix1[i,j] - matrix2[i,j];
    end;
  end;
end;

procedure MatrixMinus(var dim1, dim2: byte; var matrix3, matrix4, matrix5: rrarray); overload;
(*Computes <matrix3 = matrix4 - matrix5>
where both <matrix4> and <matrix5> are of order <dim1 x dim2>.*)
var i, j: integer;
begin
SetArray(matrix3,dim1,dim2);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    matrix3[i,j] := matrix4[i,j] - matrix5[i,j];
    end;
  end;
end;

procedure MatrixMinus(var ldim: integer; var dim: byte; var matrix3, matrix4, matrix5: rrarray); overload;
(*Computes <matrix3 = matrix4 - matrix5>
where both <matrix4> and <matrix5> are of order <ldim x dim>.*)
var i, j: integer;
begin
SetArray(matrix3,ldim,dim);
for i := 1 to ldim do
  begin
  for j := 1 to dim do
    begin
    matrix3[i,j] := matrix4[i,j] - matrix5[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorAdd(var dim: byte; var vector1, vector2: pvec); overload;
(*Add <vector1 := vector1 + vector2>.*)
var i: byte;
begin
for i := 1 to dim do vector1[i] := vector1[i] + vector2[i];
end;

procedure VectorAdd(var dim: byte; var vector5, vector6: R_Ar); overload;
(*Add <vector5 := vector5 + vector6>.*)
var i: byte;
begin
for i := 1 to dim do vector5[i] := vector5[i] + vector6[i];
end;

procedure VectorAdd(var ldim: integer; var vector5, vector6: R_Ar); overload;
(*Add <vector5 := vector5 + vector6>.*)
var i: integer;
begin
for i := 1 to ldim do vector5[i] := vector5[i] + vector6[i];
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorMinus(var dim: byte; var vector1, vector2: pvec); overload;
(*Compute <vector1 := vector1 - vector2>.*)
var i: byte;
begin
for i := 1 to dim do vector1[i] := vector1[i] - vector2[i];
end;

procedure VectorMinus(var ldim: integer; var vector5, vector6: R_Ar); overload;
(*Compute <vector5 := vector5 - vector6>.*)
var i: integer;
begin
for i := 1 to ldim do vector5[i] := vector5[i] - vector6[i];
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixT(const dim1, dim2: byte; var result, matrix1: pmatrix); overload;
(*Computes <result = transpose(matrix1)>, where <matrix1> is of order <dim1 x dim2>
and therefore <result> is of order <dim2 x dim1>.*)
var i, j, dimension: integer;
begin
if (dim1 > dim2) then dimension := dim1
else dimension := dim2;
for i := 1 to dimension do
  begin
  for j := 1 to dimension do result[i,j] := 0; (*Reset all elements: safety measure.*)
  end;
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do result[j,i] := matrix1[i,j]; (*Reverse the order of subscripts in order to make rows columns.*)
  end;
end;

procedure MatrixT(const dim: byte; var result: pmatrix; const matrix1: pmatrix); overload;
(*Computes <result = transpose(matrix1)>, where <matrix1> is of order <dim x dim>
and therefore <result> is of order <dim x dim>.*)
var i, j: integer;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do result[j,i] := matrix1[i,j]; (*Reverse the order of subscripts in order to make rows columns.*)
  end;
end;

procedure MatrixT(const dim1, dim2: byte; var matrix3: rrarray; const matrix1: pmatrix); overload;
(*Computes <matrix3 = transpose(matrix1)>, where <matrix1> is of order <dim1 x dim2>
and therefore <matrix3> is of order <dim2 x dim1>.*)
var i, j: integer;
begin
SetArray(matrix3,dim2,dim1);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do matrix3[j,i] := matrix1[i,j]; (*Reverse the order of subscripts in order to make rows columns.*)
  end;
end;

procedure MatrixT(const dim1, dim2: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = transpose(matrix4)>, where <matrix4> is of order <dim1 x dim2>
and therefore <matrix3> is of order <dim2 x dim1>.*)
var i, j: integer;
begin
SetArray(matrix3,dim2,dim1);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do matrix3[j,i] := matrix4[i,j]; (*Reverse the order of subscripts in order to make rows columns.*)
  end;
end;

procedure MatrixT(const dim: byte; var matrix3, matrix4: rrarray); overload;
(*Computes <matrix3 = transpose(matrix4)> where <matrix4> is of order <dim x dim>
and therefore <matrix4> is of order <dim x dim>.*)
var i, j: integer;
begin
SetArray(matrix3,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix3[j,i] := matrix4[i,j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MakeDynamic(var dim: byte; var matrix3: rrarray; var matrix1: pmatrix); overload;
(*Transform static array into dynamic array.*)
var i, j: integer;
begin
SetArray(matrix3,dim);
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    matrix3[i,j] := matrix1[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MakeStaticI(var dim: byte; var matrix1: pmatrix; var matrix3: rrarray); overload;
(*Transform dynamic array into static array.*)
var i, j: integer;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    matrix1[i,j] := matrix3[i,j];
    end;
  end;
end;

procedure MakeStaticII(const dim: byte; var matrix1: pmatrix; const matrix3: rrarray); overload;
(*Transform dynamic array into static array.*)
var i, j: integer;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    matrix1[i,j] := matrix3[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EvaluateCMarginal(const dim: byte; const ldim: integer; var matrix3: rrarray);
(*Evaluate column marginals (column sums).*)
var
  i, h: integer;
  sum: real;
begin
for h := 1 to dim do
  begin
  sum := 0;
  for i := 1 to ldim do sum := sum + matrix3[i,h];
  matrix3[0,h] := sum;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function MatrixTrace(var dim: byte; var matrix1: pmatrix): real;
(*Computes the trace of the square matrix <matrix1>*)
var
  i: byte;
  sum: real;
begin
sum := 0;
for i := 1 to dim do
  begin
  sum := sum + matrix1[i,i];
  end;
MatrixTrace := sum;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixI(var dim: byte; var determinant: real; var isigma, sigma, cholfac: pmatrix);
(*Invert variance-covariance matrix.*)
var ic: pmatrix;
begin
if (dim = 1) then
  begin
  if (abs(determinant) > tin) then isigma[1,1] := 1 / determinant; (*The determinant equals the variance in the one-dimensional case.*)
  end
else if (dim = 2) then (*Cuts computation time compared to full-fledged matrix inversion routines.*)
  begin
  if (abs(determinant) > tin) then (*Otherwise singular.*)
    begin
    isigma[1,1] := sigma[2,2] / determinant;
    isigma[2,2] := sigma[1,1] / determinant;
    isigma[1,2] := - (sigma[1,2] / determinant);
    isigma[2,1] := isigma[1,2];
    end;
  end
else if (dim = 3) then (*Matrix inversion based on Cholesky factor: <inverse(Sigma) = inverse(C C') = inverse(C') inverse(C) = inverse(C)' inverse(C)>: see CholeskyInversion.*)
  begin
  (*Compute the transpose (!) of <inverse(C)>.*)
  ic[1,1] := 1 / cholfac[1,1];
  ic[1,2] := - cholfac[2,1] / (cholfac[1,1] * cholfac[2,2]);
  ic[1,3] := ((cholfac[2,1] * cholfac[3,2]) - (cholfac[3,1] * cholfac[2,2])) / (cholfac[1,1] * cholfac[2,2] * cholfac[3,3]);
  ic[2,1] := 0;
  ic[2,2] := 1 / cholfac[2,2];
  ic[2,3] := - cholfac[3,2] / (cholfac[2,2] * cholfac[3,3]);
  ic[3,1] := 0;
  ic[3,2] := 0;
  ic[3,3] := 1 / cholfac[3,3];
  (*Compute <inverse(Sigma) = inverse(C)' inverse(C)>.*)
  MatrixTProductII(dim,ic,isigma);
  end
else (*That is, <dim > 3>.*)
  begin
  CholeskyI(dim,cholfac,ic); (*Inversion of the Cholesky factor; exploits the structure of the Cholesky factor.*)
  MatrixTProductII(dim,ic,isigma);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function CholeskyI(const dim: byte; const c: pmatrix; var ic: pmatrix): boolean;
(*Inversion of the Cholesky factor; exploits the structure of the Cholesky factor;
see: Finn (1974, p. 42), "A General Model for Multivariate Analysis".*)
var
  i, j, k: byte;
  sum: real;
begin
(*Note: resetting <ic> is pointless, since all (relevant) elements of <ic> are manipulated.*)
CholeskyI := false; (*Default.*)
for i := 1 to dim do
  begin
  if (abs(c[i,i]) > tin) then
    begin
    for j := 1 to (i - 1) do (*Up to main diagonal...*)
      begin
      sum := 0;
      for k := j to (i - 1) do sum := sum + (c[i,k] * ic[k,j]); (*The diagonal element of <ic> is not needed.*)
      ic[i,j] := - sum / c[i,i];
      end;
    ic[i,i] := 1 / c[i,i]; (*...main diagonal...*)
    end
  else CholeskyI := true; (*Singular within numerical precision; the resulting row is zero.*)
  for j := (i + 1) to dim do ic[i,j] := 0; (*...above the main diagonal.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ScalarDiv(const dim: byte; var matrix1: pmatrix; const ldim: integer); overload;
(*Divide static array by scalar.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do
    begin
    if (ldim <> 0) then matrix1[i,h] := matrix1[i,h] / ldim
    else matrix1[i,h] := 0;
    end;
  end;
end;

procedure ScalarDiv(const dim1: byte; var matrix1: pmatrix; const dim2: byte); overload;
(*Divide static array by constant scalar.*)
var i, h: byte;
begin
for i := 1 to dim1 do
  begin
  for h := 1 to dim1 do
    begin
    if (dim2 <> 0) then matrix1[i,h] := matrix1[i,h] / dim2
    else matrix1[i,h] := 0;
    end;
  end;
end;

procedure ScalarDiv(const dim1, dim2: byte; var matrix3: rrarray; const ldim2: integer); overload;
(*Divide dynamic array by scalar.*)
var i, h: byte;
begin
for i := 1 to dim1 do
  begin
  for h := 1 to dim2 do
    begin
    if (ldim2 <> 0) then matrix3[i,h] := matrix3[i,h] / ldim2
    else matrix3[i,h] := 0;
    end;
  end;
end;

procedure ScalarDiv(const dim: byte; var matrix3: rrarray; const ldim2: integer); overload;
(*Divide dynamic array by scalar.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do
    begin
    if (ldim2 <> 0) then matrix3[i,h] := matrix3[i,h] / ldim2
    else matrix3[i,h] := 0;
    end;
  end;
end;

procedure ScalarDiv(const ldim1: integer; const dim: byte; var matrix3: rrarray; const ldim2: integer); overload;
(*Divide dynamic array by scalar.*)
var
  i: integer;
  h: byte;
begin
for i := 1 to ldim1 do
  begin
  for h := 1 to dim do
    begin
    if (ldim2 <> 0) then matrix3[i,h] := matrix3[i,h] / ldim2
    else matrix3[i,h] := 0;
    end;
  end;
end;

procedure ScalarDiv(const ldim1, ldim2: integer; var matrix3: rrarray; const ldim: integer); overload;
(*Divide dynamic array by scalar.*)
var
  i: integer;
  h: byte;
begin
for i := 1 to ldim1 do
  begin
  for h := 1 to ldim2 do
    begin
    if (ldim <> 0) then matrix3[i,h] := matrix3[i,h] / ldim
    else matrix3[i,h] := 0;
    end;
  end;
end;

procedure ScalarDiv(const dim: byte; var vector1: pvec; const ldim: integer); overload;
(*Divide static vector by scalar.*)
var i: integer;
begin
for i := 1 to dim do
  begin
  if (ldim <> 0) then vector1[i] := vector1[i] / ldim
  else vector1[i] := 0;
  end;
end;

procedure ScalarDiv(const dim: byte; var vector5: R_Ar; const ldim: integer); overload;
(*Divide dynamic vector by scalar.*)
var i: integer;
begin
for i := 1 to dim do
  begin
  if (ldim <> 0) then vector5[i] := vector5[i] / ldim
  else vector5[i] := 0;
  end;
end;

procedure ScalarDiv(const ldim1: integer; var vector5: R_Ar; const ldim2: integer); overload;
(*Divide dynamic vector by scalar.*)
var i: integer;
begin
for i := 1 to ldim1 do
  begin
  if (ldim2 <> 0) then vector5[i] := vector5[i] / ldim2
  else vector5[i] := 0;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ScalarProd(var dim: byte; var matrix1: pmatrix; var ldim: integer); overload;
(*Multiply static array by scalar.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do matrix1[i,h] := matrix1[i,h] * ldim;
  end;
end;

procedure ScalarProd(var dim: byte; var result, matrix1: pmatrix; var ldim: integer); overload;
(*Multiply static array by scalar.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do result[i,h] := matrix1[i,h] * ldim; 
  end;
end;

procedure ScalarProd(const dim: byte; var result: pmatrix; const matrix1: pmatrix; const real1: real); overload;
(*Multiply static array by scalar.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do result[i,h] := matrix1[i,h] * real1;
  end;
end;

procedure ScalarProd(const dim: byte; var matrix3: rrarray; const ldim: integer; const matrix4: rrarray); overload;
(*Multiply dynamic array by scalar: <matrix3 = ldim * matrix4>.*)
var i, h: byte;
begin
SetArray(matrix3,dim);
for i := 1 to dim do
  begin
  for h := 1 to dim do matrix3[i,h] := matrix4[i,h] * ldim;
  end;
end;

procedure ScalarProd(const dim1, dim2: byte; var matrix3: rrarray; const ldim: integer); overload;
(*Multiply dynamic array by scalar: <matrix3 = ldim * matrix3>.*)
var i, h: byte;
begin
for i := 1 to dim1 do
  begin
  for h := 1 to dim2 do matrix3[i,h] := matrix3[i,h] * ldim;
  end;
end;

procedure ScalarProd(const dim: byte; var matrix3: rrarray; const ldim: integer); overload;
(*Multiply dynamic array by scalar: <matrix3 = ldim * matrix3>.*)
var i, h: byte;
begin
for i := 1 to dim do
  begin
  for h := 1 to dim do matrix3[i,h] := matrix3[i,h] * ldim;
  end;
end;

procedure ScalarProd(var ldim1: integer; var dim: byte; var matrix3: rrarray; var ldim2: integer); overload;
(*Multiply dynamic array by scalar.*)
var
  i: integer;
  h: byte;
begin
for i := 1 to ldim1 do
  begin
  for h := 1 to dim do matrix3[i,h] := matrix3[i,h] * ldim2;
  end;
end;

procedure ScalarProd(var dim: byte; var vector1: pvec; var ldim: integer); overload;
(*Multiply static vector by scalar.*)
var i: integer;
begin
for i := 1 to dim do vector1[i] := vector1[i] * ldim;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixVector(var dim: byte; var vector5: R_Ar; var matrix1: pmatrix);
(*Transform matrix <matrix1> into <vec(matrix1)>, where <matrix1> is assumed to be symmetric.*)
var i, j, h: integer;
begin
i := dim * (dim - 1); (*Number of non-diagonal elements.*)
i := i div 2; (*Number of lower-diagonal elements.*)
i := i + dim; (*Number of lower-diagonal plus diagonal elements.*)
vector5 := nil;
SetLength(vector5,i+1);
h := 0;
for i := 1 to dim do (*Lower-diagonal and diagonal elements.*)
  begin
  for j := 1 to i do
    begin
    inc(h);
    vector5[h] := matrix1[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorMatrix(var dim: byte; var matrix1: pmatrix; var vector5: R_Ar);
(*Invert the transformation of matrix <matrix1> transformed into <vec(matrix1)>, where <matrix1> is assumed to be symmetric.*)
var i, j, h: integer;
begin
h := 0;
for i := 1 to dim do
  begin
  for j := 1 to i do
    begin
    inc(h);
    matrix1[i,j] := vector5[h];
    matrix1[j,i] := matrix1[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorDelete(var dim: byte; var exclusion: pBVec; var vector1, vector2: pvec);
(*Delete coordinate <dim1> of <vector1> of dimension <dim>
and store the result in the vector <vector1> of dimension <dim - 1>.*)
var i, h: byte;
begin
h := 0;
for i := 1 to dim do
  begin
  if not(exclusion[i]) then
    begin
    inc(h);
    vector1[h] := vector2[i];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixDelete(var dim: byte; var exclusion: pBVec; var result, matrix1: pmatrix);
(*Delete row <dim1> and column <dim1> of matrix <matrix1> with dimension <dim x dim>.
On output, the original matrix is unchanged and the result is stored in the matrix
<result> of dimension <dim - 1 x dim - 1>.*)
var i, j, row, column: byte;
begin
row := 0;
for i := 1 to dim do
  begin
  if not(exclusion[i]) then
    begin
    inc(row);
    column := 0;
    for j := 1 to dim do
      begin
      if not(exclusion[j]) then
        begin
        inc(column);
        result[row,column] := matrix1[i,j];
        end;
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure FilterRows(var matrix3: rrarray; var ldim: integer; var dim1, dim2: byte; var matrix4: rrarray); overload;
(*Write the rows of <matrix4> that are a multiple of <dim2> to <matrix3>.*)
var i, j, h: integer;
begin
SetLength(matrix3,dim1+1,trunc(ldim / dim2)+1);
h := 0;
for i := 1 to ldim do
  begin
  if Multiple(i,dim2) then (*Boolean: true = selected row, false = non-selected row.*)
    begin
    inc(h);
    for j := 1 to dim1 do matrix3[j,h] := matrix4[j,i];
    end;
  end;
end;

procedure FilterRows(var matrix6: rrrarray; var ldim1, ldim2: integer; var dim1, dim2: byte; var matrix7: rrrarray); overload;
(*Write the rows of <matrix7> that are a multiple of <dim2> to <matrix6>.*)
var b, i, k, h: integer;
begin
SetLength(matrix6,ldim2+1,dim1+1,trunc(ldim1 / dim2)+1);
b := 0;
for k := 1 to ldim1 do
  begin
  if Multiple(k,dim2) then (*Boolean: true = selected row, false = non-selected row.*)
    begin
    inc(b);
    for i := 1 to ldim2 do (*Selected matrix.*)
      begin
      for h := 1 to dim1 do matrix6[i,h,b] := matrix7[i,h,k];
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetMatrix(const dim1, dim2: byte; var matrix1: pmatrix; const matrix2: pmatrix); overload;
(*Set <matrix1 := matrix2>.*)
var i, j: byte;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do matrix1[i,j] := matrix2[i,j];
  end;
end;

procedure SetMatrix(var ldim1, ldim2: integer; var matrix1, matrix2: pmatrix); overload;
(*Set <matrix1 := matrix2>.*)
var i, j: byte;
begin
for i := 1 to ldim1 do
  begin
  for j := 1 to ldim2 do matrix1[i,j] := matrix2[i,j];
  end;
end;

procedure SetMatrix(var dim: byte; var matrix1: pmatrix; var matrix3: rrarray); overload;
(*Set <matrix1 := matrix3>, where <matrix1> is a static array while <matrix3> is a dynamic array.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix1[i,j] := matrix3[i,j];
  end;
end;

procedure SetMatrix(var dim: byte; var matrix3, matrix4: rrarray); overload;
(*Set <matrix3 := matrix4>.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix3[i,j] := matrix4[i,j];
  end;
end;

procedure SetMatrix(const ldim: integer; const dim: byte; var matrix3: rrarray; const matrix4: rrarray); overload;
(*Set <matrix3 := matrix4>.*)
var i, j: integer;
begin
for i := 1 to ldim do
  begin
  for j := 1 to dim do matrix3[i,j] := matrix4[i,j];
  end;
end;

procedure SetMatrix(var ldim1, ldim2: integer; var matrix3, matrix4: rrarray); overload;
(*Set <matrix3 := matrix4>.*)
var i, j: integer;
begin
for i := 1 to ldim1 do
  begin
  for j := 1 to ldim2 do matrix3[i,j] := matrix4[i,j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetDiagMatrix(const dim: byte; var matrix1: pmatrix; const real1: real); overload;
(*Creates diagonal matrix of order <dim x dim> with value <real1> on the main diagonal.*)
var i, j: byte;
begin
matrix1[1,1] := real1;
for i := 2 to dim do
  begin
  matrix1[i,i] := real1;
  for j := 1 to (i - 1) do
    begin
    matrix1[i,j] := 0;
    matrix1[j,i] := 0;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetIdentityMatrix(const dim: byte; var matrix1: pmatrix); overload;
(*Creates identity matrix of order <dim x dim>.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do
    begin
    if (i = j) then matrix1[i,j] := 1 (*Element on main diagonal.*)
    else matrix1[i,j] := 0;
    end;
  end;
end;

procedure SetIdentityMatrix(var matrix3: rrarray; var dim: byte); overload;
(*Creates identity matrix of order <dim x dim>.*)
var i: byte;
begin
SetArray(matrix3,dim);
for i := 1 to dim do matrix1[i,i] := 1;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetMatrixValue(const dim: byte; var matrix1: pmatrix; const real1: real); overload;
(*Sets <matrix1> to the matrix of dimension <dim x dim> with elements equal to <real1>.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix1[i,j] := real1;
  end;
end;

procedure SetMatrixValue(var dim: byte; var matrix3: rrarray; var real1: real); overload;
(*Sets <matrix3> to the matrix of dimension <dim x dim> with elements equal to <real1>.*)
var i, j: integer;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix3[i,j] := real1;
  end;
end;

procedure SetMatrixValue(var dim1, dim2: byte; var matrix3: rrarray; var real1: real); overload;
(*Sets <matrix3> to the matrix of dimension <dim x dim> with elements equal to <real1>.*)
var i, j: integer;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do matrix3[i,j] := real1;
  end;
end;

procedure SetMatrixValue(var ldim: integer; var dim: byte; var matrix3: rrarray; var real1: real); overload;
(*Sets <matrix3> to the matrix of dimension <dim x dim> with elements equal to <real1>.*)
var i, j: integer;
begin
for i := 1 to ldim do
  begin
  for j := 1 to dim do matrix3[i,j] := real1;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetVectorValue(const dim: byte; var vector1: pvec; const real1: real);
(*Sets <matrix1> to the matrix of dimension <dim x dim> with elements equal to <real1>.*)
var i: byte;
begin
for i := 1 to dim do vector1[i] := real1;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetVector(const dim: byte; var vector1: pvec; const vector2: pvec);
(*Set <vector1 := vector2>.*)
var i: byte;
begin
for i := 1 to dim do vector1[i] := vector2[i];
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixStore(const ldim1: integer; const dim: byte; var autoRE: rrrarray; const ldim2: integer; const matrix3: rrarray); overload;
(*Store matrix.*)
var
  i: integer;
  j: byte;
begin
for i := 1 to ldim1 do
  begin
  for j := 1 to dim do autoRE[i,j,ldim2] := matrix3[i,j];
  end;
end;

procedure MatrixStore(const dim: byte; var autoCovRE: rrrarray; const ldim: integer; const matrix1: pmatrix); overload;
(*Store matrix.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do autoCovRE[i,j,ldim] := matrix1[i,j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure VectorStore(const dim: byte; var autotheta: rrarray; const ldim: integer; const vector1: pvec); overload;
(*Store vector.*)
var i: byte;
begin
for i := 1 to dim do autotheta[i,ldim] := vector1[i];
end;

procedure VectorStore(const dim: byte; var autotheta: rrarray; const ldim: integer; const vector5: R_Ar); overload;
(*Store vector.*)
var i: byte;
begin
for i := 1 to dim do autotheta[i,ldim] := vector5[i];
end;
(*--------------------------------------------------------------------------------------------------*)

procedure SetIMatrix(var dim: byte; var matrix1: pmatrix);
(*Sets <matrix1> to the identity matrix of dimension <dim x dim>.*)
var i, j: byte;
begin
for i := 1 to dim do (*Identity matrix.*)
  begin
  matrix1[i,i] := 1;
  for j := 1 to (i - 1) do
    begin
    matrix1[i,j] := 0;
    matrix1[j,i] := matrix1[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Rescale(const dim: byte; var factor: real; var matrix1: pmatrix); overload;
(*Rescale matrix by factor <factor>.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix1[i,j] := factor * matrix1[i,j];
  end;
end;

procedure Rescale(const dim: byte; const factor: real; const matrix1: pmatrix; var matrix2: pmatrix); overload;
(*Rescale matrix by factor <factor> without changing the original matrix.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to dim do matrix2[i,j] := factor * matrix1[i,j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixNorm(var dim1: byte; var matrix1: pmatrix);
(*Makes the lengths of the column vectors of a <dim1 x dim1> matrix unity by dividing each column <c[i]> by <square root[c(i)'c(i}]>.*)
var
  i, j: integer;
  k: extended;
begin
for i := 1 to dim1 do
  begin
  k := 0;
  for j := 1 to dim1 do k := k + sqr(matrix1[j,i]);
  if k <> 0 then
    begin
    for j := 1 to dim1 do matrix1[j,i] := matrix1[j,i] / sqrt(k);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CorrelationMatrix(const dim: byte; var matrix1: pmatrix);
(*Transforms below-diagonal elements of variance-covariance matrix into correlations.*)
var
  i, j: byte;
  u: real;
begin
for i := 2 to dim do
  begin
  for j := 1 to (i - 1) do
    begin
    u := matrix1[i,i] * matrix1[j,j];
    if (u > 1e-16) then u := sqrt(u)
    else u := 1e-16; (*Its inverse is 1e+16.*)
    matrix1[i,j] := matrix1[i,j] / u; 
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure TransformCov(var dim: byte; var sigma: pmatrix);
(*The aim is to preserve the correlations in the given covariance matrix
but to change the scale. Observe that correlations are scale-invariant.*)
var i, j: byte;
begin
CorrelationMatrix(dim,sigma); (*Transforms below-diagonal elements of variance-covariance matrix into correlations.*)
sigma[1,1] := 1.5; (*Outdegree.*)
for i := 2 to dim do sigma[i,i] := 1;
for i := 2 to dim do
  begin
  for j := 1 to (i - 1) do
    begin
    sigma[i,j] := sigma[i,j] * (sqrt(sigma[i,i]) * sqrt(sigma[j,j]));
    sigma[j,i] := sigma[i,j];
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function CholeskyDecomposition(var matrix1: pmatrix; const dim: byte; var determinant: real): boolean;
(*Cholesky decomposition ("taking the square root") of symmetric, positive definite matrix,
<Sigma = C transpose(C)>, where <C> is the lower-triangular matrix solving the equation.
See random effects paper: latent attachements; Moonan (1957), JASA; Burden & Faires (1997, pp. 410-411), "Numerical Analysis", Algorithm 6.6; Devroye (1986, p. 564).*)
var
  i, j, h: byte;
  sum, u: extended;
begin
CholeskyDecomposition := false;
(*Step 1: diagonal element 1:*)
if (matrix1[1,1] < 0) then
  begin
  CholeskyDecomposition := true;
  exit;
  end
else matrix1[1,1] := sqrt(matrix1[1,1]);
(*Step 2:*)
for j := 2 to dim do matrix1[j,1] := matrix1[j,1] / matrix1[1,1];
(*Step 3:*)
for i := 2 to (dim - 1) do
  begin
  (*Step 4: diagonal elements > 1:*)
  sum := 0.0;
  for h := 1 to (i - 1) do sum := sum - (matrix1[i,h] * matrix1[i,h]);
  u := matrix1[i,i] + sum;
  if (u < 0) then
    begin
    CholeskyDecomposition := true;
    exit;
    end
  else matrix1[i,i] := sqrt(u);
  (*Step 5: non-diagonal elements:*)
  for j := (i + 1) to dim do
     begin
     sum := 0.0;
     for h := 1 to (i - 1) do sum := sum - (matrix1[j,h] * matrix1[i,h]);
     matrix1[j,i] := (matrix1[j,i] + sum) / matrix1[i,i];
     end
  end;
(*Step 6: last diagonal element:*)
sum := 0.0;
for h := 1 to (dim - 1) do sum := sum - (matrix1[dim,h] * matrix1[dim,h]);
u := matrix1[dim,dim] + sum;
if (u < 0) then
  begin
  CholeskyDecomposition := true;
  exit;
  end
else matrix1[dim,dim] := sqrt(u);
(*Step 7:*)
determinant := 1;
for i := 1 to (dim - 1) do
  begin
  determinant := determinant * matrix1[i,i];
  for j := i + 1 to dim do matrix1[i,j] := 0.0;
  end;
determinant := determinant * matrix1[dim,dim];
determinant := abs(determinant); (*Should be redundant.*)
end;

(*--------------------------------------------------------------------------------------------------*)

function Cholesky(const dim: byte; var cholfac: pmatrix; const sigma: pmatrix; var determinant: real): boolean;
(*Cholesky decomposition of positive definite variance-covariance-matrix.*)
var u: extended;
begin
Cholesky := false;
SetMatrix(dim,dim,cholfac,sigma); (*Set <cholfac := sigma>.*)
(*Two aims: (a) evaluate determinant, to be exploited in matrix inversion and normal probability density evaluation, (b) compute Cholesky factor, exploited in sampling from N(0,Sigma).*)
if (dim = 1) then determinant := cholfac[1,1] (*<cholfac> is the variance-covariance matrix here.*)
else if (dim = 2) then determinant := ((cholfac[1,1] * cholfac[2,2]) - sqr(cholfac[2,1]))
else (*<dim > 2>*)
  begin
  Cholesky := CholeskyDecomposition(cholfac,dim,determinant);
  determinant := sqr(determinant); (*Determinant of <Sigma> is determinant of the Cholesky factor to the power of 2.*)
  end;
if (dim < 3) then (*Take the square root, which is done by Cholesky decomposition in the higher-dimensional case.*)
  begin
  if (cholfac[1,1] <= 0) then (*The condition "< 0" is replaced by "<= 0" because division by 0 would rise runtime errors below.*)
    begin
    Cholesky := true;
    exit;
    end
  else cholfac[1,1] := sqrt(cholfac[1,1]);
  if (dim = 2) then (*See paper Cholesky.tex.*)
    begin
    cholfac[1,2] := 0;
    cholfac[2,1] := cholfac[2,1] / cholfac[1,1];
    u := cholfac[2,2] - sqr(cholfac[2,1]);
    if (u < 0) then
      begin
      Cholesky := true;
      exit;
      end
    else cholfac[2,2] := sqrt(u);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure OpenFile(var ascii: textfile; var filename: shortstring);
(*Assign file name to textfile variable and rewrite file.*)
begin
assignfile(ascii,filename);
rewrite(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure OpenMFile(const boolean1: boolean; var filename: shortstring; var ascii: textfile);
(*If boolean condition is satisfied, prepare file for writing assignments.*)
begin
if boolean1 then OpenFile(ascii,filename);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure UpdateFrequencyDistribution(var real1, real2: real; var matrix4: rrarray);
(*Update frequency distribution by adding one data point.*)
var i: integer;
begin
i := trunc(real1 / real2) + 1;
if (real1 < 0) then matrix4[0,i] := matrix4[0,i] + 1
else matrix4[1,i] := matrix4[1,i] + 1;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ReadFile(var ascii: textfile; const filename: shortstring); overload;
(*Open and reset file.*)
begin
assignfile(ascii,filename);
if FileExists(filename) then reset(ascii);
end;

procedure ReadFile(var ascii: textfile; const filename: shortstring; var boolean1: boolean); overload;
(*Open and reset file.*)
begin
assignfile(ascii,filename);
if FileExists(filename) then
  begin
  boolean1 := true;
  reset(ascii);
  end
else boolean1 := false;  
end;

(*--------------------------------------------------------------------------------------------------*)

procedure AppendToFile(var ascii: textfile; var filename: shortstring);
(*Assign file name to textfile variable and append to file.*)
begin
assignfile(ascii,filename);
if FileExists(filename) then append(ascii)
else rewrite(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure WriteReal(var ascii: textfile; var real1: real); overload;
(*Writes real number to file to produce time series.*)
begin
writeln(ascii,' ',real1:10:10);
end;

procedure WriteReal(var ascii: textfile; const real1: extended); overload;
(*Writes real number to file to produce time series.*)
begin
writeln(ascii,' ',real1:10:10);
end;

procedure WriteReal(var ascii: textfile; var real1, real2: real); overload;
(*Writes real number to file to produce time series.*)
begin
writeln(ascii,' ',real1:10:10,' ',real2:10:10);
end;

procedure WriteReal(var boolean1: boolean; var ascii: textfile; var real1: real); overload;
(*Writes real number to file to produce time series.*)
begin
if boolean1 then writeln(ascii,' ',real1:10:10);
end;

procedure WriteReal(var boolean1: boolean; var ascii: textfile; var real1, real2: real); overload;
(*Writes real number to file to produce time series.*)
begin
if boolean1 then writeln(ascii,' ',real1:10:10,' ',real2:10:10);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixWrite(var ascii: textfile; const dim1, dim2: byte; const matrix1: pmatrix); overload;
(*Writes real matrix <pmatrix> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
begin
writeln(ascii);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    write(ascii,matrix1[i,j]:10:4);
    end;
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWrite(var ascii: textfile; var dim: byte; var ldim: integer; var matrix3: rrarray); overload;
(*Writes real matrix of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
begin
writeln(ascii);
for i := 1 to dim do
  begin
  for j := 1 to ldim do
    begin
    write(ascii,matrix3[i,j]:10:4);
    end;
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWrite(var ldim: integer; var dim: byte; var ascii: textfile;  var matrix3: rrarray); overload;
(*Writes real matrix of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
begin
for i := 1 to ldim do
  begin
  for j := 1 to dim do
    begin
    write(ascii,matrix3[i,j]:10:4);
    end;
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWrite(var ascii: textfile; const dim1, dim2: byte; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
(*Writes real matrix <pmatrix> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
begin
writeln(ascii);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    write(ascii,matrix1[i,j]:ldim1:ldim2);
    end;
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWrite(const ldim: integer; const dim: byte; var ascii: textfile; const matrix3: rrarray; const ldim1, ldim2: integer); overload;
(*Writes real matrix of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
begin
for i := 1 to ldim do
  begin
  for j := 1 to dim do
    begin
    write(ascii,matrix3[i,j]:ldim1:ldim2);
    end;
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWrite(var ascii: TFileStream; const dim1, dim2: byte; const matrix1: pmatrix); overload;
(*Writes real matrix <pmatrix> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
x : string;
begin
//writeln(ascii);
ascii.Write(pchar(eol)^,length(eol));
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    x := format('%10.4f',[matrix1[i,j]]);
    ascii.Write(pchar(x)^,length(x));
    end;
  //writeln(ascii);
  ascii.Write(pchar(eol)^,length(eol));
  end;
//writeln(ascii);
ascii.Write(pchar(eol)^,length(eol));
end;

procedure MatrixWrite(var ascii: TFileStream; const dim1, dim2: byte; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
(*Writes real matrix <pmatrix> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
x: string;
begin
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    x := format('%*.*f',[ldim1,ldim2,matrix1[i,j]]);
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixWriteLight(var ascii: textfile; const dim1, dim2: byte; const matrix1: pmatrix); overload;
(*Writes real matrix <pmatrix> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    write(ascii,matrix1[i,j]:10:4);
    end;
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWriteLight(var ascii: TFileStream; const dim1, dim2: byte; const matrix1: pmatrix); overload;
(*Writes real matrix <pmatrix> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var i, j: integer;
x: string;
begin
for i := 1 to dim1 do
  begin
  for j := 1 to dim2 do
    begin
    x := format('%10.4f',[matrix1[i,j]]);
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
end;
(*--------------------------------------------------------------------------------------------------*)

procedure MatrixWritePlot(var boolean1: boolean; var ldim: integer; var dim: byte;  var ascii: textfile; var matrix3: rrarray); overload;
(*Writes real matrix <rrarray> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var
  count, i: integer;
  h: byte;
begin
if boolean1 then
  begin
  count := 0;
  for i := 1 to ldim do
    begin
    for h := 1 to dim do
      begin
      if (count >= 950) then
        begin
        (*Comment:
        ASCII files store up to 1024 characters on one line; strings exceeding 1024 characters are cut after the
        1024th character and continued on the following line; this leads to problems when reading the files later.*)
        writeln(ascii);
        count := 0;
        end
      else count := count + 10; (*10 (!) is the number of characters written by the statement.*)
      write(ascii,' ',matrix3[i,h]:9:6);
      end;
    end;  
  writeln(ascii);
  end;  
end;

procedure MatrixWritePlot(var ldim: integer; var dim: byte;  var ascii: textfile; var matrix3: rrarray); overload;
(*Writes real matrix <rrarray> of dimension <dim1 x dim2> to ASCII file indexed by <ascii>.*)
var
  count, i: integer;
  h: byte;
begin
count := 0;
for i := 1 to ldim do
  begin
  for h := 1 to dim do
    begin
    if (count >= 950) then
      begin
      (*Comment:
      ASCII files store up to 1024 characters on one line; strings exceeding 1024 characters are cut after the
      1024th character and continued on the following line; this leads to problems when reading the files later.*)
      writeln(ascii);
      count := 0;
      end
    else count := count + 10; (*10 (!) is the number of characters written by the statement.*)
    write(ascii,' ',matrix3[i,h]:9:6);
    end;
  end;
writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixWriteCov(const dim: byte; var ascii: textfile; var matrix1: pmatrix); overload;
(*Write the vector of non-redundant elements of variance-covariance matrix <matrix1> to the file,
where the vector was obtained by applying the vec operator to <matrix1> and then deleting the above-diagonal elements.*)
var i, j: byte;
begin
for j := 1 to dim do (*For each column do...*)
  begin
  for i := j to dim do write(ascii,matrix1[i,j]:10:6); (*...write diagonal and below-diagonal values to the file.*)
  end;
writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixWriteTriangle(const dim: byte; var ascii: textfile; const matrix1: pmatrix); overload;
(*Writes lower triangle (including diagonal) of symmetric matrix to file.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to i do write(ascii,' ',matrix1[i,j]:10:4);
  writeln(ascii);
  end;
writeln(ascii);
end;

procedure MatrixWriteTriangle(const dim: byte; var ascii: textfile; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
(*Writes lower triangle (including diagonal) of symmetric matrix to file.*)
var i, j: byte;
begin
for i := 1 to dim do
  begin
  for j := 1 to i do write(ascii,' ',matrix1[i,j]:ldim1:ldim2);
  writeln(ascii);
  end;
writeln(ascii);
end;
procedure MatrixWriteTriangle(const dim: byte; var ascii: TFileStream; const matrix1: pmatrix); overload;
(*Writes lower triangle (including diagonal) of symmetric matrix to file.*)
var i, j: byte;
x: string;
begin
for i := 1 to dim do
  begin
  for j := 1 to i do begin
    x := format('%s%10.4f',[' ',matrix1[i,j]]);
    ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
end;

procedure MatrixWriteTriangle(const dim: byte; var ascii: TFileStream; const matrix1: pmatrix; const ldim1, ldim2: integer); overload;
(*Writes lower triangle (including diagonal) of symmetric matrix to file.*)
var i, j: byte;
x : string;
begin
for i := 1 to dim do
  begin
  for j := 1 to i do begin
    x := format('%s%*.*f',[' ',ldim1,ldim2,matrix1[i,j]]);
    ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
end;
(*--------------------------------------------------------------------------------------------------*)

procedure VectorWrite(var ascii: textfile; var dim: byte; var vector1: pvec);
(*Writes estimates of fixed effects (parameters) to file to produce trace plot.*)
var i: byte;
begin
for i := 1 to dim do write(ascii,' ',vector1[i]:10:6);
writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MatrixRead(var ascii: textfile; var filename: shortstring; var matrix3: rrarray; var ldim: integer; var dim: byte);
(*Read matrix <matrix3> of dimension <ldim x dim> from input file <ascii> with name <filename>.*)
var
  i, j: integer;
  u: real;
begin
ReadFile(ascii,filename); (*Open and reset file.*)
SetArray(matrix3,ldim,dim);
for i := 1 to ldim do
  begin
  for j := 1 to dim do
    begin
    read(ascii,u);
    matrix3[i,j] := u;
    end;
  readln(ascii);
  end;
closefile(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

function MatrixURead(const filename: string; const dim: byte; var matrix3: rrarray): boolean;
(*Read matrix with an unknown number of rows and a known number of columns.*)
var
  i, j: integer;
  u: real;
  condition: boolean;
  f: textfile;
begin
ReadFile(f,filename,condition); (*The boolean indicates whether the attempt to reset the file was succesful.*)
if condition then
  begin
  matrix3 := nil;
  i := 0;
  while not(eof(f)) do (*Row <i>.*)
    begin
    Inc(i);
    SetLength(matrix3,i+1,dim+1); (*Note: one row is added, but the existing rows are preserved.*)
    for j := 1 to dim do (*Column <j>.*)
      begin
      if eoln(f) then condition := false (*Failed to read <dim> column elements.*)
      else read(f,u);
      matrix3[i,j] := u;
      end;
    readln(f);
    end;
  closefile(f);
  end;
MatrixURead := condition; (*If either resetting or reading the file was problematic, the boolean is false.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Get_Cov_Vector(const dim: byte; var sigma: pmatrix; const vector1: pvec);
(*Extract the variance-covariance matrix <sigma> from the vector of non-redundant elements of <sigma>,
where the vector was obtained by applying the vec operator to <sigma> and then deleting the above-diagonal elements.*)
var i, j, h: byte;
begin
h := 0;
for j := 1 to dim do (*For each column do...*)
  begin
  for i := 1 to (j - 1) do sigma[i,j] := 0; (*Above-diagonal elements.*)
  for i := j to dim do (*...diagonal and below-diagonal elements.*)
    begin
    Inc(h);
    sigma[i,j] := vector1[h]; (*That is, <sigma[j+1,j]> up to, and including, <sigma[dim,j]>.*)
    end;
  end;
for i := 2 to dim do (*Create above-diagonal triangle of variance-covariance matrix.*)
  begin
  for j := 1 to (i - 1) do sigma[j,i] := sigma[i,j];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function NonSingular(var dim: byte; var sigma: pmatrix): boolean;
(*Test whether system is non-singular.*)
var i: byte;
begin
NonSingular := false;
for i := 1 to dim do
  begin
  if (sigma[i,i] <= 0) then NonSingular := true;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function CauchySchwartz(var dim: byte; var sigma: pmatrix): boolean;
(*Test whether the variance-covariance matrix satisfies the Cauchy-Schwartz inequality.*)
var
  i, j: byte;
  criterion: real;
begin
CauchySchwartz := false;
for i := 2 to dim do
  begin
  for j := 1 to (i - 1) do
    begin
    criterion := sqrt(sigma[i,i]) * sqrt(sigma[j,j]);
    if (sigma[i,j] > criterion) then CauchySchwartz := true;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function Multiple(const ldim: integer; const dim: byte): boolean;
(*Tests whether <ldim> is a multiple of <dim>, that is, whether
<ldim = dim x i> holds for some integer <i>.*)
begin
if (trunc(ldim / dim) = (ldim / dim)) then Multiple := true
else Multiple := false;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetArray(var vector6: R_Ar; const dim: byte); overload;
begin
vector6 := nil;
SetLength(vector6,dim+1);
end;

procedure SetArray(var vector6: R_Ar; const ldim: integer); overload;
begin
vector6 := nil;
SetLength(vector6,ldim+1);
end;

procedure SetArray(var matrix3: rrarray; const dim: byte); overload;
begin
matrix3 := nil;
SetLength(matrix3,dim+1,dim+1);
end;

procedure SetArray(var matrix3: rrarray; const dim1, dim2: byte); overload;
begin
matrix3 := nil;
SetLength(matrix3,dim1+1,dim2+1);
end;

procedure SetArray(var matrix3: rrarray; const ldim: integer; const dim: byte); overload;
begin
matrix3 := nil;
SetLength(matrix3,ldim+1,dim+1);
end;

procedure SetArray(var matrix3: rrarray; const dim: byte; const ldim: integer); overload;
begin
matrix3 := nil;
SetLength(matrix3,dim+1,ldim+1);
end;

procedure SetArray(var matrix3: rrarray; const ldim1, ldim2: integer); overload;
begin
matrix3 := nil;
SetLength(matrix3,ldim1+1,ldim2+1);
end;

procedure SetArray(var threedim: rrrarray; const dim: byte); overload;
begin
threedim := nil;
SetLength(threedim,dim+1,dim+1,dim+1);
end;

procedure SetArray(var threedim: rrrarray; const ldim: integer; const dim: byte); overload;
begin
threedim := nil;
SetLength(threedim,ldim+1,dim+1,dim+1);
end;

procedure SetArray(var threedim: rrrarray; const dim: byte; const ldim: integer); overload;
begin
threedim := nil;
SetLength(threedim,dim+1,dim+1,ldim+1);
end;

procedure SetArray(var threedim: rrrarray; const ldim1: integer; const dim: byte; const ldim2: integer); overload;
begin
threedim := nil;
SetLength(threedim,ldim1+1,dim+1,ldim2+1);
end;

procedure SetArray(var booleans: boarray; const ldim: integer); overload;
begin
booleans := nil;
SetLength(booleans,ldim+1);
end;

procedure SetZero(var matrix: rrarray; const d1, d2: integer);
(* sets matrix[i,j] := 0.0 for 0<=1<=d1, 0<=j<=d2 *)
var i,j: integer;
begin
  for i:=d1 downto 0 do for j:=d2 downto 0 do matrix[i,j] := 0.0;
end;   

(*--------------------------------------------------------------------------------------------------*)

end.


